;
; $Id: ppm.scm,v 1.11 2003/01/19 08:14:32 jp Exp $
;
; ppm.scm -- Functions and classes for reading and displaying ppms
; and other bitmaps.


(module ppm mzscheme
  (require 
   (lib "class.ss" "mzlib")
   (lib "mred.ss" "mred")
   (lib "etc.ss" "mzlib")
   )
  (provide (all-defined))
  
  
  (define (show-ppm . args )
    ; 
    ; show-ppm: [filename] -> void
    ;
    ; Show a ppm file in a window.  No frills. Takes a filename as
    ; optional first arg.  Otherwise prompts for a file.
    ;
    (let* ((filename (if (null? args) 
			 (get-file "Select a PPM File")
			 (car args)))
	   (bm (read-ppm filename))
	   (frame (instantiate frame% (filename)
                    (width (send bm get-width))
                    (height (send bm get-height))))
	   (canvas (instantiate canvas% () 
                     (parent frame) 
                     (paint-callback 
                      (lambda (c dc) 
                        (send dc draw-bitmap bm 0 0 'solid 
                              (make-object color% 0 0 0) #f)))
                     )))
      (send frame show #t)))
  
  (define (read-ppm . args)
    ;
    ; read-ppm: [filename] -> bitmap%
    ;
    ; Reads a ppm.  If a filename is given read from that filename,
    ; otherwise from current-input-port
    ;
    (if (not (null? args))
	(with-input-from-file (car args) read-ppm)
	(let* ((tag (read-no-comments))
	       (width (read-no-comments))
	       (height (read-no-comments))
	       (max-intensity (read-no-comments))
	       (bm (make-object bitmap% width height #f)))
	  ;(eat-whitespace-and-comments)
	  (read-line)
	  (case tag          
	    ((P2) (read-ascii-pgm-data bm max-intensity))
	    ((P3) (read-ascii-ppm-data bm max-intensity))
	    ((P5) (read-raw-pgm-data bm max-intensity))
	    ((P6) (read-raw-ppm-data bm max-intensity))
	    (else
	     (error (format "Unknown PPM tag: ~S" tag))
	     ))
	  bm)))
  
  (define (read-no-comments)
    ; read-no-comments: void -> obj
    ; Consume any and all comments and whitespace, then
    ; read the next object from current-input-port
    (eat-whitespace-and-comments)
    (read))
  
  (define (eat-whitespace-and-comments)
    ; eat-whitespace-and-comments: void -> void
    ; consume any and all whitespace and comments, then return
    (let ((c (peek-char)))
      (cond
        ((eof-object? c))
        ((eq? c #\#) 
         (read-line)
         (eat-whitespace-and-comments))
        ((char-whitespace? c)
         (read-char)
         (eat-whitespace-and-comments)))))
  
  (define (read-raw-ppm-data bm max-intensity)
    ; read-raw-ppm-data: bitmap% integer -> void
    ; Read the pixel data for a P6 raw ppm from current-input port and
    ; instantiate it in the given bitmap.
    (let ((bdc (make-object bitmap-dc% bm))
          (width (send bm get-width))
          (height (send bm get-height))
          (color (make-object color%)))
      (do ((y 0 (+ y 1)))
        ((or (at-eof?)
             (= y height)))
        (do ((x 0 (+ x 1)))
          ((or (at-eof?) (= x width) ))
	  (let* ((r (read-ppm-byte))
		 (g (read-ppm-byte))
                 (b (read-ppm-byte)))
	    (send color set 
		  (conform-color r max-intensity) 
		  (conform-color g max-intensity) 
		  (conform-color b max-intensity))
	    (send bdc set-pixel x y color))))
      (send bdc set-bitmap #f)))
  
  (define (read-raw-pgm-data bm max-intensity)
    ; read-raw-ppm-data: bitmap% integer -> void
    ; Read the pixel data for a P5 raw pgm from current-input port and
    ; instantiate it in the given bitmap.
    (let ((bdc (make-object bitmap-dc% bm))
          (width (send bm get-width))
          (height (send bm get-height))
          (color (make-object color%)))
      (do ((y 0 (+ y 1)))
        ((= y height))
        (do ((x 0 (+ x 1)))
          ((= x width) )
	  (let* ((c (read-ppm-byte))
		 (v (conform-color c max-intensity)))
	    (send color set v v v)
	    (send bdc set-pixel x y color))))
      (send bdc set-bitmap #f)))
  
  (define (read-ascii-ppm-data bm max-intensity)
    ; read-raw-ppm-data: bitmap% integer -> void
    ; Read the pixel data for a P3 ascii ppm from current-input port and
    ; instantiate it in the given bitmap.
    (let ((bdc (make-object bitmap-dc% bm))
          (width (send bm get-width))
          (height (send bm get-height))
          (color (make-object color%)))
      (do ((y 0 (+ y 1)))
        ((= y height))
        (do ((x 0 (+ x 1)))
          ((= x width) )
	  (let* ((r (read-ppm-val))
		 (g (read-ppm-val))
                 (b (read-ppm-val)))
	    (send color set 
		  (conform-color r max-intensity) 
		  (conform-color g max-intensity) 
		  (conform-color b max-intensity))
	    (send bdc set-pixel x y color))))
      (send bdc set-bitmap #f)))
  
  
  (define (read-ascii-pgm-data bm max-intensity)
    ; read-raw-ppm-data: bitmap% integer -> void
    ; Read the pixel data for a P2 ascii pgm from current-input port and
    ; instantiate it in the given bitmap.
    (let ((bdc (make-object bitmap-dc% bm))
          (width (send bm get-width))
          (height (send bm get-height))
          (color (make-object color%)))
      (do ((y 0 (+ y 1)))
        ((= y height))
        (do ((x 0 (+ x 1)))
          ((= x width) )
	  (let* ((c (read-ppm-val))
		 (v (conform-color c max-intensity)))
	    (send color set v v v)
	    (send bdc set-pixel x y color))))
      (send bdc set-bitmap #f)))
  
  
  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  
  (define (scale-bitmap bm target-width target-height)
    ; scale-bitmap: bitmap% real real -> bitmap%
    ; Scale the given bitmap to the given dimensions.
    (let* ((oldwidth (send bm get-width))
           (oldheight (send bm get-height)) 
           (width (round target-width))
           (height (round (if target-height target-height
                              (* oldheight (/ width oldwidth)))))
           (sx (/ oldwidth width))
           (sy (/ oldheight height))
           (result (instantiate bitmap% (width height)))
           (old-bdc (make-object bitmap-dc% bm))
           (result-bdc (make-object bitmap-dc% result))
           (color (make-object color%))
           )
      (do ((row 0 (+ row 1)))
        ((= row height))
        (do ((col 0 (+ col 1)))
          ((= col width))
          (send old-bdc get-pixel (* col sx) (* row sy) color)
          (send result-bdc set-pixel  col row color) ))
      (send old-bdc set-bitmap #f)
      (send result-bdc set-bitmap #f)
      result))
  
  (define (scale-bitmap-by bm scale-factor)
    ; scale-bitmap-by: bitmap% real -> bitmap%
    ; Scale the given bitmap uniformly by scale-factor.
    (scale-bitmap bm
                  (round (* scale-factor (send bm get-width)))
                  (round (* scale-factor (send bm get-height)))))
  
  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  
  
  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  ; bitmap panel
  
  (define bitmap-panel%
    ; bitmap-panel%
    ; A class to provide a labelled, zoomable bitmap display.
    (class vertical-panel%
      (init-field bm)
      (init-field label)
      (init-field (initial-width (send bm get-width)))
      (field (scaled-bm (scale-bitmap bm initial-width #f)))
                                      
      (super-instantiate () (min-width initial-width))
      
      (define (canvas-callback canvas dc)
        (send dc draw-bitmap scaled-bm 0 0 'solid (make-object color% 0 0 0) #f))
      
      (field (bm-canvas 
              (instantiate canvas% () 
                (parent this)
                (paint-callback canvas-callback) 
                (stretchable-width #f)
                (stretchable-height #f))))
      
      (field (msg (instantiate message% () (parent this) (label label))))
      
      (define/public (zoom-by factor)
	(let* ((current-width (send scaled-bm get-width))
	       (target-width (* current-width factor)))
	  ; don't let it scale down below 5, because round-off error
	  ; starts to screw things up.
	  (set! scaled-bm (scale-bitmap bm (if (< target-width 5) 5 target-width) #f))
	  (fix-geometry)
	  (send bm-canvas on-paint)))
      
      (define/public (fix-geometry)
        (send bm-canvas min-width (send scaled-bm get-width))
        (send bm-canvas min-height (send scaled-bm get-height)))
      
      (define/public (new-bitmap)
        (zoom-by 1)
        (fix-geometry)
        (send bm-canvas on-paint))
    
      (fix-geometry)
      ))
  
  
  
  
  (define ppm-panel%
    ; A bitmap-panel% that reads ppms.
    (class bitmap-panel%
      (init-field (filename ""))
      (inherit-field bm)
      
      (inherit new-bitmap)
      
      (public reload!)
      (define (reload!)
        (set! bm (read-ppm filename))
        (new-bitmap) )
      
      (if (file-exists? filename)
          (super-instantiate () (bm (read-ppm filename)))
          (super-instantiate ()))))
  
   
  (define h-ppm-array%
    ; A zoomable horizontal array of labelled ppms.
    (class horizontal-panel%
      (init-field (files ()))
      (init-field (labels ()))
      (init-field (initial-width 0))
      
      (inherit delete-child get-children)
      
      (define/public (reload)
        (for-each (lambda (child) (send child reload!))
                  (get-children)))
      
      (define/public (zoom-by zoom-factor)
        (for-each (lambda (child) (send child zoom-by zoom-factor))
                  (get-children)))
      
      (super-instantiate ())
      
      (define/public (set-files new-files)
        (set! files new-files))
      
      (define/public (set-labels new-labels)
        (set! labels new-labels))
      
      (define/public (reinit)
        (for-each (lambda (c) (delete-child c)) (get-children))
        (for-each (lambda (f t)
                    (instantiate ppm-panel% () 
                      (parent this) 
                      (filename f) 
                      (label t)
                      (initial-width initial-width)
                      (style '(border))
                      (stretchable-height #f)
                      (stretchable-width #f)
                      (border 2)))
                  files labels))
      ))
  
    
  
  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  ; Utility routines
  (define (read-ppm-byte)
    (let ((c (read-char)))
      (if (eof-object? c) 0
          (char->integer c))))
  
  (define (read-ppm-val)
    (let ((c (read)))
      (if (eof-object? c) 0 c)))
  
  
  (define (at-eof?)
    (eof-object? (peek-char)))
  
  (define (conform-color x max)
    (if (= max 255) 
        x
        (round (* (/ x max) 255))))
  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    
  
  )
