#|
 $Id: properties-panel.scm,v 1.6 2003/01/18 06:21:10 jp Exp $

 The properties-panel is a control that provides uniform interface for
 editing the values of a set of related named properties.  It provides
 a panel with one property per row.  Each property has a control
 for editing its value.  Several simple control types (text-field, slider, choice, check-box)
 are provided, as well as a method for adding properties with arbitrary 
 control types.

 
 
 Usage:
  
   (define p (instantiate properties-panel% () ...))
    
 To add a property with a text-field control:
    
   (send p add-text-property 'foo "Foo" "A string for foo")
   
 The above adds a text-field property with name 'foo, label "Foo" and
 initial value "A string for Foo".  Similar methods exist for the other control
 types.  See the individual function documentation below.
 
 To get a property's value
 
   (send p get-value 'foo)
   
 To set a value:
 
   (send p set value 'foo "A new string for foo")
    
 |#

(module properties-panel mzscheme
  (require (lib "mred.ss" "mred")
           (lib "class.ss" "mzlib")
           "tagged-slider.scm")
  
  (provide (all-defined))
  
  (define-struct property (message control get-value-thunk set-value-thunk))
  
  
  (define properties-panel%
    (class vertical-panel%
      
      (super-instantiate ())
      
      (field (properties (make-hash-table)))
      (field (min-label-width 0))
      (field (last-row #f))
      
      
      (define (add-row)
        (set! last-row
              (instantiate horizontal-panel% () 
                (parent this)
                (stretchable-height #f))))
      
      (add-row)
      
      
      (define/public (add-custom-property name label control-promise get-value-thunk set-value-thunk)
        ; 
        ; Generic method for adding a property.  Use specialized methods (e.g. add-text-property)
        ; when possible.
        ;
        ; Params:
        ;    name             = The lookup name of the property (generally a symbol)
        ;    label            = The string label of the property in the panel.
        ;    control-promise  = A promise which will evaluate to a control for the property
        ;    get-value-thunk  = A function which takes the control as a parameter and
        ;                       returns the value of the parameter.
        ; 
        ; The method adds the label and control to the last row of the panel, and
        ; adds a property structure to the hash table, then adds a new (empty) row to the table.
        ;
        (let ((message (instantiate message% () 
                         (parent last-row) 
                         (stretchable-width #f)
                         (label label))))
          (adjust-label-widths message)
          (hash-table-put! properties name
                           (make-property message (force control-promise) get-value-thunk set-value-thunk))
          (add-row)))
      
      (define/public (delete-property name)
        ;
        ; Deletes the named property from the panel
        ;
        (let ((p (hash-table-get properties name)))
          (hash-table-remove! properties name)
          (send this delete-child (send (property-control p) get-parent))))
      
      (define/public (add-text-property name label init-value)
        ;
        ; Add a text property to the panel.
        ; Params:
        ;   name       = The lookup name of the property (usually a symbol).
        ;   label      = The label string for the property in the panel.
        ;   init-value = The initial value of the property (a string).
        ;
        (add-custom-property name label
                             (delay
                               (instantiate text-field% () 
                                 (parent last-row) 
                                 (label #f)
                                 (init-value init-value)
                                 (callback (lambda (t e) #f))))
                             (lambda (c) (send c get-value))
                             (lambda (c v) (send c set-value v))))
      
      
      (define/public (add-choice-property name label choices init-value)
        ;
        ; Add a choice (drop-down menu) to the panel.
        ; Params:
        ;   name       = The lookup name of the property (usually a symbol).
        ;   label      = The label string for the property in the panel.
        ;   choices    = A list of string choices for the menu.
        ;
        (add-custom-property name label
                             (delay 
                               (instantiate choice% ()
                                 (parent last-row)
                                 (label #f)
                                 (choices choices)
                                 (callback (lambda (c e) #f))))
                             (lambda (c) (list-ref choices (send c get-selection)))
                             (lambda (c v) 
                               (if (integer? v)
                                   (send c set-selection v)
                                   (send c set-selection (position v choices))
                                   )))
        (set-value name init-value))
                             
      (define/public (add-slider-property name label min-value max-value init-value)
        ;
        ; Add a slider to the panel.
        ; Params:
        ;   name       = The lookup name of the property (usually a symbol).
        ;   label      = The label string for the property in the panel.
        ;   min-value  = The minimum value of the slider
        ;   max-value  = The maximum value of the slider.
        ;   init-value = The initial value of the slider.
        ;
        (add-custom-property name label
                             (delay
                               (instantiate slider% ()
                                 (parent last-row)
                                 (label #f)
                                 (min-value min-value)
                                 (max-value max-value)
                                 (init-value init-value)
                                 (callback (lambda (s e) #f))))
                             (lambda (s) (send s get-value))
                             (lambda (s v) (send s set-value v))))
      
      (define/public (add-tagged-slider-property name label min-value max-value init-value decimal-places
                                                 translation-fn)
        ;
        ; Add a tagged-slider to the panel.
        ; Params:
        ;   name       = The lookup name of the property (usually a symbol).
        ;   label      = The label string for the property in the panel.
        ;   min-value  = The minimum value of the slider
        ;   max-value  = The maximum value of the slider.
        ;   init-value = The initial-value of the slider.
        ;   decimal-places = The number of decimal-places to display in the slider's tag
        ;   translation-fn =  The string translation function
        ;
        (add-custom-property name label
                             (delay
                               (instantiate tagged-slider% ()
                                 (parent last-row)
                                 (min-value min-value)
                                 (max-value max-value)
                                 (init-value init-value)
                                 (decimal-places decimal-places)
                                 (string-translator-fn translation-fn)
                                 (callback (lambda (s e) #f))))
                             (lambda (s) (send s get-value))
                             (lambda (s v) (send s set-value v))))
     
      
      (define/public (add-check-box-property name label init-value)
        ;
        ; Add a check-box to the panel.
        ; Params:
        ;   name       = The lookup name of the property (usually a symbol).
        ;   label      = The label string for the property in the panel.
        ;   init-value = Whether or not the box is initially checked.
        (add-custom-property name label
                             (delay
                               (instantiate check-box% () 
                                 (parent last-row) 
                                 (label "")
                                 (callback (lambda (t e) #f))))
                             (lambda (c) (send c get-value))
                             (lambda (c v) (send c set-value v))))
      
      
      
      (define (adjust-label-widths new-label)
        ;
        ; Adjust the widths of all the labels in the panel to accommodate
        ; the width of a new label.  Or, set the min width of the new label to
        ; be the overall min-label-width
        ; Params:
        ;   new-label = a message% control for a new label.
        ;
        (let ((w (send new-label get-width)))
          (cond ((< min-label-width w)
                 (set! min-label-width w)
                 (hash-table-for-each properties 
                                      (lambda (k p) (send (property-message p) min-width w))))
                (else
                 (send new-label min-width min-label-width)))))
      
      (define (position value list)
        ;
        ; position - find the position of a value in a list, using equal?
        ;
        (let loop ((l list)
                   (pos 0))
          (cond 
            ((null? l) 
             (error (format "No such value ~S in ~S" value list)))
            ((equal? value (car l)) 
             pos)
            (else
             (loop (cdr l) (+ 1 pos))))))
      
      (define/public (get-value name)
        ;
        ; Get the value of the property with the given name
        ;
        (let* ((p (hash-table-get properties name))
               (get-value (property-get-value-thunk p))
               (control (property-control p)))
          (get-value control)))
      
      (define/public (get-values)
        ;
        ; Get the values of all properties as an alist of paired
        ; names and values in arbitrary order.
        ;       
        (hash-table-map properties
                        (lambda (key val) (list key (send this get-value key)))))
      
      (define/public (set-value name value)
        ;
        ; Set the value of the named property
        ;
        (let* ((p (hash-table-get properties name))
               (set-value (property-set-value-thunk p))
               (control (property-control p)))
          (set-value control value)))
      
      (define/public (set-values values)
        ;
        ; Set the values of several properties, by passing an alist
        ; of names and values. e.g '((a 1) (b 2) (c 3))
        ;
        (let loop ((l values))
          (unless (null? l)
            (let* ((pair (car l))
                   (name (car pair))
                   (value (cadr pair)))
              (set-value name value)
              (loop (cdr l))))))
      ))
  
  
  (define (test-prop-panel)
    (define f (instantiate frame% ("Test")
                (stretchable-height #f)))
    (define p (instantiate properties-panel% () 
                (parent f)
                (stretchable-height #f)))
    (instantiate grow-box-spacer-pane% () (parent f))
    
    (send p add-text-property 'foo "Foo" "11")
    (send p add-choice-property 'color "Color" '("Red" "Blue" "Green") "Blue")
    (send p add-slider-property 'volume "Volume" 0 100 10)
    (send p add-check-box-property 'on "On" #t)
    (send p add-text-property 'bar "Bar" "22")
    (send p add-tagged-slider-property 'angle "Angle" -3.14159 3.14159 0 3)
    
    (send f show #t)
    p
    )
  
  )