#ifndef MTL_FLATTEN_ITERATOR_H
#define MTL_FLATTEN_ITERATOR_H

#include <iterator>

namespace mtl {


/*
  This class acts as a flattened one-dimensional STL iterator that
  traverses element-by-element over any STL container of STL
  containers of elements.  It should behave correctly even if some or
  all sub-containers are empty or if the top-level container is empty,
  which is ensured by maintaining the internal state such that in user
  code the flatten_iterator is always either equal to the end()
  flatten_iterator or else its internal 2D and 1D iterators point to a
  valid container and element, respectively.

  An end() flatten_iterator is defined to be one whose top-level
  iterator (curr2D) points to the end() of the top-level container
  (end2D); in such a case the values of all OneD_iter parameters and
  member variables are undefined.

  Note that a flatten_iterator requires somewhat greater overhead than
  a pair of 2D and 1D iterators, since each ++ operation requires a
  check for the end of both the 1D dimension (in operator++ below) and
  the 2D dimension (in the user's loop).
*/
template <class TwoD_iter, class OneD_iter>
class flatten_iterator {
  typedef flatten_iterator self;
public:
  typedef std::forward_iterator_tag iterator_category;
  typedef typename std::iterator_traits<OneD_iter>::reference reference;
  typedef typename std::iterator_traits<OneD_iter>::value_type value_type;
  typedef typename std::iterator_traits<OneD_iter>::difference_type
    difference_type;
  typedef typename std::iterator_traits<OneD_iter>::pointer pointer;

  // Constructor for a singular iterator, not to be used until assigned
  flatten_iterator() { }

  // Constructor for an end() iterator; for such an iterator the
  // curr1D and end1D members are undefined and thus cannot be passed
  // in as arguments
  flatten_iterator(TwoD_iter start2D, TwoD_iter end2D_i)
    : curr2D(start2D), end2D(end2D_i)
    { }
       
  // Constructor for a non-end() iterator; it is illegal to use this
  // constructor for an end() flatten_iterator
  flatten_iterator(TwoD_iter curr2D_i, TwoD_iter end2D_i, OneD_iter curr1D_i, OneD_iter end1D_i)
    : curr2D(curr2D_i), end2D(end2D_i), curr1D(curr1D_i), end1D(end1D_i)
    {
      //assert(curr2D != end2D);

      // Skip leading empty containers.  This code should only get
      // executed for a begin() iterator, since curr2D and curr1D of
      // all other non-end() iterators are guaranteed to point to a
      // valid item already.
      while (curr1D == end1D && ++curr2D != end2D) {
	curr1D = (*curr2D).begin();
	end1D = (*curr2D).end();
      }
    }

  self& operator++() {
    //assert(curr2D != end2D);
    ++curr1D;
    // Skip to next valid element, if any
    while (curr1D == end1D && ++curr2D != end2D) {
      curr1D = (*curr2D).begin();
      end1D = (*curr2D).end();
    }
    return *this;
  }
  self operator++(int) {
    self tmp = *this;
    ++(*this);
    return tmp;
  }
  reference operator*() {
    //assert(curr2D != end2D);
    return *curr1D; 
  }

  // All end() iterators are identical and do not allow
  // reading the curr1D, so must check for that case first
  bool operator!=(const self& x) const {
    return (curr2D == end2D ?
	    x.curr2D != x.end2D :
	    curr2D != x.curr2D || curr1D != x.curr1D);
  }
  bool operator==(const self& x) const {
    return (curr2D == end2D ?
	    x.curr2D == x.end2D :
	    curr2D == x.curr2D && curr1D == x.curr1D);
  }
protected:
  TwoD_iter curr2D;
  TwoD_iter end2D;

  OneD_iter curr1D;
  OneD_iter end1D;
};



// Functions for constructing begin() and end() flatten_iterators for
// any given container of containers.  Currently the argument types
// are insufficent to specify the template arguments, since they do
// not include an item of type OneDContainer (nor could they, since no
// item of that type is guaranteed to be available for an empty
// container-container.)  Thus these functions must be called using
// the full template notation, as in:
//
// STL: flatten_begin< vector< vector<double> >, vector<double> >(A)  
// MTL: flatten_begin< matrix<double>::type, matrix<double>::type::OneD >(B)
//
// This restriction could (and probably should) be removed for STL
// containers by using the type TwoDContainer::value_type instead of
// OneDContainer, which would ensure that the template parameters are
// fully determined by the argument types.  Unfortunately, this would
// break the code for MTL arrays, since MTL's arrays are somewhere in
// between a container of containers and a container of elements ---
// the iterator is over containers but the value_type is the element
// type.  Sigh.
template <class TwoDContainer, class OneDContainer>
flatten_iterator<typename TwoDContainer::const_iterator,
                 typename OneDContainer::const_iterator>
flatten_begin(const TwoDContainer& A)
{
  typedef flatten_iterator<typename TwoDContainer::const_iterator,
                           typename OneDContainer::const_iterator> Fiter;
  return (A.begin() == A.end() ?
	  Fiter(A.begin(), A.end()) :
	  Fiter(A.begin(), A.end(), (*A.begin()).begin(), (*A.begin()).end() ));
}

template <class TwoDContainer, class OneDContainer>
flatten_iterator<typename TwoDContainer::const_iterator,
                 typename OneDContainer::const_iterator>
flatten_end(const TwoDContainer& A)
{
  typedef flatten_iterator<typename TwoDContainer::const_iterator,
                           typename OneDContainer::const_iterator> Fiter;
  return Fiter(A.end(),A.end());
}

template <class TwoDContainer, class OneDContainer>
flatten_iterator<typename TwoDContainer::iterator,
                 typename OneDContainer::iterator>
flatten_begin(TwoDContainer& A)
{
  typedef flatten_iterator<typename TwoDContainer::iterator,
                           typename OneDContainer::iterator> Fiter;
  return (A.begin() == A.end() ?
	  Fiter(A.begin(), A.end()) :
	  Fiter(A.begin(), A.end(), (*A.begin()).begin(), (*A.begin()).end() ));
}

template <class TwoDContainer, class OneDContainer>
flatten_iterator<typename TwoDContainer::iterator,
                 typename OneDContainer::iterator>
flatten_end(TwoDContainer& A)
{
  typedef flatten_iterator<typename TwoDContainer::iterator,
                           typename OneDContainer::iterator> Fiter;
  return Fiter(A.end(),A.end());
}


// Specialized version of the above begin() and end() functions that works 
// only for containers with a OneD member typedef which declares the type
// of the contained containers, namely MTL matrices.  With such an
// assumption, the template parameters may be deduced correctly; however,
// these will not work for e.g. an STL vector< vector<double> >.
//
template <class Matrix>
flatten_iterator<typename Matrix::const_iterator,
		   typename Matrix::OneD::const_iterator>
const_flatten_begin(const Matrix& A)
{
  typedef flatten_iterator<typename Matrix::const_iterator,
			     typename Matrix::OneD::const_iterator> Fiter;
  return (A.begin() == A.end() ?
	  Fiter(A.begin(), A.end()) :
	  Fiter(A.begin(), A.end(), (*A.begin()).begin(), (*A.begin()).end() ));
}

template <class Matrix>
flatten_iterator<typename Matrix::const_iterator,
                 typename Matrix::OneD::const_iterator>
const_flatten_end(const Matrix& A)
{
  typedef flatten_iterator<typename Matrix::const_iterator,
                           typename Matrix::OneD::const_iterator> Fiter;
  return Fiter(A.end(),A.end());
}

template <class Matrix>
flatten_iterator<typename Matrix::iterator,
                 typename Matrix::OneD::iterator>
flatten_begin(Matrix& A)
{
  typedef flatten_iterator<typename Matrix::iterator,
                           typename Matrix::OneD::iterator> Fiter;
  return (A.begin() == A.end() ?
	  Fiter(A.begin(), A.end()) :
	  Fiter(A.begin(), A.end(), (*A.begin()).begin(), (*A.begin()).end() ));
}

template <class Matrix>
flatten_iterator<typename Matrix::iterator,
                 typename Matrix::OneD::iterator>
flatten_end(Matrix& A)
{
  typedef flatten_iterator<typename Matrix::iterator,
                           typename Matrix::OneD::iterator> Fiter;
  return Fiter(A.end(),A.end());
}

}

#endif /* MTL_FLATTEN_ITERATOR_H */
