/** \file
  Low-level classes for specifying and operating on axis-aligned
  rectangles.  These are e.g. useful as bounding boxes within other
  classes.  For speed, these do not use virtual functions, and to
  allow external control via e.g. parameter-setting mechanisms they
  leave their member data public.

  $Header: /u/nn/cvsroot/lissom/src/aarectangle.h,v 1.9 2003/09/30 00:02:19 jbednar Exp $
*/

#ifndef __AARECTANGLE_H__
#define __AARECTANGLE_H__

#include <iostream>
#include <algorithm>

#include "genericalgs.h"

/** Namespace for rectangles and other 2D shapes, including bounding regions. */
namespace Boundary {
  
using std::min;
using std::max;


/** Low-level class containing a pair of two-dimensional points in X and Y.
    (rl,cl) is the point corner closest to the origin.
*/
template<class T=int>
struct TwoOrderedPoints {
  /** The member data is public since the class is low-level. */
  T xl,yl,xh,yh;
  
  inline TwoOrderedPoints(const T& xlow=0, const T& ylow=0, 
			  const T& xhigh=0,const T& yhigh=0)
    : xl(xlow), yl(ylow), xh(xhigh), yh(yhigh)  {  }
};



/** Basic specification for an axis-aligned rectangle determined by
    (rl,cl), the corner closest to the origin, and (rh,ch), the corner
    furthest from it.

  Either corner may be inclusive or exclusive, depending on how this
  class is used; it is thus not a complete bounding box.  To make it
  fast enough to use widely, it does not use any virtual functions; if
  such generality is needed use a full BoundingBox instead.
*/
template<class T=int,class Boolean=bool,class Float=double>
struct AARectangle {
private:
  typedef AARectangle<T,Boolean,Float> self;
  
public:
  typedef TwoOrderedPoints<T> Points;
  typedef T value_type;
  
  /** Typical constructor: square box */
  inline AARectangle(const T& xlow=0, const T& ylow=0, 
		     const T& xhigh=0,const T& yhigh=0)
    : p(xlow,ylow,xhigh,yhigh)  {  }
  
  /** Returns true if the given coordinates are inside the box,
      treating both corners as inclusive. */
  inline Boolean inside(T x, T y) const
    {  return (x>=p.xl && y>=p.yl && x<=p.xh && y<=p.yh);  }
  
  /** Returns true if the given coordinates are inside the box,
      treating the lower corner as inclusive and the upper as
      exclusive. */
  inline Boolean upperexclusive_inside(T x, T y) const
    {  return (x>=p.xl && y>=p.yl && x<p.xh && y<p.yh);  }
  
  /** Expands this rectangle to be an axis-aligned rectangle
      covering all area covered by the itself or the given
      rectangle.  Note that this is not the same as the union
      of the two areas, since it may include area not covered
      by either box so that the result remains a rectangle.
  */
  
  inline self& operator+=( const self& o ) {
    if (empty()) { p=o.p; }
    else         { p.xl=min(p.xl,o.p.xl); p.yl=min(p.yl,o.p.yl); p.xh=max(p.xh,o.p.xh); p.yh=max(p.yh,o.p.yh); }
    return *this;
  }
  
  /** Multiply the size times the indicated amounts in each dimension. */  
  inline self& scale( T xscale, T yscale ) {
    p.xl=p.xl*xscale; p.yl=p.yl*yscale; p.xh=p.xh*xscale; p.yh=p.yh*yscale;
    return *this;
  }
  
  /** Rotate the box by the given angle (in radians) and take the smallest
      axis-aligned rectangle which encloses the result.
  */
  self& rotate( Float theta ) {
    const Float cosmt  = cos(-theta);
    const Float sinmt  = sin(-theta);
    
    const T xlc = T(p.xl*cosmt);  const T xls = T(p.xl*sinmt);
    const T xhc = T(p.xh*cosmt);  const T xhs = T(p.xh*sinmt);
    const T ylc = T(p.yl*cosmt);  const T yls = T(p.yl*sinmt);
    const T yhc = T(p.yh*cosmt);  const T yhs = T(p.yh*sinmt);
    
    p.xl = min(min(min(xlc-yls,xlc-yhs),xhc-yls),xhc-yhs);
    p.xh = max(max(max(xlc-yls,xlc-yhs),xhc-yls),xhc-yhs);

    p.yl = min(min(min(xls+ylc,xls+yhc),xhs+ylc),xhs+yhc);
    p.yh = max(max(max(xls+ylc,xls+yhc),xhs+ylc),xhs+yhc);
    
    return *this;
  }
  
  /** Translate the position by the indicated amounts. */
  inline self& translate( T xoffset, T yoffset ) {
    p.xl+=xoffset; p.yl+=yoffset; p.xh+=xoffset; p.yh+=yoffset;
    return *this;
  }
  
  /** Set the rectangle to the given dimensions. */
  inline self& set(T xlow, T ylow, T xhigh, T yhigh ) {
    p.xl=xlow; p.yl=ylow; p.xh=xhigh; p.yh=yhigh;
    return *this;
  }

  /** Set the rectangle to the given dimensions. */
  inline self& set (const Points& o) {  p=o;  }

  /* Returns a rectangle that is the intersection of the given box with this one */
  inline self intersection(const self& o) {
    return self(Generic::crop(o.p.xl,o.p.xh,p.xl),
		Generic::crop(o.p.yl,o.p.yh,p.yl),
		Generic::crop(o.p.xl,o.p.xh,p.xh),
		Generic::crop(o.p.yl,o.p.yh,p.yh));
  }
  
  /** The height of the rectangle, assuming inclusive bounds. */
  inline T height() const {  return (std::abs(p.yh-p.yl));  }

  /** The width of the rectangle, assuming inclusive bounds. */
  inline T width()  const {  return (std::abs(p.xh-p.xl));  }
  
  /** Discouraged: returns the points making up the corners of the box. */
  inline Points corners()  const {  return p;  }
  
  /** Regardless of whether the corners are exclusive or inclusive,
      enclosing a single point or line is always considered empty
      since zero area is enclosed. */
  inline Boolean empty() const {  return (height()==0 || width()==0); }

  /** Output a human-readable representation to the given stream */
  inline std::ostream& put(std::ostream &s) const
    {  return (s << "((" << p.xl << "," << p.yl << ") (" << p.xh << "," << p.yh << "))"); }

private:
  Points p;
};
  
  
  
/** General-purpose axis-aligned rectangle, identical
    to AARectangle<> except that it also allows boxes to be
    infinite. */
template<class T=int,class Boolean=bool,class Float=double>
class AAArbitraryRectangle {
private:
  typedef AAArbitraryRectangle<T,Boolean,Float> self;
  typedef AARectangle<T,Boolean,Float> PlainRect;
  
public:
  typedef TwoOrderedPoints<T> Points;
  typedef T value_type;

  /** Default constructor: infinite extent */
  inline AAArbitraryRectangle() : infinite(true) { }
  
  /** Typical constructor: square box */
  inline AAArbitraryRectangle(T xlow, T ylow, T xhigh, T yhigh )
    : infinite(false), rect(xlow,ylow,xhigh,yhigh) { }
  
  inline Boolean isinfinite() const {  return infinite;  }
  
  inline Boolean inside(T x, T y) const
    {  return (infinite || rect.inside(x,y));  }
  
  inline Boolean upperexclusive_inside(T x, T y) const
    {  return (infinite || rect.upperexclusive_inside(x,y));  }
  
  /** Make the bounding box infinite. */
  inline self& set() {  infinite=true; return *this;  }

  inline self& set(T xlow, T ylow, T xhigh, T yhigh)
    {  infinite=false; rect.set(xlow,ylow,xhigh,yhigh); return *this;  }
  
  /**
     Expand this box to enclose the given one.
     
     An infinite box already encloses all boxes, while an empty() box
     will enclose any other box exactly, regardless of the original
     location of the empty rect.
  */
  inline self& operator+=( const self& b ) {
    if (infinite || b.infinite) infinite=true;
    else rect+=b.rect;
    return *this;
  }
  
  /** Multiply the size times the indicated amounts in each dimension. */
  inline self& scale( T xscale, T yscale )
    {  if (!infinite) rect.scale(xscale,yscale);  return *this;  }
  
  /** Rotate the box by the given angle (in radians) and take the smallest
      axis-aligned rectangle which encloses the result. */
  inline self& rotate( Float theta ) 
    {  if (!infinite) rect.rotate(theta);  return *this;  }
  
  /** Translate the position by the indicated amounts. */
  inline self& translate( T xoffset, T yoffset ) 
    {  if (!infinite) rect.translate(xoffset,yoffset);  return *this;  }
  
  /** Output a human-readable representation to the given stream */
  inline std::ostream& put(std::ostream &s) const
    {  return (infinite? s << "(Infinite)" : rect.put(s)); }

  /* Returns a rectangle that is the intersection of the given box with this one */
  inline self& intersection(const self& o) {
    return (o.isinfinite()? *this : 
	    (isinfinite()?  o     :
	     self(rect.intersection(o.rect))));
  }
  
  /** The height of the rectangle, assuming inclusive bounds, or zero if infinite. */
  inline T height() const {  return (infinite? 0 : rect.height()); }

  /** The width of the rectangle, assuming inclusive bounds, or zero if infinite. */
  inline T width()  const {  return (infinite? 0 : rect.width());  }
  
  /** Discouraged: returns a plain axis-aligned rectangle, which is
      empty if this rectangle is infinite.  */
  inline PlainRect aarectangle()  const {  return (infinite? PlainRect() : rect);  }
  
  /** Discouraged: returns the points making up the corners of the box. */
  inline Points corners()  const {  return rect.corners();  }
  
  /** Regardless of whether the corners are exclusive or inclusive,
      enclosing a single point or line is always considered empty
      since zero area is enclosed. */
  inline Boolean empty() const
    {  return (!infinite && rect.empty()); }
  
private:
  Boolean infinite;
  PlainRect rect;
};


} /* namespace Boundary */ 
#endif /* __AARECTANGLE_H__ */
