/** \file
  AllocatingPointer, OwningPointer, and MemoryReference classes

  $Header: /u/nn/cvsroot/lissom/src/allocatingpointer.h,v 1.17 2001/04/01 07:05:54 jbednar Exp $
*/

#ifndef __ALLOCATING_POINTER_H__
#define __ALLOCATING_POINTER_H__

#include <functional>


/**
  Returns a copy of the given object.  It allocates only the nominal
  type T, and will thus slice objects which happen to be of a derived
  type rather than the nominal type.
*/
template<class T>
struct CopyAsNominalType : public std::unary_function<T,T*>
{  T* operator() (const T& value) {  return new T(value);  }  };



/**
  Returns a copy of the given object.  Calls the nominal types'
  clone() operation, which if virtual and implemented in all derived
  types will construct a clone object of the actual type instead of
  the nominal type.  This makes it possible to work with unknown
  derived classes.  Of course, to be able to use this class, type T
  must support such a clone() operation.
*/
template<class T>
struct CopyUsingClone : public std::unary_function<T,T*>
{  T* operator() (const T& value) {  return value.clone();  }  };



/** Simple pointer-like class to hide whether the object pointed to is
  an external object with its own lifetime, or an object allocated and
  managed by this class.  The differences from regular pointer
  semantics are intended to be the following:

   1) On construction a value may be allocated (if called with a scalar).
   2) On destruction any allocated value will be deallocated.
   3) When copied, the copy will only point to the same memory
      location if that location has its own lifetime, i.e. was
      an external object; otherwise a new object is allocated
      and set to have the same value as the original.

  A few operations not supported by pointers are also provided,
  and not all operations supported by pointers have necessarily
  been implemented.  (E.g. use (*ap). instead of ap-> for now.)

  Using the default Cloner template argument, this class cannot be
  used with an object of an unknown type derived from the nominal
  templated type, since it needs to be able to create a copy of the
  object.  If the objects offer a virtual clone() procedure, you can 
  supply a different Cloner (e.g. CopyUsingClone) which supports
  copying of the actual types; this will allow unknown derived types
  to be supplied without fear of slicing.
*/
template<class T, class Cloner=CopyAsNominalType<T>, bool default_own=false >
class OwningPointer {
  typedef OwningPointer<T,Cloner> self;

public:
  /** Constructor for an internally-allocated variable */
  OwningPointer(const T& value)
    : vp(allocate(value)), allocated(vp) { }
  
  /** Constructor for a pointer to an external value */
  OwningPointer(T* value_pointer, bool owns_object=default_own)
    : vp(value_pointer), allocated(owns_object? value_pointer : 0) { }

  /** Copy constructor; copies other's pointer only if it
      refers to an externally-managed object */
  OwningPointer(const self& other)
    : vp       (other.externalpointer()? other.vp : allocate(*(other.vp))),
      allocated(other.externalpointer()? 0        : vp    ) { }

  ~OwningPointer()
    {  delete allocated;  }
  
  /** Override the underlying value with a new one. */
  void operator= (const T& new_value) { *vp=new_value; }

  /** Copy assignment */
  self& operator= (const self& o) {
    if (this == &o) return *this;    
    if (allocated) {
      delete allocated; allocated=0;
    }

    if (o.externalpointer()) vp=o.vp;
    else allocated=vp=allocate(*(o.vp));

    return *this;
  }

  /** Conversion to Boolean; acts like underlying pointer. */
  inline      operator bool()  const {  return bool(vp); }

  /** Returns the current underlying value. */
  inline const T& operator*()  const {  return *vp; }

  /** Const pointer dereference. */
  inline const T* operator->() const {  return vp;  }

  /** Pointer dereference. */
  inline       T* operator->()       {  return vp;  }

  /** Returns a pointer to the underlying value; use with caution. */
  inline T* valueptr() const {  return  vp;  }

  /** Returns a modifiable pointer to the underlying value;
      use with extreme caution.  
  */
  inline T*& valueptrref()   {  return  vp;  }

  /* Automatic conversion to type T (disabled) */
  //operator T() const { return *vp; }

  /** \name    Operations that may be overridden by derived classes.
      
    Could support clone() here except that the Cray compiler
    doesn't allow the return type to vary even trivially, and
    a type ValueGenerator was chosen for the standard return type.
  */
  //@{
  /** If our pointer is external, overwrite it with the given one */
  void relink(T* ptr) {  if (externalpointer()) vp = ptr;  }
  //@}
  
private:
  /** Pointer to underlying object, whether allocated or external. */
  T* vp;

  /** If non-zero, points to an allocated object; this should be
      deleted when we are destroyed or when a new object is allocated */
  T* allocated;

  /** Returns true if vp is an external pointer, i.e. does not
      point to an allocated value  */
  inline bool externalpointer() const {  return (allocated!=vp);  }

protected:
  /** Allocates and returns a pointer to a new item set to the given value */
  inline T* allocate(const T& value) const
    {  return Cloner()(value);  }
};



/**
  Version of OwningPointer that adds a default argument to the value
  constructor, thus making it parameter-less, which allows it to be
  used in e.g. arrays.
*/
template<class T, class Cloner=CopyAsNominalType<T>, bool default_own=false >
class AllocatingPointer : public OwningPointer<T,Cloner,default_own> {
public:
  /** Constructor for an internally-allocated variable */
  AllocatingPointer(const T& value=T())
    : OwningPointer<T,Cloner,default_own>(allocate(value),true) { }
  
  /** Constructor for a pointer to an external value */
  AllocatingPointer(T* value_pointer, bool owns_object=default_own)
    : OwningPointer<T,Cloner,default_own>(value_pointer,owns_object) { }

  /** Override the underlying value with a new one. */
  void operator= (const T& new_value) { *valueptr()=new_value; }
};



/** Keeps track of at most one pointer, deleting it when this object is destroyed.

  Essentially, this class peforms simple single-use garbage
  collection.  In general, an AllocatingPointer is more useful, but
  this class is required in some cases.  Specifically, this class can
  handle objects of an unknown derived type, even if the type does not
  support a clone() operation (e.g. the standard numeric or string
  types.)  It does so by sacrificing the ability to copy the
  pointed-to object; the copy simply has no pointer at all.
*/
template<class T>
class MemoryReference {
  typedef MemoryReference<T> self;
  
public:
  /** Constructor; accepts 0 or a pointer to an object for which we are responsible */
  MemoryReference(T* pointer=0)
    : allocated(pointer) { }

  /** Copy constructor; does not copy pointer. */
  MemoryReference(const MemoryReference&) : allocated(0) { }

  ~MemoryReference()  {  delete allocated;  }
  
  /** Copy assignment; does not copy pointer. */
  self& operator= (const MemoryReference&) { }

private:
  /** If non-zero, points to an allocated object; this should be
      deleted when we are destroyed or when a new object is allocated */
  T* allocated;
};


#endif /* __ALLOCATING_POINTER_H__ */
