/** \file
  StateSaver and LISSOMBinaryStateSaver class declarations.  The use of these classes
  is deprecated.
  
  $Header: /u/nn/cvsroot/lissom/src/binarysave.h,v 1.14 2003/03/15 01:47:31 jbednar Exp $
*/

#ifndef __BINARYSAVE_H_
#define __BINARYSAVE_H_

#include <cstdio>
#include <vector>

#include "tristate.h"
#include "ind_types.h"
#include "kernel.h"


/******************************************************************************/
/* StateSaver class                                                           */
/******************************************************************************/


/** Abstract base class for writing a state to a file when that state
    is directly accessible to this class (e.g.\ if the class is a
    friend of the classes to be saved, or if it is in global
    variables.)

    This class exists only for backwards compatibility with old LISSOM
    code. Avoid using it if possible, instead using something with a
    thin interface based e.g. on a series of calls for different
    Matrix arguments.

    All but write should be able to be const member functions, but
    that may take some work.
*/
class StateSaver {
public:
  virtual ~StateSaver() { }

  /** Write state to the given file */
  virtual int  read(       FILE *infile  )=0;

  /** Read state from the given file */
  virtual int  write(      FILE *outfile )=0;

  /** Write a readable representation of the given file to stdout */
  virtual void display_as_text( FILE *infile  )=0;
};



/******************************************************************************/
/* Defines                                                                    */
/******************************************************************************/

/**
   Number of neurons to be handled in one buffer for saving and loading.
   The larger the number, the more efficient, but the more memory is 
   required.  Must be smaller than N, and N must be an even multiple of it.
*/
#define NEURONS_PER_BUFFER 1
#define parts_per_row ((lm.N)/NEURONS_PER_BUFFER)

/** Units for token lengths.
    
  Each token must be an even multiple of this number.
  Token headers must also be this length, as must
  the size of LFBufferDataType.
*/
#define LF_TOKEN_GRAIN_SIZE        4
/** The size of this should match LF_TOKEN_GRAIN_SIZE. */
#define LFBufferDataType_to_use      i32
#define LFBufferDataType_swap_endian i32_swap_endian




/******************************************************************************/
/* LISSOMBinaryStateSaver class                                               */
/******************************************************************************/

/** StateSaver that can save the state of a LISSOM network to a binary
    file.  This entire class hierarchy is deprecated; see the
    StateSaver documentation.
*/
class LISSOMBinaryStateSaver : public StateSaver {
public:
  virtual ~LISSOMBinaryStateSaver() { }
  
  LISSOMBinaryStateSaver(BasicLissomMap& m) : 
    tokensWritten(0), tokensRead(0), status(0), errorsEncountered (0),
    bigendian(endianness()), lm(m) {  }
  
  /** Call this routine before calling any others */
  static void init_hook( void ) {  binarysave_init_hook();  }

  virtual int  read(       FILE *infile  ) {  return binaryWeightsFileRead(  infile  );  }
  virtual int  write(      FILE *outfile ) {  return binaryWeightsFileWrite( outfile );  }
  virtual void display_as_text( FILE *infile  ) {    printTokens(            infile  );  }

  /** Type of each item in the buffer */
  typedef LFBufferDataType_to_use LFBufferDataType;

private:
  typedef BasicLissomMap::a_weight a_weight;
  typedef BasicLissomMap::l_weight l_weight;
  
  /*  Typedefs */
  
  /** Type of each token; allows access to parts of 64-bit word */
  struct LFTokenType {
    i32 right;
    i32 left;
  };

  /** Buffer for constructing binary file to write */
  struct LFBuffer {
    LFBuffer(const int initial_size=0) : data(initial_size) { }
    int start;
    int end;
    inline int  length() const { return (end-start + 1); }
    inline void ensure_space_available(const unsigned additional_length) {
      const unsigned required_length = length()+additional_length;
      if (required_length >= data.size())
	data.resize(required_length);
    }
    std::vector<LFBufferDataType> data;
  };

  /*   Parameters */
  static Tristate save_afferent_weights_only;
  static Tristate save_all_lateral_weights;
  static Tristate load_afferent_weights_only;
  static Tristate load_lateral_inhibitory_weights;

  /*   Variables */
  int tokensWritten; /**< Total number of tokens output so far. */
  int tokensRead; /**< Total number of tokens input  so far. */
  
  LFBuffer  binaryBuffer; 
  int       status;
  /** Number of errors encountered in this load or save operation. */
  int       errorsEncountered; 
  
  /** Initially undefined; define before use. */
  int bigendian;


  /*   Prototypes for public functions */
  static void binarysave_init_hook( void );
  
  int  binaryWeightsFileWrite( FILE *file );
  int  binaryWeightsFileRead( FILE *file );
  void printTokens( FILE *file );
    
  /*   Prototypes for private functions */
  LFTokenType makeToken(int name, int length, int infoa, int infob );
  int  tokenName(   LFTokenType token );
  int  tokenLength( LFTokenType token );
  int  tokenInfoa(  LFTokenType token );
  int  tokenInfob(  LFTokenType token );
  void printToken( LFTokenType token, int offset );
  const char *tokenNameString(LFTokenType token);
	 
  int writeToken(FILE *file, LFTokenType token );
  LFTokenType peekToken(FILE *file );
  LFTokenType readToken(FILE *file );
  LFTokenType readTokenExpecting(int tokenNameExpected, FILE *file);
  int writeItem(FILE *file, void *itemptr);
  LFBufferDataType readItem(FILE* file);
  int writeFloat(FILE *file, f32 aFloat );
  f32 readFloat(FILE *file);
  int writeInteger(FILE *file, int anInteger );
  int readInteger(FILE *file);
  void printTokensInBuffer( LFBuffer *buffer );
  
  int binaryWeightsFileReadInArbitraryOrder(FILE *file, LFBuffer *buffer);
  int binaryWeightsFileReadInFixedOrder(FILE *file, LFBuffer *buffer);
  
  int binaryWeightsBufferFill( LFBuffer *buffer, int current_map_row, 
				int current_local_row, int part_of_row );
  
  int binaryWeightsBufferProcess( LFBuffer *buffer, int verify);
  int binaryWeightsBufferWriteLatWeights( LFBuffer *buffer, int token_name,
					  int i, int j, int radius, int array_width,
					  l_weight *weights);
  int binaryWeightsBufferReadLatWeights(  LFBuffer *buffer, LFTokenType token,
					  int i, int j, int radius, int array_width,
					  l_weight *weights, int verify, const char *description);
  
  void LFBufferGet(    LFBuffer *buffer, int pe );
  void LFBufferPut(    LFBuffer *buffer, int pe );
  void LFBufferClear(  LFBuffer *buffer );
  int  LFBufferLength( LFBuffer *buffer );
  int  LFBufferWrite(  LFBuffer *buffer, FILE *file );
  int  LFBufferRead(   LFBuffer *buffer, FILE *file, int length);
  int  LFBufferWriteToken(LFBuffer *buffer, LFTokenType token );
  LFTokenType LFBufferReadToken(LFBuffer *buffer );
  int  LFBufferWriteItem(LFBuffer *buffer, void *itemptr);
  LFBufferDataType LFBufferReadItem(LFBuffer *buffer);
  int  LFBufferWriteFloat(LFBuffer *buffer, f32 aFloat );
  f32  LFBufferReadFloat(LFBuffer *buffer);
  int  LFBufferWriteInteger(LFBuffer *buffer, int anInteger );
  int  LFBufferReadInteger( LFBuffer *buffer);
  int  compareParameter(int weight_file_parameter, int param_file_parameter, const char *description);
  int  compareOrSetInt(int *destination, int value, int verify, const char *description);

#define compareOrSet_proto(type) \
type compareOrSet_ ## type(type *destination, type value, int verify, const char *description)
  
  compareOrSet_proto(f64);
  compareOrSet_proto(a_weight);
  compareOrSet_proto(l_weight);

  BasicLissomMap& lm;
};



#endif
