/** \file
  Contains Bounded namespace, holding bounded numeric classes.

  $Header: /u/nn/cvsroot/lissom/src/boundednumber.h,v 1.6 2001/04/07 02:19:17 jbednar Exp $
*/

#ifndef __BOUNDEDNUMBER_H__
#define __BOUNDEDNUMBER_H__

#include <cmath>
#include <cassert>
#include <algorithm>

#include "genericalgs.h"
#include "bounded.h"


namespace Bounded {


  
/** Function object wrapper (a Boundary) for Generic::crop */
template<class T>
struct Crop {  T operator() (T l, T u, T x) { return Generic::crop(l,u,x); }  };

 
/** Function object wrapper (a Boundary) for Generic::wrap */
template<class T>
struct Wrap {  T operator() (T l, T u, T x) { return Generic::wrap(l,u,x); }  };




/* Forward declarations to allow Float and Integer to convert to
   and from one another */
template<class, LargeInt, LargeInt, class> class Integer;
template<class, LargeInt, LargeInt, class> class Float;


 
/** Returns the maximum value representable in the given unsigned
    integer type. Should be unnecessary once all compilers support
    numeric_limits.
*/
#define MAX_UNSIGNED(T) ((sizeof(T)<<8)-1)

 
  
/** Bounded number with integral type.

  The template can be instantiated for any integral type smaller than
  or equal to type LargeInt.

  An instance of this class or the Float class stores a single number,
  silently enforcing upper and lower bounds for construction,
  assignment, and copying from other types.  Unlike the implicit
  aliasing of e.g. a char, the boundaries may be smaller than the
  maximum representable in that type (although for integers the
  default range is the full range available) and values outside the
  range can either be truncated (cropped) at the boundaries, or
  aliased (wrapped) into the valid range.
*/
template<class T, LargeInt MaxVal=LargeInt(MAX_UNSIGNED(T)),
         LargeInt MinVal=0, class Boundary=Crop<LargeInt> >
class Integer {
private:
  typedef T value_type; /**< Underlying datatype; could be made public */
  value_type n;         /**< Actual numeric data value */
  typedef Integer self;

  inline const LargeInt crop(const LargeInt x)
    {  return Boundary()(Min,Max,x); }

  /** Sets to cropped value */
  inline value_type set(const LargeInt x)
    {  return n=value_type(Boundary()(Min,Max,x));  }

  //  template<class X> friend X operator+(const X &A, const X &B);
  
public:
  static const value_type Min=MinVal; /**< Minimum legal value */
  static const value_type Max=MaxVal; /**< Maximum legal value */
  
  Integer()               : n(set(0))   { } /**< Default constructor */
  Integer(value_type val) : n(set(val)) { } /**< Constructs using specified value. */
  Integer(const self& other)
    : n(other.n)                        { } /**< Copy constructor. */
  
  /** Construct from a Bounded::Float */
  template<class OT, LargeInt OMaxVal, LargeInt OMinVal, class OBoundary>
  Integer(const Float<OT,OMaxVal,OMinVal,OBoundary>& other)
    : n(set(value_type(Max*other.mag())))  { }

  /** Set to a numeric value; assumes LargeInt is an encompassing type */
  void operator= (const LargeInt& val) {  set(val);  assert(n<=Max); }

  /** Multiplicative scaling */
  self& operator*=(const LargeFloat m) {  set(LargeInt(n*m));  return *this; }
  
  /** Integer addition */
  self operator+(const self &B) const {
    self tmp;
    tmp.set(n+B.n);
    return tmp;
  }

  
  /** Returns the current numeric value as a fraction of the maximum */
  Magnitude mag() const { assert(n<=Max); return LargeFloat(n)/LargeFloat(Max);  }

  /** Returns the current numeric value in the native units */
  value_type raw() const { assert(n<=Max); return n;  }
};


 
/** Bounded number with floating-point type.  See the Integer class
    for more details.

  The template can be instantiated for any floating-point type smaller
  than or equal to type LargeFloat.

  Unfortunately, because the C++ standard explicitly disallows
  floating-point template literal arguments (without giving any
  rationale!), the MaxVal and MinVal must always be integers
  representable in type LargeInt, even though the underlying type is
  floating-point.  Luckily bounds are usually integers anyway.
*/
template<class T, LargeInt MaxVal=1, LargeInt MinVal=0, class Boundary=Crop<LargeFloat> >
class Float {
private:
  typedef T value_type; /**< Underlying datatype; could be made public */
  value_type n;         /**< Actual numeric data value */

  inline const LargeFloat crop(const LargeFloat x)
    {  return Boundary()(Min,Max,x); }
  
public:
  static const value_type Min; /**< Minimum legal value */
  static const value_type Max; /**< Maximum legal value */

  Float()               : n(crop(0))   { } /**< Default constructor */
  Float(value_type val) : n(crop(val)) { } /**< Constructs using specified value. */
  Float(const Float<T,MaxVal,MinVal,Boundary>& other)
    : n(other.n)                       { } /**< Copy constructor. */
  
  /** Construct from a Bounded::Integer */
  template<class OT, LargeInt OMaxVal, LargeInt OMinVal, class OBoundary>
  Float(const Integer<OT,OMaxVal,OMinVal, OBoundary>& other)
    : n(crop(value_type(Max*(other.mag()))))  { }
    
  /** Set to a numeric value; assumes LargeFloat is an encompassing type */
  void operator= (const LargeFloat& val) {  n=value_type(crop(val)); assert(n>=Min&&n<=Max); }
  
  /** Returns the current numeric value as a fraction of the maximum */
  Magnitude mag() const { assert(n>=Min&&n<=Max); return LargeFloat(n)/LargeFloat(Max);  }

  /** Returns the current numeric value in the native units */
  value_type raw() const { assert(n>=Min&&n<=Max); return n;  }
};


template<class T, LargeInt MaxVal, LargeInt MinVal, class Boundary>
const T Float<T,MaxVal,MinVal,Boundary>::Min=MinVal;
template<class T, LargeInt MaxVal, LargeInt MinVal, class Boundary>
const T Float<T,MaxVal,MinVal,Boundary>::Max=MaxVal;

 
} /* namespace Bounded */

#endif /* __BOUNDEDNUMBER_H__ */
