/** \file
  BoundedString class.

  $Header: /u/nn/cvsroot/lissom/src/boundedstring.h,v 1.2 2001/04/01 07:05:54 jbednar Exp $
*/

#ifndef __BOUNDEDSTRING_H__
#define __BOUNDEDSTRING_H__

#include <string>
using std::string;

namespace Bounded {



/**
  Simple wrapper for a C array of characters that acts like a regular
  string, but has a fixed maximum size (silently cropping) and a
  c_str() pointer that can be passed around safely.

  Automatic conversion operator to a string is supported, which allows
  an object of this class to be passed directly into most functions
  expecting a string.  However, some overload ambiguities may remain.
  If so, explicit conversion to string can be used where needed,
  either at the usage point or by implementing the appropriate
  function here using such a conversion.
*/
template <size_t N, class T=char>
class BoundedString {
public:
  BoundedString() {}
  
  BoundedString(const basic_string<T>& o) {  operator=(o);  }
  
  /** Assignment copies all characters up to the bound, then adds terminator */
  void operator= (const basic_string<T>& o) 
    {  data[o.copy(data,N)]=0;  }
  
  bool operator== (const char* o) const
    {  return (string(*this)==string(o));  }
  
  /** Automatic conversion to type string */
  operator basic_string<T>() const
    {  return basic_string<T>(data);  }
  
  /** Returns a pointer to a C array of characters.  Unlike regular string,
      this pointer will remain valid for the entire lifetime of this object. */
  const T* c_str() const { return &(data[0]);  }
  
private:
  /** Character array, with an extra element for the terminator character */
  T data[N+1]; 
};



/** Stream output for a BoundedString */
template <size_t N, class T>
std::ostream& operator<<(std::ostream &s, const BoundedString<N,T>& b)
{  return s << string(b);  }



} /* namespace Bounded */

#endif /* __BOUNDEDSTRING_H__ */
