/** \file
  BoundingBox class hierarchy: Classes that keep track of a region in
  a two-dimensional space, computing whether a point is within that
  region and supporting various affine transformations of that region.
  Useful for e.g. avoiding computations for points outside the region.
  Rectangles are implemented using the lower-level Rectangle classes;
  see rectangle.h.

  $Header: /u/nn/cvsroot/lissom/src/boundingbox.h,v 1.20 2003/12/18 19:21:49 jbednar Exp $
*/

#ifndef __BOUNDINGBOX_H__
#define __BOUNDINGBOX_H__

#include <iostream>
#include "aarectangle.h"
#include "genericalgs.h"


namespace Boundary {

  
/**  Abstract base class specifying interface suitable for any bounding box.

  Type Float is used for rotation, which requires floating point
  even if T is an integer.
*/
template<class T=double, class Boolean=bool, class Float=double>
class BoundingBox {
  /** Base to use in interfaces */
  typedef BoundingBox<T,Boolean,Float> self;
  typedef BoundingBox<T,Boolean,Float> base;

public:
  typedef AARectangle<T,Boolean,Float> Rectangle;
  typedef T value_type;
  
  /** Whether the given point is enclosed by the box, inclusively */
  virtual Boolean inside(T x, T y) const =0;

  /** Scale the size of the box in X and Y dimensions */
  virtual base& scale(T xscale, T yscale) =0;

  /** Rotate the box around the origin */
  virtual base& rotate(Float angle) =0;

  /** Translate the origin of the box elsewhere */
  virtual base& translate(T xoffset, T yoffset) =0;

  /** Returns an axis-aligned rectangle fully enclosing this box, or
      an empty rectangle if no such finite rectangle exists. */
  virtual Rectangle aarectangle() const =0;
  
  /** Constructs an identical copy of self and returns it. */
  virtual base* clone() const=0;

  /** Virtual interface to the operator<<, since << cannot be virtual.
    This implementation is just a stub, and should be overriden in
    derived classes.  Outputs a human-readable representation to the
    given stream.
  */
  virtual std::ostream& put(std::ostream &s) const {  return s;  };
};



/** Stream output for a BoundingBox */
template<class T,class Boolean,class Float>
std::ostream& operator<<(std::ostream &s, const BoundingBox<T,Boolean,Float>& b) {  return b.put(s);  }



/** General-purpose axis-aligned rectangular bounding box.

  Basically just a wrapper around the given BoxType using the virtual
  function interface of BoundingBox<>, which allows this class to be
  compatible with other bounding box classes.
*/
template<class T=int,class Boolean=bool,class Float=double,
         class BoxType=AAArbitraryRectangle<T,Boolean,Float> >
class AARBoundingBox : public BoundingBox<T,Boolean,Float> {
private:
  typedef AARBoundingBox<T,Boolean,Float,BoxType> self;
  typedef BoundingBox<T,Boolean,Float> base;

public:
  /** Default constructor: infinite extent */
  AARBoundingBox() { }

  /** Typical constructor: square box */
  AARBoundingBox(T xlow, T ylow, T xhigh, T yhigh )
    : box(xlow,ylow,xhigh,yhigh), upperexclusive(false) { }

  /** Constructor taking a center and a radius for a square box
      with odd dimensions */
  AARBoundingBox(T centerx, T centery, T radius )
    : box(centerx-radius,centery-radius,centerx+radius,centery+radius),
      upperexclusive(false) { }
  
  virtual Boolean inside(T x, T y) const {
    return (upperexclusive?
	    box.upperexclusive_inside(x,y) :
	    box.inside(x,y));
  }

  /** Make the bounding box infinite. */
  virtual self& set() {  box.set(); return *this;  }

  /** Set the bounding box to the given dimensions. */
  virtual self& set(T xlow, T ylow, T xhigh, T yhigh) 
    {  box=BoxType(xlow,ylow,xhigh,yhigh); return *this;  }

  /** Expand this box to enclose the given one. */
  virtual self& operator+=( const self& b )
    {  box += b.box;  return *this;  }
  
  virtual base& scale( T xscale, T yscale )
    {  box.scale(xscale,yscale);  return *this;  }
  
  /** Rotate the box by the given angle (in radians) and take the smallest
      axis-aligned rectangle which encloses the result. */
  virtual base& rotate( Float theta ) {  box.rotate(theta);  return *this;  }

  virtual base& translate( T xoffset, T yoffset )
    {  box.translate(xoffset,yoffset);  return *this;  }

  virtual typename base::Rectangle aarectangle() const
    {  return (box.isinfinite()? typename base::Rectangle() : box.aarectangle());  }
  
  virtual base* clone() const {  return new self(*this); };

  virtual std::ostream& put(std::ostream &s) const
    {  return box.put(s); }

  /** Enclosing a single point or line is considered empty since zero
      area is enclosed. */
  virtual Boolean empty() const
    {  return box.empty(); }

  /** Make this box use upperexclusive_inside() instead of inside() for
      its own inside(); should probably be moved to a different class */
  virtual void make_upperexclusive()
    {  upperexclusive=True;  }
  
protected:  
  BoxType box;
  bool upperexclusive;
};



/** General-purpose ellipsoidal bounding box.

  For simplicity, this is just a minor variation on AARBoundingBox,
  but if speed is important it can be made a fully-separate class
  instead so that it can cache the values of the parameters needed
  by inside().

  To allow an arbitrary radius to be used even for an integer-bounded
  box, a "trim" value can be passed in; this will be added to the
  integer radius to get the floating-point radius.  This somewhat-
  unorthodox implementation was chosen not because it's better than
  just using a floating-point radius in the first place, but because
  it made it easier to interface with older code structured in the
  same way.
*/
template<class T=int,class Boolean=bool,class Float=double,
         class BoxType=AAArbitraryRectangle<T,Boolean,Float> >
class BoundingEllipse : public AARBoundingBox<T,Boolean,Float,BoxType> {
  typedef BoundingBox<T,Boolean,Float> base;
  typedef BoundingEllipse<T,Boolean,Float,BoxType> self;
  typedef AARBoundingBox<T,Boolean,Float> aarbase;
  
public:
  /** Default constructor: infinite extent */
  BoundingEllipse() : aarbase(), trimx(0), trimy(0) { }
  
  /** Constructor for a circular area */
  BoundingEllipse(T centerx, T centery, T radius, Float trim=0 )
    : aarbase(centerx-radius,centery-radius,centerx+radius,centery+radius),
    trimx(trim), trimy(trim) { }

  /** Constructor for an ellipsoidal area */
  BoundingEllipse(T centerx, T centery, T radiusx, T radiusy, Float trimx_i=0, Float trimy_i=0)
    : aarbase(centerx-radiusx,centery-radiusy,centerx+radiusx,centery+radiusy),
    trimx(trimx_i), trimy(trimy_i) { }
  
  /** Returns true if the given coordinates are inside the ellipse
      bounded by the corresponding rectangle. */
  virtual Boolean inside(T x, T y) const {
    if (box.isinfinite()) return true;
    
    /* These could be cached */
    const Float	radiusx        	= (box.width())/2.0+trimx;
    const Float	radiusy        	= (box.height())/2.0+trimy;
    const Float	div_radiusx_sq 	= 1/(radiusx*radiusx);
    const Float	div_radiusy_sq 	= 1/(radiusy*radiusy);
    const T centerx             = box.corners().xl+T(radiusx);
    const T centery             = box.corners().yl+T(radiusy);
    
    const T xdistance           = x-centerx;
    const T ydistance           = y-centery;
    const T xdistance_sq        = xdistance*xdistance;
    const T ydistance_sq        = ydistance*ydistance;

    return (xdistance_sq*div_radiusx_sq + ydistance_sq*div_radiusy_sq < 1);
  }


  /** Rotate the box by the given angle (in radians). Not implemented. */
  virtual base& rotate( Float theta ) {
    (void)theta;
    /* This routine should be straightforward to implement, since it would
       just need to rotate each corner about the center.   However, it would
       be tedious to do, and the inside() routine would need to be modified
       to match.  So it is left undone since this routine is not
       currently needed.
    */
    abort();
    return *this;
  }

  virtual base* clone() const {  return new self(*this); }

private:
  Float trimx,trimy;
};


/** BoundingBox consisting of the intersection between two
    already-defined bounding boxes.  */
template<class T=int,class Boolean=bool,class Float=double,
	 class BBA=BoundingEllipse<T,Boolean,Float>,
	 class BBB=AARBoundingBox<T,Boolean,Float> >
class Intersection : public BoundingBox<T,Boolean,Float> {
  typedef BoundingBox<T,Boolean,Float> base;
  typedef Intersection<T,Boolean,Float,BBA,BBB> self;
  typedef AARBoundingBox<T,Boolean,Float> aarbase;
  
public:
  /** Default constructor; infinite extent if no arguments supplied */
  Intersection(const BBA& bb_a=BBA(), const BBB& bb_b=BBB())
    : a(bb_a), b(bb_b) { }
  
  /** Returns true if the given coordinates are inside the ellipse
      bounded by the corresponding rectangle. */
  virtual Boolean inside(T x, T y) const 
    {  return (a.inside(x,y) && b.inside(x,y));  }

  /** Scale the size of the box in X and Y dimensions */
  virtual base& scale(T xscale, T yscale)
    {  a.scale(xscale,yscale); b.scale(xscale,yscale);  return *this;  }

  /** Rotate the box around the origin */
  virtual base& rotate(Float angle)
    {  a.rotate(angle); b.rotate(angle);  return *this;  }

  /** Translate the origin of the box elsewhere */
  virtual base& translate(T xoffset, T yoffset)
    {  a.translate(xoffset,yoffset); b.translate(xoffset,yoffset);  return *this;  }
  
  /** Returns an axis-aligned rectangle fully enclosing this box, or
      an empty rectangle if no such finite rectangle exists. */
  virtual typename base::Rectangle aarectangle() const {  return a.aarectangle().intersection(b.aarectangle());  }

  virtual base* clone() const {  return new self(*this); };

  virtual std::ostream& put(std::ostream &s) const
    {  s<<"(";  a.put(s);  s<<" && ";  b.put(s);  s<<")";  return s;  };
  
private:
  BBA a;
  BBB b;
};



/** Circular bounding box. */
template<class T=int,class Boolean=bool,class Float=double>
class BoundingCircle : public BoundingBox<T,Boolean,Float> {
private:
  typedef BoundingCircle<T,Boolean,Float> self;
  typedef BoundingBox<T,Boolean,Float> base;

public:
  /** Default constructor: infinite extent */
  BoundingCircle(): infinite(true) { }

  /** Typical constructor */
  BoundingCircle(T centerx_, T centery_, Float radius_ )
    : infinite(false), centerx(centerx_), centery(centery_), radius(radius_)
    {  freshencache();  }

  inline Boolean isinfinite() const {  return infinite;  }
  
  virtual Boolean inside(T x, T y) const {
    if (isinfinite()) return true;
    
    const T xdistance           = x-centerx;
    const T ydistance           = y-centery;
    const T xdistance_sq        = xdistance*xdistance;
    const T ydistance_sq        = ydistance*ydistance;

    return (xdistance_sq + ydistance_sq <= radius_sq);
  }

  /** Only supports scaling with xscale==yscale */
  virtual base& scale( T xscale, T yscale )
    {  radius*=xscale; (void)yscale; assert(xscale==yscale);  freshencache();  return *this;  }
  
  /** Rotating a circle does nothing.  */
  virtual base& rotate( Float ) {  return *this;  }

  virtual base& translate( T xoffset, T yoffset )
    {  centerx+=xoffset;  centery+=yoffset;  freshencache();  return *this;  }

  /** Provides an axis-aligned rectangle based on the circle, but not necessarily
      enclosing all of it due to integer truncation. */
  virtual typename base::Rectangle aarectangle() const {
    return (isinfinite()?
	    typename base::Rectangle() :
	    typename base::Rectangle(T(centerx)-T(radius),T(centery)-T(radius),
				     T(centerx)+T(radius),T(centery)+T(radius)));
  }
  
  virtual base* clone() const {  return new self(*this); };

  virtual std::ostream& put(std::ostream &s) const
    {  return (s << "((" << centerx << "," << centery << ") radius " << radius << ")");  }

private:
  Boolean infinite;

  T centerx;
  T centery;
  Float	radius;

  /* \name Cached values computed from radius,radiusy and centerx,centery */
  //@{
  Float	radius_sq;
  void freshencache() {  radius_sq = radius*radius;  }
  //@}
};




} /* namespace Boundary */

// For temporary backwards compatibility
using namespace Boundary;

#endif /* __BOUNDINGBOX_H__ */
