/** \file
  ExprParser class.
  
  $Header: /u/nn/cvsroot/lissom/src/exprparser.h,v 1.15 2002/10/29 18:32:06 jbednar Exp $
*/

#ifndef ___EXPRPARSER_H___
#define ___EXPRPARSER_H___

#include <stdio.h>
#include <string>
using std::string;

#include "../src/msg.h"
#include "../src/polymorph.h"



/**
  Basic C-like expression parsing.

  This parser works reasonably well, but it is not at all robust or
  well-written.  In particular, the string handling is almost all
  based on old C code, which is much more complicated than the
  equivalent code using a C++ string class would be.  Rather than
  being improved incrementally, it should probably be reimplemented
  using lex/yacc or just replaced altogether with a real interpreter,
  e.g. for Scheme.

  In any case, this interface is for interfacing to legacy code; new
  code should use the ParamStringParser interface if possible.

  At present, the template argument is just a dummy to allow the full
  definition to be included in the header file, but in future, the
  numeric expression type might be a good one to actually use as the
  templated type.
*/
template <class T=int>
class ExprParser {
public:
  /** Constructor.  Takes an expression and an optional handler
      for various sorts of messages. */
  ExprParser(const char* expr, Msg::LevelHandler<>** mhp_i=&(Msg::LevelHandler<>::default_instance()))
    : remaining(expr)  {  mhp = mhp_i;  };

  virtual ~ExprParser() { }

  /** \name Parsing for specific datatypes */
  //@{ (E.g. called within a command)

  /**  Public routine to read an expression of type double. */
  double       cmdf() {  return (double)read_float();  }
  
  /** Integer equivalent of cmdf. Returns type long but warns if can't fit in int. */
  long         cmdi() {
    exprtype result = read_float();
    if (result >= INT_MAX || result <= INT_MIN)
      message(Msg::Warning,"Value " + String::stringrep(double(result)) +
	      " outside the representable range of an integer (" +
	      String::stringrep(INT_MIN) + "," + String::stringrep(INT_MAX) +
	      "); may cause overflow or underflow");
    
    return (long)(floor(0.5+result));
  }
  
  /**  Public routine to read an expression of string type. */
  const char * cmds() {  return getstringval();  }
  //@}

  /** If the given string is the next thing on the remaining chars,
      consume it and return true; otherwise do nothing.  Useful for
      testing for e.g. a keyword or special character. */
  bool consumeifpresent(string s) {
    if (strncmp(s.c_str(), remaining, s.length()) == 0) {
      remaining += s.length();
      return true;
    }
    return false;
  }

  string whatsleft() {  return string(remaining);  }
  
  /** Returns true if no further characters remain. */
  bool empty() const {  return (*remaining==0);  }
  
  static void message(Msg::MessageLevel m, const string& s)
    {  if (mhp && *mhp) (**mhp)(m,s);  }
  
  /** Unified return type for numeric expressions */
  typedef
#ifndef LONG_DOUBLE_UNAVAILABLE
  long
#endif
  double exprtype;

  exprtype        getnumericval();

protected:
  virtual const Typeless& getval();
  const char*     remaining;

private:
  virtual const char* getstringval();
  exprtype        read_float(bool warn=true);
  exprtype        apply_op(exprtype lval, string op, exprtype rval);
  exprtype        next_term(string previous_op="");
  string          next_op();
  exprtype        parse_expr();
  int             op_has_precedence(const string& previous_op, const string& op);
  static Msg::LevelHandler<>** mhp;
};

/* Static variables */
template <class T>
Msg::LevelHandler<>** ExprParser<T>::mhp=&(Msg::LevelHandler<>::default_instance());


/**
  Parses the given string to find the value of the first
  scalar item found (i.e., not an expression).

  This implementation assumes that the scalar is a literal numeric
  value representable in type double.  Derived classes may implement
  return of other types via the Typeless wrapper.
*/
template <class T>
const Typeless& ExprParser<T>::getval()
{
  const char* str = remaining;
  char *rem;
  static Polymorph<double> numval(0.0);
  numval = strtod(str,&rem);
  remaining = rem;

  if (str == remaining) {
    while (*remaining && (*remaining != ')'))
      remaining++;  /* Skip this item */
    message(Msg::Error,"Can't parse expression: `" + string(str) + "'");
    numval=-1;
  }
  
  return numval;
}



/**
  Same as getval but keeps result as a string, without evaluating it.
  This version isn't particularly useful, but derived versions may
  implement parameter substitution.

  It returns a pointer to a static buffer, so the lifetime of the
  result is only until the next call to this function.
*/
template <class T>
const char *ExprParser<T>::getstringval()
{
  const size_t bufsize=65536; // Arbitrary limit
  static char buf[bufsize];
  const char* maxbufptr=buf+bufsize;

  char* bufptr=buf;

  while (remaining && *remaining && bufptr<maxbufptr) {
    /* Copy all characters directly */
    while (*remaining && bufptr<maxbufptr)
      *(bufptr++) = *(remaining)++;
  }
  if (bufptr>=maxbufptr)
    message(Msg::Error,"Reached maximum string size; characters may be lost");
  
  *bufptr='\0';
    
  return buf;
}



/**
  Same as getval but works only with numeric types, and always
  converts them to "exprtype" (long double, if avail, or double, if
  not) so they can be returned cleanly.
*/
template <class T>
typename ExprParser<T>::exprtype ExprParser<T>::getnumericval()
{
  const char* str = remaining;
  
  if (*remaining == '(') {
    exprtype subexpr;
    (remaining)++;    
    subexpr=(parse_expr());
    
    if (*remaining != ')')
      message(Msg::Error,"Missing right parenthesis");
    else
      (remaining)++;
    return subexpr;
  }

  if (*remaining == '-') {
    (remaining)++;        
    return -1.0*next_term();
  }
  
  const Typeless& val = getval();
  if (val.isnumeric())
    return val.numericvalue();
  else
    message(Msg::Error,"Invalid param type in expression `" + string(str) + "'");
 
  return Uninitialized;
}



/** Currently handles single-character operators only. */
template <class T>
typename ExprParser<T>::exprtype ExprParser<T>::apply_op(exprtype lval, string op, exprtype rval)
{
  if      (op == "+")  return lval +  rval;
  else if (op == "-")  return lval -  rval;
  else if (op == "*")  return lval *  rval;
  else if (op == "<")  return lval <  rval;
  else if (op == ">")  return lval >  rval;
  else if (op == "==") return lval == rval;
  else if (op == "!=") return lval != rval;
  else if (op == "<=") return lval <= rval;
  else if (op == ">=") return lval >= rval;
  else if (op == "<?") return std::min(lval,rval);
  else if (op == ">?") return std::max(lval,rval);
  else if (op == "@;") return lval * sin(rval); /* Appallingly bad notation! */
  else if (op == "@:") return lval * cos(rval); /* Appallingly bad notation! */
  
  else if (op == "%" || op == "/") {
    if (rval==0) {
      message(Msg::Error,"Division by zero");
      return Uninitialized;
    }
    if (op == "/") return lval / rval;
    else           return fmod(lval,rval);
  }
  
  message(Msg::Error,"Unknown operator " + op);
  return Uninitialized;
}



template <class T>
int ExprParser<T>::op_has_precedence(const string& previous_op, const string& op)
{
  (void) previous_op; /* Currently ignored */
  return (op=="*" || op=="/" | op=="%");
}



template <class T>
string ExprParser<T>::next_op()
{
  const char* start=remaining;

  while (ispunct(*remaining) && !(*remaining=='(') && !(*remaining==':' && *(remaining+1)==':'))
    (remaining)++;

  const char* end=remaining;

  return string(start,end-start);
}



template <class T>
typename ExprParser<T>::exprtype ExprParser<T>::next_term(string previous_op)
{
  exprtype lval = getnumericval();
  string op = next_op();
  
  /* Reduce strings of operators with precedence */
  while (op_has_precedence(previous_op, op)) {
    if (! *remaining) {
      message(Msg::Error,"Missing right value in `" + string(remaining) + "'");
      return lval;
    }
    lval=apply_op(lval,op,getnumericval());
    previous_op = op;
    op = next_op();
  }
  remaining -= op.length(); /* Push back unused operator */

  return lval;
}



template <class T>
typename ExprParser<T>::exprtype ExprParser<T>::parse_expr()
{
  const char* str = remaining;
  exprtype lval = next_term();

  while (*remaining && *remaining != ')') {
    string op= next_op();
    
    if (! *remaining) {
      message(Msg::Error,"Missing right value in `" + string(str) + "'");
      return lval;
    }
    
    lval = apply_op(lval,op,next_term(op));
  }

  return lval;
}



/** Read in a floating point expression. */
template <class T>
typename ExprParser<T>::exprtype ExprParser<T>::read_float(bool warn)
{
  const char* str = remaining;
  exprtype val = parse_expr();

  if (str == remaining) {
    message(Msg::Error,"No number found in string: `" + string(str) + "'");
    return Uninitialized;
  }
  
  if (warn && *remaining)
    message(Msg::Warning,"Trailing characters in expression ignored: `" + string(remaining) + "'");
  return val;
}



#endif /* ___EXPRPARSER_H___ */
