/** \file
  Generic algorithms which require components from the standard library; also
  includes those without such requirements by reference.

  $Header: /u/nn/cvsroot/lissom/src/generic_stdlib.h,v 1.14 2001/11/24 23:27:39 jbednar Exp $
*/

#ifndef __GENERIC_STDLIB_H__
#define __GENERIC_STDLIB_H__

#include <map>
#include <utility>
#include <functional>
#include <string>
using std::string;
#include <vector>
#include <complex>

#include "genericalgs.h"


namespace Generic {

/** Specialization of general delete_contents algorithm for a map of pointers. */
template<class key_type, class data_type>
void delete_contents(std::map<key_type, data_type*>& m)
{
  typedef typename std::map<key_type, data_type*>::iterator iterator;
  for (iterator i=m.begin(); i!=m.end(); i++) {
    delete (*i).second;
    m.erase(i);
  }
}


/** Variant of make_pair which keeps the left item const; useful for maps. */
template <class T1, class T2>
inline std::pair<const T1, T2> make_pair_leftconst(const T1& x, const T2& y) {
  return std::pair<const T1, T2>(x, y);
}


/**
  Version of unary_function suitable for non-templated functions which
  must take a specific type of argument.
*/
template <class Arg, class Result>
struct unary_virtual_function : public std::unary_function<Arg, Result> {
  virtual ~unary_virtual_function() { }
  /** Call the function */
  virtual Result operator() ( Arg ) =0;
};



/**
  Predicate object applicable to any container of pointers to
  objects which have a name() member function.
*/
template <class T, class N=string>
struct name_match : public std::unary_function<T*,bool> {
  name_match(const N& name_i) : name(name_i) { }
  /** Returns whether the supplied object has a matching name. */
  bool operator() ( T* a) {  return name == a->name();  }
private:
  const N& name;
};



/** Function object returning the maximum of the two given objects of
    identical type. */
template <class T>
struct max_fnobj : public std::binary_function<T, T, T> {
  T operator()(const T& x, const T& y) const { return std::max(x,y); }
};



/**
  Insert a named object into any pointer container supporting
  push_back, overwriting and deleting any existing object with that
  name (using the objects name() function).  Since the existing object
  may be deleted, the container must own the objects to which it
  points.
*/
template <class T, class C, class N>
void insert_named(C& container, const N& name, T* objptr)
{
  typename C::iterator i =
    std::find_if(container.begin(),container.end(),name_match<T>(name));
  if (i==container.end())
    container.push_back(objptr);
  else {
    delete (*i);
    (*i) = objptr;
  }
}


/**
  Returns a pointer to the first object found with the given name()
  in the given container.  The type T is the underlying type (not
  the pointer type) of the objects pointed to by each element of
  the container; there doesn't appear to be any way to deduce it
  from C::value_type (which should be T*).
*/
template <class T, class C, class N>
T* find_named(C& container, const N& name)
{
  typename C::iterator i =
    std::find_if(container.begin(),container.end(),name_match<T>(name));
  return (i==container.end() ? 0 : (*i));
}



/**
  Returns a pointer to the first object found with the given name()
  in the given container.  (const version of find_named.)
*/
template <class T, class C, class N>
const T* find_named_const(const C& container, const N& name)
{
  typename C::const_iterator i =
    std::find_if(container.begin(),container.end(),name_match<T>(name));
  return (i==container.end() ? 0 : (*i));
}



/** Sum of the elements of a vector */
template <class T> inline T sum(const std::vector<T>& A)
{  return std::accumulate(A.begin(),A.end(),T(0));  }
 


/** Function object for constructing a complex number from
    a (rho,theta) pair.  For use with e.g. std::inner_product.
*/
template<class T>
struct Polar : public std::binary_function<T,T,std::complex<T> > {
  inline std::complex<T> operator() (const T& rho,const T& theta) const

/* Workaround for bug in at least egcs-1.1.2, possibly GCC 2.95 as well */
#if (__GNUC__ && (__GNUC__ < 3))
    {  return polar<T>(rho,theta);  }
#else
    {  return std::polar<T>(rho,theta);  }
#endif
    
};


} /* namespace Generic */ 
#endif /* __GENERIC_STDLIB_H__ */
