/** \file
  Generic algorithms not assuming any particular datatype.

  $Header: /u/nn/cvsroot/lissom/src/genericalgs.h,v 1.12 2003/01/16 04:11:58 jbednar Exp $
*/

#ifndef __GENERICALGS_H__
#define __GENERICALGS_H__

#include <algorithm>

/**
  Use whenever a pair of iterators are expected, as in:
  std::for_each(ISEQ(container),std::mem_fun(&Container::update)).

  (based on a proposal in Stroustrup's "C++, 3rd Ed.", section 18.3.1)
*/
#ifndef ISEQ
#define ISEQ(container) (container).begin(),(container).end()
#endif


/**
  Use to avoid the tricky C++ member function calling syntax.

  (based on the 3/2000 C++ FAQ LITE question 30.5)
*/
#ifndef callMemberFunction
#define callMemberFunction(object,ptrToMember)  ((object).*(ptrToMember))
#endif


/** Namespace for generic algorithms not assuming any particular datatype. */
namespace Generic {



/** Crops to the given (inclusive) range, truncating at each end of the range. */
template <class T>
inline const T crop(const T lower, const T upper, const T x) {
  return std::max(lower,std::min(upper,x));
}
  

 
/** Circularly aliases (wraps) to the given (inclusive) range, as angles do.
 I have no idea how I came up with this algorithm; it can probably be simplified.
 Files using this routine will need to add:
 \verbatim
 #include <cmath>
 \endverbatim
*/
template <class T>
inline const T wrap(const T lower, const T upper, const T x) {
  const T range=upper-lower;
  return lower + (T)fmod(x-lower + 2*range*(1-floor(x/(2*range))),
			 range);
}
 

 
/** Round the given floating-point value to the nearest integer */
template <class T>
inline int round(const T x) {  return int(x+0.5);  }



/**
  Deletes all pointed-to items and removes them from the given container;
  any standard container of pointers should work.
*/
template<class Container>
void delete_contents(Container& c)
{
  while(!c.empty()) {
    delete c.back();
    c.pop_back();
  }
}



/**
  Version of delete_contents for containers of pointers to containers;
  deletes all pointed-to items in the subcontainers, then deletes the
  subcontainers themselves.
*/
template<class Container>
void nested_delete_contents(Container& c)
{
  while(!c.empty()) {
    delete_contents(*c.back());
    delete c.back();
    c.pop_back();
  }
}



/** Apply a lower threshold to any container, with a cropping value
    that can be different from the threshold. */
template <class ForwardIterator, class T>
void lower_threshold(ForwardIterator first, ForwardIterator last,
		     const T& threshold, const T& new_value)
{  for (; first != last; ++first) if (*first < threshold) *first=new_value;  }



/** Copy a container into a new one, for each element applying a lower threshold. */
template <class InputIterator, class ForwardIterator, class T>
void lower_threshold(InputIterator   ifirst, InputIterator   ilast,
		     ForwardIterator ofirst,
		     const T& threshold, const T& new_value)  {
  for (; ifirst != ilast; ++ifirst,++ofirst)
    *ofirst = (*ifirst < threshold) ? new_value : *ifirst;
}



/** Apply an upper threshold to any container, with a cropping value
    that can be different from the threshold.  */
template <class ForwardIterator, class T>
void upper_threshold(ForwardIterator first, ForwardIterator last,
		     const T& threshold, const T& new_value)
{  for (; first != last; ++first) if (*first > threshold) *first=new_value;  }



/** Copy a container into a new one, for each element applying an upper threshold. */
template <class InputIterator, class ForwardIterator, class T>
void upper_threshold(InputIterator   ifirst, InputIterator   ilast,
		     ForwardIterator ofirst,
		     const T& threshold, const T& new_value)  {
  for (; ifirst != ilast; ++ifirst,++ofirst)
    *ofirst = (*ifirst > threshold) ? new_value : *ifirst;
}



/** Same as std::accumulate but accepts an object to compute the value
    instead of one to compute and do the summation both; often the
    former is already available but a special function would have to
    be written to get the latter. */
template <class InputIterator, class T, class Operation >
T accumulate(InputIterator beg, InputIterator end, T init, Operation op)
{
  for ( ; beg != end; ++beg)
    init = init + op(*beg);
  return init;
}

/** Same as std::accumulate called with a binary operator, but splits
    the operation into two operations, one called first on the value
    and the other called to combine the two.  The standard version
    requires that the same function do both; such functions are rarely
    already available.  */
template <class InputIterator, class T, class Operation, class BinaryOperation >
T accumulate(InputIterator beg, InputIterator end, T init, Operation op, BinaryOperation binop)
{
  for ( ; beg != end; ++beg)
    init = binop(init,op(*beg));
  return init;
}


/** Substitutes for system function not always available */
template <class T>
inline T hypot(T x, T y)
{  return sqrt(x*x+y*y);  }


} /* namespace Generic */ 
#endif
