/** \file
  OneDHistogram class and associated functions.

  $Header: /u/nn/cvsroot/lissom/src/histogram.h,v 1.15 2002/10/29 18:32:06 jbednar Exp $
*/

#ifndef __HISTOGRAM_H__
#define __HISTOGRAM_H__


#include <iterator>
#include <functional>
#include <vector>
#include <numeric>
#include <complex>

#include "../src/index_iterator.h"
#include "../src/genericalgs.h"
#include "../src/generic_stdlib.h"
#include "../src/bounded.h"
#include "../src/mathconstants.h"


namespace  Histo {


/** Abstract base class for Histogram-like objects consisting of a
 finite one-dimensional list of bins with counts and values (possibly
 identical to the counts) in each bin, plus total counts and values.

 This class is designed to include some classes which are not,
 strictly speaking, histograms, e.g. if their total counts are less
 than the sum of the individual counts, or if individual counts are
 less than zero.
*/
template<class Count=int, class Value=double>
class OneDBinList {
protected:
  std::vector<Count> counts;
  std::vector<Value> values;
  Count total_c;
  Value total_v;

  template<class C, class V>  
  friend OneDBinList<C,V> operator-(const OneDBinList<C,V> &A, const OneDBinList<C,V> &B);
  
private:
  typedef Generic::index_iterator< std::vector<Value>, Scale::Linear<Value> > index_i;
  typedef Scale::Linear<Value> ScaleType;
  
public:
  typedef typename std::vector<Value>::size_type size_type;
  typedef Value value_type;

  /** A value whose nominal range is 0..1.0 */
  typedef Bounded::Magnitude Magnitude;

  /** For vector averages; complex acts like a geometrical vector and
      is universally available so it is used here. */
  typedef std::complex<Value> Vector;

  static size_type num_bins_default;

  /** To avoid divide-by-zero errors in various member functions,
      requires that num_bins>0. */
  OneDBinList(const size_type num_bins=num_bins_default)
    : counts(num_bins), values(num_bins), total_c(0), total_v(0),
      cached_vector_sum_count(0), cached_vector_sum(0,0),
      cached_sum_value_count(0),  cached_sum_value(0)
    {  assert(num_bins>0);  }


  /** Add the specified value to the histogram bin specified */
  // Could specify as h[bin]+=val instead
  OneDBinList& add(size_type bin, Value val=1){
    total_c++;
    total_v+=val;
    counts[bin]++;
    values[bin]+=val;
    return *this;
  }

  /** Store the given value as the value for the given bin if it's the
      highest value seen so far.  Useful for maintaining a
      distribution of maxiumum values for different bins, as opposed
      to the summations typical to histogram bins.  Note that each
      call will increase the total_value() and total_count() (and thus
      decrease the value_mag() and count_mag()) even if the value
      doesn't happen to be the maximum seen so far, since each data
      point still helps improve the sampling and thus the confidence.

  */
  OneDBinList& keep_peak(size_type bin, Value val) {
    total_c++;
    total_v+=val;
    counts[bin]++;
    if (val > values[bin]) values[bin]=val; 
    return *this;
  }

  /** \name    Summary access functions

      Report statistics about the whole histogram.
  */
  //@{
  inline size_type num_bins() const {  return counts.size();  }

  /** The sum of the current values.
      
      The result is cached when first computed, and the cached value
      is used unless the total_count() changes (i.e. unless new values
      have been added since the sum was computed.)  Thus this routine
      may be called multiple times, e.g. for computation of other
      quantities such as average_value(), with no performance penalty.
  */
  inline Value sum_value() const {
    if (total_count() != cached_sum_value_count) {
      cached_sum_value = std::accumulate(ISEQ(values),Value(0));
      cached_sum_value_count = total_count();
    }
    return cached_sum_value;
  }
  
  /** The sum of the current counts. */
  inline Count sum_count()    const {  return std::accumulate(ISEQ(counts),Count(0));  }

  /** The arithmetic mean of the current values. */
  inline Value average_value()const {  return sum_value()/Value(num_bins());  }

  /** The largest value() of any bin. */
  Value max_value()    const {  return *(std::max_element(ISEQ(values)));  }
  Value min_value()    const {  return *(std::min_element(ISEQ(values)));  }
  Count max_count()    const {  return *(std::max_element(ISEQ(counts)));  }
  Count min_count()    const {  return *(std::min_element(ISEQ(counts)));  }

  /** The bin_number with the largest value */
  size_type max_value_bin() const
    {  return std::max_element(ISEQ(values))-values.begin();  }

  /** The bin with the largest value, as a number between 0.0 and 1.0 */
  //Magnitude max_value_bin_mag() const {  return bin_mag(max_value_bin());  }
  
  /** Returns value(max_value_bin()) as a proportion of the
      sum_value(), scaled to 0.0 to 1.0.  This quantity is a measure
      of the how strongly the histogram is biased towards the
      max_value_bin().  For smooth, single-lobed distributions with an
      inclusive, non-cyclic range, this quantity is an analog to
      vector_selectivity.  To be a precise analog for arbitrary
      distributions, it would need to compute some measure of the
      selectivity that works like the weighted_average() instead of
      the max_value_bin().  The scaling is such that if all bins are
      identical, the selectivity is 0.0, and if all bins but one are
      zero, the selectivity is 1.0.
  */
  inline Magnitude relative_selectivity() const {
    /* With only one bin, one is always fully selective; special
       case to avoid divide-by-zero errors.  */
    if (num_bins()<=1) return 1.0;
    const Magnitude proportion = safe_divide(value(max_value_bin()),sum_value());
    const Magnitude offset = 1.0/num_bins();
    const Magnitude scaled = (proportion-offset)/(1.0-offset);
    /* Since we take the peak, negative selectivities shouldn't be
       possible, but just in case... */
    return Generic::crop(Magnitude(0.0),Magnitude(1.0),scaled);
    return scaled;
  }

  /** Returns the sum of each value times its bin number */
  Value weighted_sum() const
    {  return std::inner_product(ISEQ(values),index_i(values.begin()),Value(0));  }

  /** Returns the average bin_number as a number from 0.0 to 1.0,
      weighting each bin by its value.  This quantity is a continuous
      measure of the max_value_bin() of the distribution for a
      non-wrapping distribution, i.e.  one whose upper and lower
      bounds are distinct.  */
  inline Magnitude weighted_average() const
    {  return safe_divide((safe_divide(weighted_sum(),sum_value())),Value(num_bins()-1));  }


  /** Returns the number of times that undefined values have been
      returned from calculations for any instance of this class,
      e.g. calls to vector_direction() or vector_selectivity() when no
      value is non-zero.  Useful for warning users when the values are
      not meaningful.  */
  inline static size_type undefined_values() {  return undefined_vals;  }
  inline static void      undefined_values_reset(size_type val=0) {  undefined_vals=val;  }
  
  /** The sum of all of the values ever provided for each bin.  For a
      simple histogram this will be the same as sum_value(), but for
      other OneDBinList types it often differs from sum_value(). */
  inline Value total_value()  const {  return total_v;  }
  
  /** The total number of values that have ever been provided for each
      bin.  For a simple histogram this will be the same as
      sum_counts(), but for other OneDBinList types it often
      differs. */
  inline Count total_count()  const {  return total_c;  }
  
  //@}
  

  
  /** \name    Vector summary functions

      Summary functions computed from the vector_sum().
  */
  //@{
  /** Returns the vector sum of all the values indexed by their
      bin_number.  A vector is constructed from each bin: the length
      is the value() of the bin, and the angle is the bin_number as a
      percentage of a circle, i.e. 2*PI*bin_number()/num_bins().
      
      The result is cached when first computed, and the cached value is
      used unless the total_count() changes (i.e. unless new values have
      been added since the sum was computed.)  Thus this routine may be
      called multiple times, e.g. for the other vector quantities such
      as vector_average(), with no performance penalty.
  */
  Vector vector_sum() const {
    if (total_count() != cached_vector_sum_count) {
      cached_vector_sum = std::inner_product
	(ISEQ(values), index_i(values.begin(),ScaleType(2*(Math::pi)/Value(num_bins()))),
	 Vector(0,0),

/* Workaround for bug in at least egcs-1.1.2, possibly GCC 2.95 as well */	 
#if (__GNUC__ && (__GNUC__ < 3))
	 plus<Vector>(),
#else
	 std::plus<Vector>(),
#endif
	 
	 Generic::Polar<Value>());
      cached_vector_sum_count = total_count();
    }
    
    return cached_vector_sum;
  }

  /** Returns the vector_average of all the values indexed by their
      bin_number, i.e. the vector_sum()/num_bins(). */
  inline Vector vector_average() const {  return vector_sum()/Value(num_bins());  }

  /** Returns the direction of the vector_sum() on a scale from 0 to
      1.0.  This quantity is a continuous measure of the
      max_value_bin() of the distribution, assuming the distribution
      wraps around from the high bin to the low bin.  It has more
      precision than the bin_number of the peak (i.e.
      bin_mag(max_value_bin())) because it is computed from the entire
      distribution instead of just the bin_number of the peak bin.
      However, it requires uniform sampling for correct results, which
      is not always possible.

      The vector_direction() is not meaningful when
      abs(vector_sum())==0, since a zero-length vector has no
      direction.  To find out whether such cases occurred, you can
      compare the value of undefined_values() before and after
      a series of calls to this function.
  */
  inline Magnitude vector_direction() const {
    if (std::abs(vector_sum())==0) undefined_vals++; /* Optional check */
    return Generic::wrap(0.0,1.0,std::arg(vector_sum())/(2*(Math::pi)));
  }

  /** Returns abs(vector_sum())/sum_value(), i.e. the ratio of the
      length of the vector sum to the sum of the vector lengths.  This
      is a vector-based measure of the peakedness of the histogram.
      If only a single bin has a non-zero value(), the selectivity
      will be 1.0, and if all bins have the same value() then the
      selectivity will be 0.0.  Other distributions will result in
      intermediate values.

      For a histogram with a sum_value() of zero (i.e. all bins
      empty), the selectivity is undefined.  Assuming that one will
      usually be looking for high selectivity, we return zero in such
      a case so that high selectivity will not mistakenly be claimed.
      To find out whether such cases occurred, you can compare the
      value of undefined_values() before and after a series of
      calls to this function.
  */
  inline Magnitude vector_selectivity() const
    {  return safe_divide(std::abs(vector_sum()),sum_value());  }
  //@}


  
  /** \name    Element access functions

      Return the value of a single bin in the histogram.  The _mag functions
      return the value as a portion of the total_value() or total_count().
  */
  //@{
  inline Value         value(size_type i) const {  return values[i];  }
  inline Count         count(size_type i) const {  return counts[i];  }
  inline Magnitude value_mag(size_type i) const {  return safe_divide(values[i]*1.0,total_value());  }
  inline Magnitude count_mag(size_type i) const {  return safe_divide(Magnitude(counts[i]),Magnitude(total_count()));  }


  /** Returns the value of the given bin as a proportion of the
      sum_value().  This quantity is on a scale from 0.0 (the
      specified bin is zero and others are nonzero) to 1.0 (the
      specified bin is the only nonzero bin).  If the histogram is
      empty the result is undefined; in such a case zero is returned
      and the value returned by undefined_values() is incremented.
  */
  Magnitude relative_value(size_type bin) const
    {  return safe_divide(value(bin),sum_value());  }
  //@}

  
  /** \name    Helper functions */
  //@{
  /** Converts a binmag (a value from 0 to 1.0) into the corresponding
      bin_number, assuming an exclusive upper bound.  (I.e., assumes
      that the highest bin number corresponds to 1.0-(1/num_bins())
      instead of 1.0).  
      
      A tiny constant float_adjust is added to the binmag after the
      binmag is scaled into the range num_bins(), to correct for any
      floating-point truncation errors during previous calculations.
      Such errors can be significant for binmags that were originally
      integer bin values (e.g. bin 1 out of 3), because when they are
      transformed back into bin_values here, any slight floating-point
      error can cause a full bin's worth of error (e.g. choosing bin 0
      instead of bin 1).  For any num_bins() likely to fit into memory
      there should be no effect on the accuracy for magnitudes arising
      from other, continuous sources.
  */
  size_type bin_number(Magnitude binmag) {
    return Generic::crop(size_type(0),size_type(num_bins())-size_type(1),
			 size_type(binmag*num_bins()+float_adjust));
  }


  /** Converts a binmag (a value from 0 to 1.0) into the corresponding
      bin_number, assuming an inclusive upper bound.  (I.e., assumes
      that the highest bin number corresponds to 1.0 instead of
      1.0-(1/num_bins())).  See bin_number(). */
  size_type bin_number_inclusive(Magnitude binmag) {
    return Generic::crop(size_type(0),size_type(num_bins()),
			 size_type(binmag*(num_bins()-1)+float_adjust));
  }
  

  /** Converts an integer bin_number into the corresponding Magnitude,
      a number from 0 to 1.0. */
  Magnitude bin_mag(size_type bin) {  return Magnitude(bin)/Magnitude(num_bins());  }
  //@}

private:

  /** Division routine that avoids division-by-zero errors
      (returning zero in such cases) but keeps track of them
      for undefined_values(). */
  template <class T>
  inline T safe_divide(const T numerator, const T denominator) const 
    {  return (denominator==0? undefined_vals++, 0 : numerator/denominator);  }

  static size_type undefined_vals;

  /** These variables are marked mutable because they don't affect the
      outwardly visible state.  */
  //@{
  mutable Count  cached_vector_sum_count;
  mutable Vector cached_vector_sum;
  mutable Count  cached_sum_value_count;
  mutable Value  cached_sum_value;
  //@}

  static const value_type float_adjust; 
};



/* Static variables */
template<class Count, class Value> typename OneDBinList<Count,Value>::size_type
OneDBinList<Count,Value>::num_bins_default=1;  /**< Arbitrary default value */
template<class Count, class Value> typename OneDBinList<Count,Value>::size_type
OneDBinList<Count,Value>::undefined_vals=0;

/** 1E-16 would be too small to help for doubles.  1E-15 appeared
    sufficient in tests, but we chose a slightly larger (yet still 
    tiny) value for safety. */
template<class Count, class Value> const typename OneDBinList<Count,Value>::value_type
OneDBinList<Count,Value>::float_adjust=1E-14; 


/** Computes the difference of two histograms.  The scale becomes the
  sum of the scales of the two source histograms.  (The total_value and
  total_count combine additively, since the number of samples
  increases even when subtracting.)
*/  
template<class Count, class Value>
OneDBinList<Count,Value> operator-(const OneDBinList<Count,Value> &A, const OneDBinList<Count,Value> &B)
{
  typename OneDBinList<Count,Value>::size_type N = A.num_bins();
  assert(N==B.num_bins());

  OneDBinList<Count,Value> tmp(N);

  for (typename OneDBinList<Count,Value>::size_type i=0; i<N; i++) {
    tmp.values[i] = A.values[i] - B.values[i];
    tmp.counts[i] = A.counts[i] - B.counts[i];
  }
  
  tmp.total_v = A.total_v + B.total_v;
  tmp.total_c = A.total_c + B.total_c;
  
  return tmp;
}



/** General-purpose one-dimensional histogram.

  Same as OneDBinList except that the individual counts are guaranteed
  to be non-negative and to add up to the total counts, i.e. guaranteed
  to be a real Histogram.  Having it be a public derived class of
  OneDBinList may be a semantic error, since not all operations
  on OneDBinList should be used on this class; this decision may
  need to be revisited at some point.
*/
template<class Count=int, class Value=double>
class OneDHistogram : public OneDBinList<Count,Value> {
public:
  typedef typename OneDBinList<Count,Value>::size_type size_type;
  OneDHistogram(const size_type num_bins)
    : OneDBinList<Count,Value>(num_bins) { }
};


 
/** Difference between two one-dimensional histograms. */
template<class Count=int, class Value=double>
class OneDHistogramDifference : public OneDBinList<Count,Value> {
public:
  OneDHistogramDifference(const OneDBinList<Count,Value> &A,
			  const OneDBinList<Count,Value> &B)
    : OneDBinList<Count,Value>(A-B) { }
};
 

 
} /* namespace Histo */
#endif /* __HISTOGRAM_H__ */
