/** \file
  HistogramImage class.

  Can be extended to support a OneDHistogramDifference as well, but
  will need some minor modifications (zero_pixel would be set to the
  center of the plot area, bin_fullness_range *might* be cut in half;
  bin_fullness_pixelnum and draw_element would need to be modified.)

  $Header: /u/nn/cvsroot/lissom/src/histogramimage.h,v 1.14 2004/09/29 17:39:19 jbednar Exp $
*/

#ifndef __HISTOGRAM_IMAGE_H__
#define __HISTOGRAM_IMAGE_H__

// Uses roundabout paths to the current directory to avoid Cray C bug
#include "../src/image.h"
#include "../src/pixel.h"
#include "../src/colorlookup.h"
#include "../src/histogram.h"
#include "../src/msg.h"
#include "../src/stringutils.h"

namespace  Plot {


/** Image of a histogram, with an optional border.
    
  The border is not as important here as it is for the MatrixImage<>
  class, but it is supported so that images from both classes can be
  freely intermixed, with all having a uniform border.
*/
template<class PixelType=RGBPixel<>, class PixelMatrix= typename MatrixType<PixelType>::rectangular >
class OneDHistogramImage : public AARImage<PixelType,PixelMatrix> {
private:

  /** Copied here explicitly to avoid apparent bug in GCC 2.95 */
  typedef typename AARImage<PixelType,PixelMatrix>::PixelSubscript PixelSubscript;

  /** Index into a histogram */
  typedef int Subscript;
  
  /** Width of surrounding border */
  PixelSubscript border;
  
  /** Number of bins in the histogram */
  double num_bins;
  
  /** Ratio between the number of bins in the histogram and the image size */
  double size_scale;
  
  /** Whether the plot should be horizontal or vertical */
  bool vertical;
  
  /** Range of bin_fullnesses to show out of the full range 1.0 */
  Bounded::Magnitude bin_fullness_range;
  
  /** Number of the lowest pixel for the value axis */
  PixelSubscript min_pixelnum;
  
  /** Number of the origin pixel for the value axis */
  PixelSubscript zero_pixelnum;
  
  /** Number of the highest pixel for the value axis */
  PixelSubscript max_pixelnum;

  /** Background color */
  PixelType bg;

  /** Foreground color */
  PixelType fg;

  /** If true, mark the zero axis in the plot */
  bool markzero;
  
  /** If true, outline each bin in the foreground color */
  bool outline;
  
  /** Returns the top pixel row corresponding to the given histogram element */
  inline PixelSubscript bin_number_pixelnum(const Subscript b) const
    {  return PixelSubscript(b*size_scale+border);  }
  
  /** Returns pixel corresponding to the tip of the specified bin fullness. */
  /* Will need to be modified for a symmetric plot */
  inline PixelSubscript bin_fullness_pixelnum(const Bounded::Magnitude bin_fullness) const {
    return Generic::crop(min_pixelnum,max_pixelnum,
			 min_pixelnum+Subscript((1+max_pixelnum-min_pixelnum)*
						(bin_fullness/bin_fullness_range)));
  }
  
  void draw_element(const Subscript bin_number, const Bounded::Magnitude bin_fullness,
		    PixelType p=default_fg, const Bounded::Magnitude previous_bin_fullness=0);
  void draw_element_vertical(const Subscript bin_number, const Bounded::Magnitude bin_fullness,
			     PixelType p=default_fg, const Bounded::Magnitude previous_bin_fullness=0);

  static void message(Msg::MessageLevel m, const string& s, const bool terminate=True) {
    // Written in this awkward fashion to avoid bug (ICE) in GCC 2.95
    const bool nonnull = (mhp!=0 && (*mhp)); 
    if (nonnull) (**mhp)(m,s,terminate);
  }

  static Msg::LevelHandler<>** mhp;
  
public:
  /** \name Constructors and Destructors */
  //@{
  
  /** Empty image constructor */
  OneDHistogramImage()
    : border(0), num_bins(0), size_scale(1.0), vertical(false),
      bin_fullness_range(0), min_pixelnum(0), zero_pixelnum(0), 
      max_pixelnum(0), markzero(false), outline(true) { }
  
  template <class Histogram>
  OneDHistogramImage(const Histogram &h, bool vertical=false,
		     double size_scale_i=1.0, double aspect_ratio_i=1.0, 
		     Bounded::Magnitude bin_fullness_range_i=1.0, int borderwidth=0,
		     const ColorLookup<PixelType>& L=HueColorLookup<PixelType>(),
		     const PixelType& bordercolor=default_border, 
		     const PixelType& bg_=default_bg,
		     const PixelType& fg_=default_fg,
		     bool show_outline=true,
		     string title="");
  
    //@}
};

/* Static variables */
template<class PixelType, class PixelMatrix>
Msg::LevelHandler<>** OneDHistogramImage<PixelType,PixelMatrix>::
mhp=&(Msg::LevelHandler<>::default_instance());



/** Constructor taking a single histogram, to plot one dimension.

  The aspect ratio is the ratio between the number of bins and the
  length of the representation to use for the bin fullnesses.

  L determines the mapping from a bin number (on a scale 0-1.0) to a
  color.
*/
template<class PixelType, class PixelMatrix>
template <class Histogram>
OneDHistogramImage<PixelType,PixelMatrix>::OneDHistogramImage
  (const Histogram &h, bool vertical_i, double size_scale_i, double aspect_ratio_i, 
   Bounded::Magnitude bin_fullness_range_i, int borderwidth, const ColorLookup<PixelType>& L,
   const PixelType& bordercolor, const PixelType& bg_, const PixelType& fg_, bool show_outline, string title)
  : AARImage<PixelType>(2*borderwidth+Subscript(h.num_bins()*size_scale_i*(vertical_i?1.0:aspect_ratio_i)),
			2*borderwidth+Subscript(h.num_bins()*size_scale_i*(vertical_i?aspect_ratio_i:1.0))),
  border(borderwidth), num_bins(h.num_bins()), size_scale(size_scale_i), vertical(vertical_i),
  bin_fullness_range(bin_fullness_range_i), min_pixelnum(borderwidth), zero_pixelnum(borderwidth), 
  max_pixelnum(borderwidth+PixelSubscript(h.num_bins()*size_scale_i*aspect_ratio_i)),
  bg(bg_), fg(fg_), markzero(false), outline(show_outline) {

  draw_border(borderwidth,bordercolor);
  typedef typename Histogram::size_type Subscript;
  message(Msg::Overwhelm,"Histogram for "+title+":"+"\n",false);

  Bounded::Magnitude previous_bin_fullness=0;
  if (vertical)
    for (Subscript b=0; b<h.num_bins(); b++) {
      const Bounded::Magnitude bin_fullness=h.value_mag(b);
      draw_element_vertical(b,bin_fullness,L((b*1.0)/h.num_bins()),previous_bin_fullness);
      message(Msg::Overwhelm,String::stringrep(b)+"\t"+String::stringrep(bin_fullness)+"\n",false);
      previous_bin_fullness=bin_fullness;
    }
  else
    for (Subscript b=0; b<h.num_bins(); b++) {
      const Bounded::Magnitude bin_fullness=h.value_mag(b);
      draw_element(b,bin_fullness,L((b*1.0)/h.num_bins()),previous_bin_fullness);
      message(Msg::Overwhelm,String::stringrep(b)+"\t"+String::stringrep(bin_fullness)+"\n",false);
      previous_bin_fullness=bin_fullness;
    }

  message(Msg::Verbose,"Histogram statistics for "+title+
	  ":\t WeightedAverage="+String::stringrep(h.weighted_average())+
	  " VectorDirection="+String::stringrep(h.vector_direction())+"\n",false);
}



/** Draws a representation of the specified bin using the given color.
    
  bin_fullness ranges from 0 to 1.  The background and foreground
  colors are used for the area not representing the histogram bin
  and the zero axis (if any), respectively.
*/
template<class PixelType, class PixelMatrix>
void OneDHistogramImage<PixelType,PixelMatrix>::draw_element_vertical
  (const Subscript bin_number, const Bounded::Magnitude bin_fullness, PixelType p,
   const Bounded::Magnitude previous_bin_fullness)
{
  const PixelSubscript prevp = bin_fullness_pixelnum(previous_bin_fullness);
  const PixelSubscript pivot = bin_fullness_pixelnum(bin_fullness);
  const PixelSubscript binpl = bin_number_pixelnum(bin_number);
  const PixelSubscript binph = bin_number_pixelnum(bin_number+1);
    
  /* Representation consists of three boxes: bar, background, and (optional) origin, 
     plus an optional outline */
  draw_rectangle(  binpl,   zero_pixelnum, binph, pivot,           p );
  draw_rectangle(  binpl,   pivot,         binph, max_pixelnum,    bg);
  if (outline) {
    const PixelSubscript off = (previous_bin_fullness<bin_fullness? -1 : 0);
    draw_rectangle(binpl,     pivot,       binph, pivot+1,         fg);
    draw_rectangle(binpl+off, prevp+off+1, binpl+off+1, pivot-off, fg);
  }
  if (markzero)
    draw_rectangle(binpl,   zero_pixelnum, binph, zero_pixelnum+1, fg);
}


 
/** Draws a representation of the specified bin using the given color.
    
  bin_fullness ranges from 0 to 1.  The background and foreground
  colors are used for the area not representing the histogram bin
  and the zero axis (if any), respectively.
*/
template<class PixelType, class PixelMatrix>
void OneDHistogramImage<PixelType,PixelMatrix>::draw_element
  (const Subscript bin_number, const Bounded::Magnitude bin_fullness, PixelType p,
   const Bounded::Magnitude previous_bin_fullness)
{
  const PixelSubscript prevp = bin_fullness_pixelnum(previous_bin_fullness);
  const PixelSubscript pivot = bin_fullness_pixelnum(bin_fullness);
  const PixelSubscript binpl = bin_number_pixelnum(bin_number);
  const PixelSubscript binph = bin_number_pixelnum(bin_number+1);
    
  /* Representation consists of three boxes: bar, background, and (optional) origin, 
     plus an optional outline */
  draw_rectangle(  nrows()-zero_pixelnum, binpl,   nrows()-pivot,           binph, p );
  draw_rectangle(  nrows()-pivot,         binpl,   nrows()-max_pixelnum,    binph, bg);
  if (outline) {
    const PixelSubscript off = (previous_bin_fullness<bin_fullness? -1 : 0);
    draw_rectangle(nrows()-pivot,         binpl,   nrows()-pivot-1,         binph, fg);
    draw_rectangle(nrows()-prevp-off-1,   binpl+off,nrows()-pivot+off,      binpl+off+1,fg);
  }
  if (markzero)	   		           		           
    draw_rectangle(nrows()-zero_pixelnum, binpl,   nrows()-zero_pixelnum+1, binph, fg);
}


} /* namespace Plot */
#endif /* __HISTOGRAM_IMAGE_H__ */
