/** \file
  ImageGrid class.

  $Header: /u/nn/cvsroot/lissom/src/imagegrid.h,v 1.6 2002/10/29 18:32:06 jbednar Exp $
*/

#ifndef __IMAGEGRID_H__
#define __IMAGEGRID_H__

#include "matrix.h"
#include "image.h"

namespace  Plot {


/** Image composed of a rectangular grid of smaller images with an optional border between them. */
template<class PixelType=RGBPixel<>, class PixelMatrix=typename MatrixType<PixelType>::rectangular >
class ImageGrid : public AARImage<PixelType,PixelMatrix> {

  /** Copied here explicitly to avoid apparent bug in GCC 2.95 */
  typedef typename AARImage<PixelType,PixelMatrix>::PixelSubscript PixelSubscript;

  /** Total height required for all rows */
  template <class MatrixOfImages>
  PixelSubscript totalheight(const MatrixOfImages& ms, int borderwidth=0, int internalborderwidth=0 ) {
    PixelSubscript t=borderwidth;
    for (typename MatrixOfImages::size_type r=0; r<ms.nrows(); r++)
      t += mat::max_nrows(mat::submatrix(ms,r,r+1,0,ms.ncols()))+internalborderwidth;
    return t - (ms.nrows()>0?internalborderwidth : 0) + borderwidth;
  }

  /** Total width required for all columns */
  template <class MatrixOfImages>
  PixelSubscript totalwidth(const MatrixOfImages& ms, int borderwidth=0, int internalborderwidth=0 ) {
    PixelSubscript t=borderwidth;
    for (typename MatrixOfImages::size_type c=0; c<ms.ncols(); c++)
      t += mat::max_ncols(mat::submatrix(ms,0,ms.nrows(),c,c+1))+internalborderwidth;
    return t - (ms.ncols()>0?internalborderwidth : 0) + borderwidth;
  }

  
public:
  /** Constructor taking a matrix of arbitrary-sized images and concatenating them.
      Borders use the default_bg instead of the default border to give the effect
      of putting outlined images on a larger page, with the same border around and
      between the images. */
  template <class MatrixOfImages>
  ImageGrid(const MatrixOfImages& ms, int borderwidth=0, int internalborderwidth=0,
	    const PixelType bg=default_bg)
    : AARImage<PixelType>(totalheight(ms,borderwidth,internalborderwidth),
			  totalwidth (ms,borderwidth,internalborderwidth),bg) {
	draw_border(borderwidth, bg);
	typedef typename MatrixOfImages::size_type Subscript;
	Subscript cursorr=borderwidth;
	
	for   (Subscript r=0; r<ms.nrows(); r++) {
	  /* Accommodate the maximum image size in this row */
	  typename MatrixOfImages::size_type stepheight = mat::max_nrows(mat::submatrix(ms,r,r+1,0,ms.ncols()));
	  Subscript cursorc=borderwidth;
	  
	  for (Subscript c=0; c<ms.ncols(); c++) {
	    /* Accommodate the maximum image size in this column */
	    typename MatrixOfImages::size_type stepwidth = mat::max_ncols(mat::submatrix(ms,0,ms.nrows(),c,c+1));
	    typename MatrixOfImages::value_type image = ms[r][c];
	    /* The image is centered in the cell available for it */
	    Subscript offr = (stepheight-image.nrows())/2; assert(int(offr)>=0);
	    Subscript offc = (stepwidth -image.ncols())/2; assert(int(offc)>=0);

	    /* Copy the pixels from the subimage into the larger image */
	    /* (There's probably a matrix routine to do this already.) */
	    for   (Subscript sr=0; sr<image.nrows(); sr++)
	      for (Subscript sc=0; sc<image.ncols(); sc++)
	        draw_pixel(cursorr+offr+sr, cursorc+offc+sc, image.get_pixel(sr,sc));
	    cursorc += stepwidth+internalborderwidth;
	  }
	  cursorr += stepheight+internalborderwidth;
	}
   }
};

 

/** Image composed of a rectangular grid of smaller images (passed as pointers)
    with an optional border between them. */
template<class PixelType=RGBPixel<>, class PixelMatrix=typename MatrixType<PixelType>::rectangular >
class ImagePointerGrid : public AARImage<PixelType,PixelMatrix> {

  /** Copied here explicitly to avoid apparent bug in GCC 2.95 */
  typedef typename AARImage<PixelType,PixelMatrix>::PixelSubscript PixelSubscript;

  /** Total height required for all rows */
  template <class MatrixOfImages>
  PixelSubscript totalheight(const MatrixOfImages& ms, int borderwidth=0, int internalborderwidth=0 ) {
    PixelSubscript t=borderwidth;
    for (typename MatrixOfImages::size_type r=0; r<ms.nrows(); r++)
      t += mat::max_pnrows(mat::submatrix(ms,r,r+1,0,ms.ncols()))+internalborderwidth;
    return t - (ms.nrows()>0?internalborderwidth : 0) + borderwidth;
  }

  /** Total width required for all columns */
  template <class MatrixOfImages>
  PixelSubscript totalwidth(const MatrixOfImages& ms, int borderwidth=0, int internalborderwidth=0 ) {
    PixelSubscript t=borderwidth;
    for (typename MatrixOfImages::size_type c=0; c<ms.ncols(); c++)
      t += mat::max_pncols(mat::submatrix(ms,0,ms.nrows(),c,c+1))+internalborderwidth;
    return t - (ms.ncols()>0?internalborderwidth : 0) + borderwidth;
  }

  
public:
  /** Constructor taking a matrix of arbitrary-sized images and concatenating them.
      Borders use the default_bg instead of the default border to give the effect
      of putting outlined images on a larger page, with the same border around and
      between the images. */
  template <class MatrixOfImages>
  ImagePointerGrid(const MatrixOfImages& ms, int borderwidth=0, int internalborderwidth=0,
		   const PixelType bg=default_bg)
    : AARImage<PixelType>(totalheight(ms,borderwidth,internalborderwidth),
			  totalwidth (ms,borderwidth,internalborderwidth),bg) {
	draw_border(borderwidth, bg);
	typedef typename MatrixOfImages::size_type Subscript;
	Subscript cursorr=borderwidth;
	
	for   (Subscript r=0; r<ms.nrows(); r++) {
	  /* Accommodate the maximum image size in this row */
	  typename MatrixOfImages::size_type stepheight = mat::max_pnrows(mat::submatrix(ms,r,r+1,0,ms.ncols()));
	  Subscript cursorc=borderwidth;
	  
	  for (Subscript c=0; c<ms.ncols(); c++) {
	    /* Accommodate the maximum image size in this column */
	    typename MatrixOfImages::size_type stepwidth = mat::max_pncols(mat::submatrix(ms,0,ms.nrows(),c,c+1));
	    typename MatrixOfImages::value_type p = ms[r][c];

	    if (p) {
	      /* The image is centered in the cell available for it */
	      Subscript offr = (stepheight-p->nrows())/2; assert(int(offr)>=0);
	      Subscript offc = (stepwidth -p->ncols())/2; assert(int(offc)>=0);
	      
	      /* Copy the pixels from the subimage into the larger image */
	      /* (There's probably a matrix routine to do this already.) */
	      for   (Subscript sr=0; sr<p->nrows(); sr++)
	      	for (Subscript sc=0; sc<p->ncols(); sc++)
	      	  draw_pixel(cursorr+offr+sr, cursorc+offc+sc, p->get_pixel(sr,sc));
	    }
	    cursorc += stepwidth+internalborderwidth;
	  }
	  cursorr += stepheight+internalborderwidth;
	}
   }
};

 
} /* namespace Plot */
#endif /* __IMAGEGRID_H__ */
