/** \file
  Endian-ness checking, swapping routines, and other functions for achieving
  cross-platform compatibility.
  
  $Header: /u/nn/cvsroot/lissom/src/ind_types.c,v 1.21 2003/01/11 06:50:43 jbednar Exp $  
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <ctype.h>

#include "ind_types.h"


/******************************************************************************/
/* Defines and typedefs                                                       */
/******************************************************************************/

typedef int endtype; /**< Type on which to test endian-ness */



/******************************************************************************/
/* Routines for achieving compatibility                                       */
/******************************************************************************/

/**
  Verify that the platform-independent types have appropriate values for
  this architecture
*/
#define CHECK_TYPE(typename,expectedsize)                    \
if (sizeof(typename) != expectedsize) {                       \
  printf("Warning -- problem with ind_types: sizeof(%3s)=%d, not %d.\n"   \
	 ,#typename,int(sizeof(typename)),int(expectedsize));   \
  errors++;                                                      \
}


#define WARN_MISSING(typename) if (warn_if_missing) printf("Note: no %s type defined.\n",#typename)


int check_ind_types( int warn_if_missing )
{
  int errors=0;
  (void)warn_if_missing;
    
#ifndef NO_I8
  CHECK_TYPE(i8,1);
#else
  WARN_MISSING(i8);
#endif
#ifndef NO_I16
  CHECK_TYPE(i16,2);
#else
  WARN_MISSING(i16);
#endif
#ifndef NO_I32
  CHECK_TYPE(i32,4);
#else
  WARN_MISSING(i32);
#endif
#ifndef NO_I64
  CHECK_TYPE(i64,8);
#else
  WARN_MISSING(i64);
#endif
#ifndef NO_F32
  CHECK_TYPE(f32,4);
#else
  WARN_MISSING(f32);
#endif
#ifndef NO_F64
  CHECK_TYPE(f64,8);
#else
  WARN_MISSING(f64);
#endif

  return errors;
}
#undef CHECK_TYPE
#undef WARN_MISSING


/**
  Returns 0 for little-endian, 1 for big-endian, and -1 for unknown.
  
  Many compilers (e.g. GCC) define a compile-time constant for endianness,
  but this routine allows it to be checked, and should work on most
  ordinary architectures
*/
int endianness(void)
{
  int little_so_far,big_so_far;
  unsigned i;
  
  union {
    endtype s;
    char c[sizeof(endtype)];
  } endu;

  /* Initialize integer */
  endu.s = 0;
  for (i=0; i < sizeof(endtype); i++) 
    endu.s += (i+1) << 8*i;
  
#ifdef DEBUG
  /* To test:  endu.s = swap_endian(endu.s); */

  printf ("Bytes ordered 0 to %d:",sizeof(endtype)-1);
  for (i=sizeof(endtype); i > 0; i--)
    printf(" %d",(int)endu.c[i-1]);
  printf("\n");
#endif

  /* Test for little endian */
  little_so_far = 1;
  for (i=0; i < sizeof(endtype); i++)
    if (little_so_far && ((unsigned)endu.c[i] != i+1))
      little_so_far = 0;
  
  if (little_so_far) return 0;
  
  /* Test for big endian */
  big_so_far = 1;
  for (i=0; i < sizeof(endtype); i++)
    if (big_so_far && ((unsigned)endu.c[i] != sizeof(endtype)-i))
      big_so_far = 0;
  
  if (big_so_far) return 1;
  
  return -1;
}



/** Family of procedures to swap endian-ness for values of different types. */
#define swap_endian_proc(type)            \
type type ## _swap_endian(type num)       \
{					  \
  unsigned i;				  \
  union {				  \
    type num;				  \
    char c[sizeof(type)];		  \
  } in,out;				  \
					  \
  in.num = num;				  \
  for (i=0; i < sizeof(type); i++)	  \
    out.c[i] = in.c[sizeof(type)-1-i];	  \
  return out.num;                         \
}
swap_endian_proc(i32)
#ifndef NO_I64
swap_endian_proc(i64)
#endif


     
/******************************************************************************/
/* Substitutes for system routines not required by ANSI C                     */
/******************************************************************************/


#if (defined(NO_SNPRINTF_AVAIL) || defined(NO_TEMPNAM_AVAIL))
#include <stdarg.h>
/**
  Provides a fake version of snprintf for systems without it.
  This is not a full implementation, since it doesn't enforce
  the bound, but since it at least checks it that's probably better
  than nothing.
*/
EXTERNAL_C_LINKAGE int my_snprintf( char *str, size_t n, const char *format, ...)
{
  size_t chars;
  va_list args;
  va_start (args, format);
  chars=vsprintf (str, format, args);
  va_end (args);
  if (chars+1>n) {
    printf("Warning -- Data may be overwritten due to missing snprintf() function.\n");
    return -1;
  }
  return chars;
}
#endif

#ifdef NO_TEMPNAM_AVAIL
/* Provides a fake version of tempnam for systems without it.
   Uses the current directory for all files, because it does not
   do any checking to see if a file currently exists by that name.
   Should still work as long as no more than a single instance of
   this executable is running in the current directory, which
   is reasonable because such a program would overwrite other
   files with fixed names as well.  Cannot use global directories
   like /tmp, because multiple concurrent executables would then
   overwrite each other's files.  
*/
EXTERNAL_C_LINKAGE char *my_tempnam(const char *dir, const char *pfx)
{
  (void)dir; /* Currently ignored */
  static int timescalled=0;
  const char* default_prefix="temp_";
  char* str = (char *)malloc(32);
  my_snprintf(str,32,"%5s%06d",(pfx? pfx : default_prefix),timescalled);
  
  timescalled++;
  return str;
}
#endif

