/** \file
  Define platform-independent types and other platform-specific support.
  
  This file can be used on its own (i.e. without ind_types.c) if none
  of the actual functions defined in ind_types.c are used.

  $Header: /u/nn/cvsroot/lissom/src/ind_types.h,v 1.36 2001/12/01 07:33:41 jbednar Exp $
*/

#ifndef __IND_TYPES_H__
#define __IND_TYPES_H__

#include <limits.h>
#include <float.h>


/******************************************************************************/
/* General-purpose definitions                                                */
/******************************************************************************/

#ifndef NO_ERROR
#define NO_ERROR        0
#endif

#ifndef MIN
#define MIN(x,y)        (((x) < (y)) ? (x) : (y))
#endif
#ifndef MAX
#define MAX(x,y)        (((x) > (y)) ? (x) : (y))
#endif

#ifndef INT_MAX
#error "Warning: INT_MAX undefined; default used"
#define INT_MAX 2147483647
#endif

#ifndef INT_MIN
#error "Warning: INT_MIN undefined; default used"
#define INT_MIN -2147483647
#endif



/******************************************************************************/
/* C/C++ compatibility                                                        */
/******************************************************************************/

/**
  Put this before any external C functions called so that they will work
  with both C and C++.
*/

#ifdef __cplusplus
#define EXTERNAL_C_LINKAGE extern "C"
#else 
#define EXTERNAL_C_LINKAGE
#endif



/******************************************************************************/
/* Platform-independent types                                                 */
/******************************************************************************/


/* Cray MPP machines T3D and T3E */
#if     defined(_CRAYMPP)  
  typedef char      i8;
  typedef short     i32;
  typedef int       i64;
  typedef float     f32;
  typedef double    f64;
#define NO_I16  /* Not supported on Cray MPP */


/* Add appropriate definitions for other architectures here */


/* Generic 32-bit platform with GCC compiler */
#elif (defined(__GNUC__) && !defined(ANSI_COMPATIBLE))
  typedef char      i8;
  typedef short     i16;    
  typedef long      i32;
#ifndef LONG_LONG_UNAVAILABLE
  typedef long long i64;
#endif
  typedef float     f32;
  typedef double    f64;

/* Assume generic 32-bit platform with no 64-bit type */
#else   
  typedef char      i8;
  typedef short     i16;    
  typedef long      i32;
  typedef float     f32;
  typedef double    f64;
#define NO_I64  /* Not supported in ANSI C */

#endif



/******************************************************************************/
/* Strict ANSI compatibility and/or better cross-platform support             */
/******************************************************************************/

#ifdef ANSI_COMPATIBLE
#define NO_RINT_AVAILABLE
#define NO_TEMPNAM_AVAIL
#define	NO_SNPRINTF_AVAIL
#define NO_STRNCASECMP_AVAIL
#define	NO_HYPOT_AVAIL
#define	NO_RAND48_AVAIL
#define NO_ASSERT_FUNC_INFO_AVAIL
#define NO_SIGNAL_HANDLING
#define	LONG_LONG_UNAVAILABLE
#define LONG_DOUBLE_UNAVAILABLE
//#define EXPLICIT_INSTANTIATION_FOR_STATICS
#endif

#ifndef ROUND
#define ROUND(a)        (floor((0.5+(a))))
#endif


#ifdef NO_TEMPNAM_AVAIL
EXTERNAL_C_LINKAGE char *my_tempnam(const char *dir, const char *pfx);
#define TEMPNAME my_tempnam
#else
#define TEMPNAME tempnam
#endif

#ifdef NO_SNPRINTF_AVAIL
EXTERNAL_C_LINKAGE int   my_snprintf(char *str, size_t n, const char *format, ...);
#define SNPRINTF my_snprintf
#else
#define SNPRINTF snprintf
#endif


/******************************************************************************/
/* Public functions for achieving compatibility                               */
/******************************************************************************/

int check_ind_types( int warn_if_missing );
int endianness( void );

#define swap_endian_proto(type) type type ## _swap_endian(type num)
swap_endian_proto(i32);
/* i32 i32_swap_endian(i32 num); */
#ifndef NO_I64
swap_endian_proto(i64);
/* i64 i64_swap_endian(i64 num); */
#endif

#endif /* __IND_TYPES_H */

