/** \file
  index_iterator class.

  $Header: /u/nn/cvsroot/lissom/src/index_iterator.h,v 1.2 2001/04/01 03:20:15 jbednar Exp $
*/

#ifndef __INDEX_ITERATOR_H__
#define __INDEX_ITERATOR_H__


#include <iterator>
#include <functional>
#include "../src/scale.h"

namespace Generic {


/** Simple iterator adapter class providing element indexes instead of
    elements.

    Given an input_iterator, creates an input_iterator that when
    dereferenced returns the number of times it has been incremented
    so far, rather than the underlying value.  (The underlying value
    may still be accessed through the value() operation.)  This class
    is useful for treating a container as if it held indexes rather
    than data values, for use with standard algorithms.  Example:

    <pre>
      typedef std::vector<int> V;
      typedef index_iterator<V> iiv;
      V v;
      //...
      
      // List the elements of v, in order, followed by their indexes, in order:
      std::copy(    v.begin(),     v.end(), std::ostream_iterator<int>(std::cout," ")); std::cout << "\n"
      std::copy(iiv(v.begin()),iiv(v.end()),std::ostream_iterator<int>(std::cout," ")); std::cout << "\n"
    </pre>

    If desired a function of the specified TransformFnType can be supplied;
    it will be applied to the index before returning it, e.g. to scale it
    to a desired range.
*/
template <class C, class TransformFnType=Scale::Identity<typename C::value_type> >
class index_iterator {
  typedef index_iterator self;
  typedef C Container;
  
  typedef typename Container::iterator        iterator;
  typedef typename Container::const_iterator  const_iterator;
  typedef typename Container::size_type       size_type;

public:
  /** \name Types required for iterators */
  //@{
  typedef std::input_iterator_tag             iterator_category;
  typedef typename Container::value_type      value_type;
  typedef typename Container::difference_type difference_type;
  typedef typename Container::value_type*     pointer;
  typedef typename Container::value_type      reference;
  //@}
  
  /** Constructor */
  index_iterator(const const_iterator& i_, TransformFnType fn_=TransformFnType())
    : i(i_), c(0), fn(fn_) {}

  /** \name Operations required for input_iterators */
  //@{
  bool       operator==(const self& o) {  return (o.i == i);       }
  bool       operator!=(const self& o) {  return (o.i != i);       }
  self&      operator++ ()             {  ++i; ++c;  return *this; }
  self&      operator++ (int)          {  i++; c++;  return *this; }
  //pointer  operator-> ()             {  return &c; } // Not meaningful for an index
  reference  operator*  ()             {  return fn(c);  }
  //@}
  
  /** Returns the value() of the underlying iterator */
  const value_type& value() const { return *i;  }
  
private:
  const_iterator i;
  size_type c;

  TransformFnType fn;
};



} /* namespace Generic */
#endif /* __INDEX_ITERATOR_H__ */
