/** \file
  Public headers for ipc.c.
  $Header: /u/nn/cvsroot/lissom/src/ipc.h,v 1.31 2004/05/13 23:59:14 jbednar Exp $
*/

#ifndef ___IPC_H___
#define ___IPC_H___

#include <stdlib.h>
/* platform-independent types i32, etc. */
#include "ind_types.h"


/* Defines and typedefs */

/* Assume long long type is available except under a few architectures */
#ifdef _CRAYMPP
#ifdef __cplusplus
#define LONG_LONG_UNAVAILABLE
#endif
#endif

/* Datatypes for copy operations */
/* Based on MPI datatypes, with C equivalent shown as comments */
typedef unsigned int ipc_datatype;
typedef int ipc_status;

#define IPC_RAW8 		((ipc_datatype)1)  /**<	any  8-bit type  */
#define IPC_RAW32 		((ipc_datatype)2)  /**<	any 32-bit type  */
#define IPC_RAW64 		((ipc_datatype)3)  /**<	any 64-bit type  */
#define IPC_CHAR 		((ipc_datatype)4)  /**<	signed char	 */
#define IPC_SHORT 		((ipc_datatype)5)  /**<	signed short     */
#define IPC_INT 		((ipc_datatype)6)  /**<	signed int	 */
#define IPC_LONG 		((ipc_datatype)7)  /**<	signed long 	 */
#define IPC_LONG_LONG 		((ipc_datatype)8)  /**<	signed long long */
#define IPC_UNSIGNED_CHAR 	((ipc_datatype)9)  /**<	unsigned char    */
#define IPC_UNSIGNED_SHORT 	((ipc_datatype)10) /**<	unsigned short   */
#define IPC_UNSIGNED 		((ipc_datatype)11) /**<	unsigned int	 */
#define IPC_UNSIGNED_LONG	((ipc_datatype)12) /**<	unsigned long    */
#define IPC_FLOAT 		((ipc_datatype)13) /**<	float	         */
#define IPC_DOUBLE 		((ipc_datatype)14) /**<	double	         */
#define IPC_LONG_DOUBLE 	((ipc_datatype)15) /**<	long double      */


#define IPC_NO_ERROR 0

#define IPC_EXIT_NORMAL                	0
#define IPC_EXIT_WRONG_USAGE           	1
#define IPC_EXIT_FILE_PROBLEM	        2
#define IPC_EXIT_PARAM_VALUE_ERROR 	3
#define IPC_EXIT_OUT_OF_MEMORY     	4
#define IPC_EXIT_TOO_MANY_ERRORS     	5

/*
  Values for debug_level in ipc_notify() and ipc_msg_level
*/
#define IPC_NONE     -10
#define IPC_ERROR     -6
#define IPC_WARNING   -4
#define IPC_CAUTION   -2
#define IPC_ALERT      0
#define IPC_REQUESTED  1
#define IPC_SUMMARY    2
#define IPC_STD        4
#define IPC_VERBOSE    6
#define IPC_OVERWHELM 10

/*
  Values for print_pe on messages
  IPC_ONE defaults to parent PE
*/
#define IPC_ONE  0
#define IPC_ALL -1

/* Parameters that can be changed by external programs */
extern int ipc_msg_level;
extern int ipc_exit_on_error_num;
extern int ipc_max_errors;  
extern int ipc_max_warnings;
extern int ipc_msg_forceall_level;
extern int ipc_msg_synch_level;

/* Documentation for parameters */
extern const char* ipc_msg_level_docstring;
extern const char* ipc_exit_on_error_num_docstring;
extern const char* ipc_max_errors_docstring;
extern const char* ipc_max_warnings_docstring;
extern const char* ipc_msg_forceall_level_docstring;
extern const char* ipc_msg_synch_level_docstring;

/* General interprocess communication routines */
int        ipc_datatype_size(ipc_datatype datatype);
int        ipc_num_processes(void);
int        ipc_my_process(void); 
void       ipc_barrier(void); 
void       ipc_set_barrier(void);

/*  User interface for copying data to and from a remote process */

/* C user interface -- no inherent type-checking */
#ifndef __cplusplus
ipc_status ipc_put(void *data, ipc_datatype datatype, size_t count, int process);
ipc_status ipc_get(void *data, ipc_datatype datatype, size_t count, int process);
ipc_status ipc_put_to(void *target_data, void *source_data, ipc_datatype datatype, size_t count, int process);
ipc_status ipc_get_to(void *target_data, void *source_data, ipc_datatype datatype, size_t count, int process);


/* C++ user interface -- two overloaded versions per ipc_datatype:
   one C-compatible version allowing weak form of runtime 
   type-checking, the other fully type-safe but incompatible. */
#else
#define IPC_CALL_PROTO(name, type,ipc_type) \
ipc_status ipc_ ## name         (type *data,                                        size_t count, int process); \
ipc_status ipc_ ## name         (type *data,                 ipc_datatype datatype, size_t count, int process); \
ipc_status ipc_ ## name ## _to  (type *target, type *source,                        size_t count, int process); \
ipc_status ipc_ ## name ## _to  (type *target, type *source, ipc_datatype datatype, size_t count, int process) 

/* IPC_CALL_PROTO(put, signed char      ,IPC_CHAR            );  */
IPC_CALL_PROTO(put, signed short     ,IPC_SHORT           ); 
IPC_CALL_PROTO(put, signed int       ,IPC_INT             ); 
IPC_CALL_PROTO(put, signed long      ,IPC_LONG            ); 
#ifndef LONG_LONG_UNAVAILABLE
IPC_CALL_PROTO(put, signed long long ,IPC_LONG_LONG       );
#endif
IPC_CALL_PROTO(put, unsigned char    ,IPC_UNSIGNED_CHAR   ); 
IPC_CALL_PROTO(put, unsigned short   ,IPC_UNSIGNED_SHORT  ); 
IPC_CALL_PROTO(put, unsigned int     ,IPC_UNSIGNED        ); 
IPC_CALL_PROTO(put, unsigned long    ,IPC_UNSIGNED_LONG   ); 
IPC_CALL_PROTO(put, float            ,IPC_FLOAT           ); 
IPC_CALL_PROTO(put, double           ,IPC_DOUBLE          ); 
#ifndef LONG_DOUBLE_UNAVAILABLE
IPC_CALL_PROTO(put, long double      ,IPC_LONG_DOUBLE     ); 
#endif

/* IPC_CALL_PROTO(get, signed char      ,IPC_CHAR         );  */
IPC_CALL_PROTO(get, signed short     ,IPC_SHORT           ); 
IPC_CALL_PROTO(get, signed int       ,IPC_INT             ); 
IPC_CALL_PROTO(get, signed long      ,IPC_LONG            ); 
#ifndef LONG_LONG_UNAVAILABLE
IPC_CALL_PROTO(get, signed long long ,IPC_LONG_LONG       );
#endif
IPC_CALL_PROTO(get, unsigned char    ,IPC_UNSIGNED_CHAR   ); 
IPC_CALL_PROTO(get, unsigned short   ,IPC_UNSIGNED_SHORT  ); 
IPC_CALL_PROTO(get, unsigned int     ,IPC_UNSIGNED        ); 
IPC_CALL_PROTO(get, unsigned long    ,IPC_UNSIGNED_LONG   ); 
IPC_CALL_PROTO(get, float            ,IPC_FLOAT           ); 
IPC_CALL_PROTO(get, double           ,IPC_DOUBLE          ); 
#ifndef LONG_DOUBLE_UNAVAILABLE
IPC_CALL_PROTO(get, long double      ,IPC_LONG_DOUBLE     ); 
#endif
#undef IPC_CALL_PROTO

#endif


/**
  Raw interface for use when absolute lengths of data are 
  known when code is written. (Primarily for code using 
  device-independent types in ind_type.h.)
*/
#define IPC_PUT_RAW_PROTO(bits) \
ipc_status ipc_put ## bits(void *data, size_t count, int process);
IPC_PUT_RAW_PROTO(8)
IPC_PUT_RAW_PROTO(32)
IPC_PUT_RAW_PROTO(64)
#undef IPC_PUT_RAW_PROTO

#define IPC_GET_RAW_PROTO(bits) \
ipc_status ipc_get ## bits(void *data, size_t count, int process);
IPC_GET_RAW_PROTO(8)
IPC_GET_RAW_PROTO(32)
IPC_GET_RAW_PROTO(64)
#undef IPC_GET_RAW_PROTO

/* Message, warning, and exit routines */
void ipc_pe_msg_delay( double scale );
void ipc_init(void);
void ipc_init_logfile(const char *basefilename);
     
void ipc_log(    int print_pe,                    const char *format, ...);
int  ipc_notify( int print_pe, int message_level, const char *format, ...);
void ipc_notify2(int terminateline, int print_pe, int message_level, const char *format, ...);
void ipc_exit(   int status,                      const char *format, ...);
void ipc_abort(  int status,                      const char *format, ...);


/*
  Functions provided as convenient markers for a debugger
  to break when an error or warning occurs; not intended
  to be called directly
*/
void ipc_error(   void );
void ipc_warning( void );


#ifdef __cplusplus
#ifndef NO_STRINGS
#include <string>
using std::string;

namespace Msg {
/**
  Message handlers implemented using ipc_ calls for multiprocessor
  safety and message counting.  These can be passed in to functions or
  templates with the same parameter list but which are not based on
  ipc_ calls, in order to give users a choice between ipc_ and
  non-ipc_ implementations of messaging.
*/
//@{
void ipc_overwhelm (const string& s, const bool terminate=true);
void ipc_verbose   (const string& s, const bool terminate=true);
void ipc_notify    (const string& s, const bool terminate=true);
void ipc_requested (const string& s, const bool terminate=true);
void ipc_error     (const string& s, const bool terminate=true);
void ipc_warning   (const string& s, const bool terminate=true);
//@}
}
#endif
#endif

#endif

     
