/** \file
  Code to generate two-dimensional matrices of function values as kernels for convolution.

  $Header: /u/nn/cvsroot/lissom/src/kernelfactory.h,v 1.9 2003/03/29 21:45:17 jbednar Exp $
*/

#ifndef __KERNEL_FACTORY_H__
#define __KERNEL_FACTORY_H__

#include <string>
using std::string;
#include <functional>

#include "stringparser.h"
#include "pnm_io.h"
#include "radialfunction.h"
#include "neuralregion.h"
#include "generic_stdlib.h"


/** Generates two-dimensional matrices of function values as kernels for convolution. */
class KernelFactory {
public:
  /** Type to use for a generated kernel */
  typedef InternalNeuralRegion::WeightMatrix  KernelMatrix;

  KernelFactory() : max_blur_type(Blur_None) { }
  virtual ~KernelFactory() { }

  KernelMatrix create( int index, double world_size_scale, StringArgs arglist );

  /** Concrete type to use for a kernel function pointer */
  typedef double (*kernel_fnPtr)  (double x_sq, double sig_sq);

  /** Convenient wrapper for RadialFunction's version */
  inline KernelMatrix create(kernel_fnPtr fnptr, double radius,
			     double max_radius, string& errors,
			     bool circular=true, bool normalize=true) {
    return RadialFunction::matrix<KernelMatrix>
      (fnptr,radius,max_radius,errors,circular,normalize);
  }

  /** Call before using any of the parameters and commands in this class */
  void register_params_and_commands( void );

  /** Possible values for a blur_type. The last size should always be Blur_None; new ones
      should be added just before it. */
  enum BlurTypes {Blur_SquareAverage,Blur_CircularAverage,Blur_Gaussian,Blur_DoG,Blur_LoG,
		  Blur_GoD,Blur_DoGGoD,Blur_PGM,Blur_SquareRandom,Blur_CircularRandom,Blur_SoG,Blur_Cauchy,Blur_None};
  
  /** Highest index defined in BlurTypes. */
  const BlurTypes max_blur_type;

private:
  /* Static defaults */
  static double blur_radius;
  static double blur_radius_surround_multiplier;
  static double blur_scale;
  static double blur_range_multiplier;
  static int    blur_type;
  static char   default_kernel_filename[];
};


/** Wraps a KernelFactory call into a function object taking a size_scale parameter. */
class KernelFactoryWrapper : public Generic::unary_virtual_function<double,KernelFactory::KernelMatrix> {
public:
  KernelFactoryWrapper(KernelFactory factory, StringArgs args, int num=0)
    : e(num), arglist(args), kf(factory)  { }
  /** Call our factory's create() with the given size_scale. */
  virtual KernelFactory::KernelMatrix operator() ( double size_scale )
    {  return kf.create(e,size_scale,arglist);  }
private:
  int e;
  StringArgs arglist;
  KernelFactory& kf;
};

#endif /* __KERNEL_FACTORY_H__ */
