/** \file
  LoopSpec class and associated functions.

  $Header: /u/nn/cvsroot/lissom/src/loopspec.h,v 1.4 2002/07/09 03:16:59 jbednar Exp $
*/

#ifndef __LOOPSPEC_H__
#define __LOOPSPEC_H__

#include <string>
using std::string;
#include <iostream>

#include "../src/valuegen.h"



/** Loop counter specification.  Useful for implementing loops with
    arbitrary run-time modifiable nesting.

    Could be made to inherit from ValueGen if the overhead of a
    virtual function call will always be considered small inside the
    innermost loop; this happens to be true at present but might not
    remain so.
*/
template<class T>
class LoopSpec {
public:
  typedef LoopSpec self;
  typedef T value_type;
  
  /** The max_count determines how many times next() can be called
      before valid() becomes false.

      If a non-empty name string is provided, it will be used for
      stream output.

      If a ValueGenerator is supplied, the value() of the
      ValueGenerator is returned for the value() of this object (after
      modification by discrete_levels, if any, below).  The
      ValueGenerator is advanced once for every reset (to ensure that
      a valid value exists for e.g. random distributions) and again
      once per next().  The ValueGenerator should returna number
      in the range 0 to 1.0.

      Usually, however, no ValueGenerator is needed.  If none is
      supplied, the value() will be a number from 0 to 1.0 reflecting
      the number of times next() has been called so far out of the
      total count specified.  If the total count is zero, the value()
      will be 0.

      If discrete_inputs is present and greater than zero, the value()
      will be discretized to the given number of levels before it is
      returned, which can be useful with Random valuegenerators,
      e.g. to discretize to max_count levels.  The discretization is
      done by multiplying by discrete_levels, truncating to an
      integer, and then dividing by discrete_levels; this procedure is
      meaningful for real variables 0..1.0 but may not be useful for
      other ranges of values.  To avoid any spurious rounding errors,
      discretization is skipped even if specified, if discrete_levels
      happens to be the same as max_count and there is no
      ValueGenerator.

      If exclusive is true, then the max_count is taken to be
      exclusive, and valid() counts will range from 0 to max_count-1,
      inclusively.  Otherwise valid counts will range from 0 to
      max_count, inclusively.  The conversion to a continuous value()
      does not depend on the exclusivity or inclusivity.
  */
  LoopSpec(const string& name_, const int max_count_, const bool exclusive=true, ValueGenerator<T>* g=0,
	   const int discrete_levels_=-1)
    : namestr(name_), max_count(max_count_),
      end(exclusive? max_count : max_count+1), valuegen(g),
      /* Skips discretization if it will be a no-op. */
      discrete_levels((discrete_levels_==max_count && g==0)? -1 : discrete_levels_),
      counter(0) { }
  
  inline bool   next()        {  if (valuegen) valuegen->next(); return (counter++ < end);  }
  inline void   reset()       {  if (valuegen) {  valuegen->reset(); valuegen->next();  }  counter=0;  }

  inline bool   valid() const {  return (counter < end);  }
  inline int    count() const {  return counter;  }
  inline int    num()   const {  return end;      }
  inline T  operator*() const {  return value();  }

  inline string name()  const {  return namestr;  }

  inline T      value() const {
    const T val = (!valuegen? (max_count? T(counter)/T(max_count) : 0) : valuegen->value());
    return (discrete_levels <=0 ? val : T(int(val*discrete_levels))/T(discrete_levels));
  }

private:
  const string namestr;
  const int max_count;
  const int end;
  OwningPointer<ValueGenerator<T> > valuegen;
  const int discrete_levels;

  int counter;
};



/** Stream output for a LoopSpec.  If a name is defined, the value is
    shown as x=y; otherwise just the value (y) is shown.  */
template<class TT>
std::ostream& operator<<(std::ostream& s, const LoopSpec<TT>& ls)
{  s << (ls.name()==""? string() : (ls.name() + "=")) << ls.value();  return s;  }



/**
   Accepts a container of LoopSpecs and executes them recursively as
   a set of nested loops around the given Action, which must be 
   callable as a function.
   
   Actually, the loops need not be LoopSpecs, per se, as long as they
   support next(), reset(), and valid().
*/
template<class LoopContainer, class LoopContainerIterator, class Action>
bool LoopSpecs_ExecuteNested(LoopContainer& loops, const LoopContainerIterator& current, Action& fn)
{
  if (current == loops.end()) return false;
  else {
    (*current)->reset();
    while ((*current)->valid()) {
      if (!LoopSpecs_ExecuteNested(loops,current+1,fn))
	fn();
      (*current)->next();
    }
  }
  return true;
}


#endif /* __LOOPSPEC_H__ */

