/** \file
  MatrixAdapter<> class.
  
  $Header: /u/nn/cvsroot/lissom/src/matrixadapter.h,v 1.9 2001/12/13 09:04:26 jbednar Exp $
*/

#ifndef __MATRIXADAPTER_H__
#define __MATRIXADAPTER_H__

#include "matrix.h"
#include "generic_stdlib.h"
#include "allocatingpointer.h"
#include "scale.h"
#include "aarectangle.h"

namespace mat {


/**
  Adapter that allows you to view an existing matrix as another matrix
  with arbitrary linear value scaling, width and height extension or
  reduction, spatial offsetting, etc.\ as needed.  Increases and
  decreases in size are not anti-aliased or smoothed, so the result
  may be blocky.  All computation is done as the elements are
  accessed, and no temporary matrix is used.

  One use for this adapter is for plotting a matrix whose values are
  in an inconvenient range, which you want to situate in a larger or
  smaller area, or of which you want only a portion.

  The value_type of the adapter can be a different type from that of
  the underlying matrix as long as a suitable conversion exists
  (e.g. it is valid to have a 'double' adapter for an 'int'
  matrix).  To achieve this, specify the desired value_type as the
  T template parameter.
*/
template <class M,
	  class T=typename M::value_type,
	  class ScaleT=Scale::Linear<T>,
	  class RegionT=Boundary::AAArbitraryRectangle<double,int> >
class MatrixAdapter {
  typedef MatrixAdapter<M,T,ScaleT,RegionT> self;
public:
  
  typedef typename M::size_type  size_type;
  typedef T                      value_type;
  typedef ScaleT                 ScaleType;
  typedef RegionT                Region;
  typedef int                    Coordinate;
  
  /**
     Typical constructor; all arguments optional except the
     underlying matrix.
     
     Linear value scaling is provided as val=orig*a+b.  The height and
     width default to the natural size of the matrix; they can be
     overridden to any nonzero value, and will be taken relative to
     the original size of the matrix, regardless of any subsequent
     scaling.  (Specified height and width must be nonzero due to a
     limitation of GCC 2.95.2 and earlier, which do not allow member
     functions in default initializers.)  Values outside the native
     size of the array are assumed to be zero, and thus the offset b
     is returned.  The underlying array may be placed at an arbitrary
     origin (or,oc) within the full region, and the full region is
     then scaled in size by sr and sc.

     This constructor takes a reference to an external object with its
     own lifetime; the lifetime must be larger than that of this
     object.
     
     It is possible that a and b should be made 'double' here instead
     of value_type, so that floating-point scales may be supplied for
     integer matrices.
  */
  
  /** Constructor for a reference to an external matrix with its own
      lifetime that is larger than that of this object. */
  inline MatrixAdapter(const M&         x,
		       const ScaleType& vscale_i=ScaleType(),
		       const Region&    region=Region(),
		       const double&    size_multiplier=1.0)
    : m(x), vscale(vscale_i),
      height(size_type(size_multiplier*(!region.isinfinite()? region.height() : x.nrows()))),
      width( size_type(size_multiplier*(!region.isinfinite()? region.width()  : x.ncols()))),
      originr(Coordinate(region.corners().yl)), originc(Coordinate(region.corners().xl)),
      size_divisor(size_multiplier? 1.0/size_multiplier : 0)
    {  assert(region.corners().xh<=region.corners().xl); assert(region.corners().yh<=region.corners().yl); }
  
  /** Constructor to a newly-allocated object which will be owned by this
      object and deleted when this object is destroyed. */
  inline MatrixAdapter(M*               x,
		       const ScaleType& vscale_i=ScaleType(), 
		       const Region&    region=Region(),
		       const double&    size_multiplier=1.0)
    : m(*x), allocated(x), vscale(vscale_i),
      height(size_type(size_multiplier*(!region.isinfinite()? region.height() : x->nrows()))),
      width( size_type(size_multiplier*(!region.isinfinite()? region.width()  : x->ncols()))),
      originr(Coordinate(region.corners().yl)), originc(Coordinate(region.corners().xl)),
      size_divisor(size_multiplier? 1.0/size_multiplier : 0)
    {  assert(region.corners().xh<=region.corners().xl); assert(region.corners().yh<=region.corners().yl); }
  
  /** Element access.  Note that the semantics of this call depend on
      the definitions in matrix.h.  That is, if matrix.h is set up to
      use a 1-based offset for operator(), so will this function
      (via the MatIndex() call), and otherwise it will used a 0-based
      offset.  That way any MatrixAdapter can be used like a regular
      matrix defined in matrix.h.
  */
  inline value_type operator()(const size_type row, const size_type col) const {
    const size_type  lower_bound=MatIndex(0);
    const value_type default_value=0;
    const Coordinate r=row-lower_bound;
    const Coordinate c=col-lower_bound;
    /* The 0.000001 is to compensate for floating point truncation
       errors for some common size scales, e.g. multiples of 3. */
    const Coordinate sr=Coordinate(floor((r*size_divisor)+0.000001))-originr;
    const Coordinate sc=Coordinate(floor((c*size_divisor)+0.000001))-originc;

    return value_type(vscale(inbounds(sr,sc)? mat::elem(m,sr,sc): default_value));
  }
    
  
  /** Number of rows in the virtual matrix */
  inline size_type nrows() const { return height; }

  /** Number of columns in the virtual matrix */
  inline size_type ncols() const { return width; }
  
protected:
  const M& m;
  MemoryReference<M> allocated;

  ScaleT vscale;
  size_type  height,width;
  Coordinate originr,originc;
  double size_divisor;

  /** Returns true if the given coordinates are within the matrix  */
  inline bool inbounds(const size_type r, const size_type c) const {
    return (int(r)>=0 && r<m.nrows() &&
	    int(c)>=0 && c<m.ncols());
  }
};

 
/** Wraps up a C array as a no-frills matrix; wrap this with a MatrixAdapter for more features. */
template <class T, class Array>
class CArrayMatrixAdapter {
  typedef CArrayMatrixAdapter<T,Array> self;
public:
  
  typedef size_t size_type;
  typedef T value_type;
  
  inline CArrayMatrixAdapter(const Array& x, const size_type nrows, const size_type ncols)
    : a(x), height(nrows), width(ncols) { }

  inline value_type operator()(const size_type row, const size_type col) const {
    const size_type  lower_bound=MatIndex(0);
    assert((row-lower_bound)<height);
    assert((col-lower_bound)<width);
    //return a[row-lower_bound][col-lower_bound];
    return a[col-lower_bound][row-lower_bound];  // Temporary: Swaps to match flipped conventions
  }
  
  /** Number of rows in the virtual matrix */
  inline size_type nrows() const { return height; }

  /** Number of columns in the virtual matrix */
  inline size_type ncols() const { return width; }
  
protected:
  const Array& a;
  const size_type height, width;
};

 

/** Abstract base class providing a simple unified virtual-function
    interface to matrices of various types.

  Normally, functions are written to handle different matrix types by
  accepting a template parameter for the matrix type.  Sometimes,
  however, e.g. when keeping a list of matrices of different types,
  one needs to have run-time polymorphism instead of just compile-time
  polymorphism.  If such a list uses this base class as its element
  type, any compatible matrix may be wrapped with one of the classes
  below so that it can be stored on the list.  The main requirement is
  that the value_type of each matrix be the same.

  Of course, since this interface relies on virtual functions, it will
  be somewhat slower than accepting a template parameter.  
*/
template <class T>
class MatrixInterface {
  typedef MatrixInterface<T> self;
public:
  
  typedef int size_type;
  typedef T   value_type;
  
  virtual ~MatrixInterface() { };
 
  virtual value_type operator()(const size_type row, const size_type col) const =0;
  
  /** Number of rows in the virtual matrix */
  virtual size_type nrows() const =0;

  /** Number of columns in the virtual matrix */
  virtual size_type ncols() const =0;
};

 
  
/** Wraps up a MatrixAdapter<> as a MatrixInterface<>, without
    introducing any virtual function overhead into the actual
    MatrixAdapter<> class.

  This class is intended to duplicate the semantics of
  MatrixAdapter<>, so see the documentation for that class for more
  information.
*/
template <class M,
	  class T=typename M::value_type,
	  class ScaleT=Scale::Linear<T>,
	  class RegionT=Boundary::AAArbitraryRectangle<double,int> >
class MatrixInterfaceAdapter : public MatrixInterface<T> {
  typedef MatrixInterfaceAdapter<M,T,ScaleT,RegionT> self;

public:  
  typedef typename MatrixInterface<T>::size_type  size_type;
  typedef typename MatrixInterface<T>::value_type value_type;
  typedef ScaleT   ScaleType;
  typedef RegionT  Region;

  /** Constructor for a reference to an external matrix with its own
      lifetime that is larger than that of this object. */
  inline MatrixInterfaceAdapter(const M&         x,
				const ScaleType& vscale=ScaleType(), 
				const Region&    region=Region(),
				const double&    size_multiplier=1.0)
    : ma(x,vscale,region,size_multiplier) { }
  
  /** Constructor for a newly-allocated matrix which will be owned by
      this object and deleted when this object is destroyed. */
  inline MatrixInterfaceAdapter(M*               x,
				const ScaleType& vscale=ScaleType(), 
				const Region&    region=Region(),
				const double&    size_multiplier=1.0)
    : ma(x,vscale,region,size_multiplier) { }
  
  inline value_type operator()(const size_type row, const size_type col) const 
    {  return value_type(ma(typename M::size_type(row),
			    typename M::size_type(col)));  }
  
  /** Number of rows in the virtual matrix */
  inline size_type nrows() const { return ma.nrows(); }

  /** Number of columns in the virtual matrix */
  inline size_type ncols() const { return ma.ncols(); }
  
protected:
  MatrixAdapter<M,value_type> ma;
};



/** Convenience class for packaging a constant scalar value as a matrix
    of arbitrary size. */
template <class T=double>
class MatrixInterfaceConstant : public MatrixInterface<T> {
  typedef MatrixInterfaceConstant<T> self;

public:
  typedef typename MatrixInterface<T>::size_type  size_type;
  typedef typename MatrixInterface<T>::value_type value_type;
  
  /** Typical constructor. */
  inline MatrixInterfaceConstant(const value_type& a=1.0, size_type nr=1, size_type nc=1)
    : val(a), height(nr), width(nc) { }
  
  inline value_type operator()(const size_type, const size_type) const 
    {  return val;  }
  
  /** Number of rows in the virtual matrix */
  inline size_type nrows() const { return height; }

  /** Number of columns in the virtual matrix */
  inline size_type ncols() const { return width; }
  
protected:
  const T val;
  const size_type height,width;
};


/** Adapter that presents as a matrix the result of a scalar
  two-argument operation on the elements of two identically-sized
  matrices.  It could e.g. add them, take the maximum, multiply them
  element-wise, etc., depending on the supplied BinaryFunction.
*/
template <class M, class BinaryFunction, class T=typename M::value_type >
class MatrixInterfaceBinaryAdapter : public MatrixInterface<T> {
  typedef MatrixInterfaceBinaryAdapter<M,T> self;
  
public:
  typedef typename MatrixInterface<T>::size_type  size_type;
  typedef typename MatrixInterface<T>::value_type value_type;

  /** Typical constructor. */
  inline MatrixInterfaceBinaryAdapter(const M& A, const M& B, const BinaryFunction& fn)
    : MA(A), MB(B), op(fn) {
      assert(MA.nrows()==MB.nrows());
      assert(MA.ncols()==MB.ncols());
    }
  
  virtual value_type operator()(const size_type row, const size_type col) const 
    {  return value_type( op( MA(row,col), MB(row,col) ));  }
   
  /** Number of rows in the virtual matrix */
  virtual size_type nrows() const { return MA.nrows(); }

  /** Number of columns in the virtual matrix */
  virtual size_type ncols() const { return MA.ncols(); }
  
protected:
  const M& MA;
  const M& MB;
  const BinaryFunction& op;
};


} /* namespace mat */


#endif /* __MATRIXADAPTER_H__ */
