/** \file
  MatrixImage class.

  $Header: /u/nn/cvsroot/lissom/src/matriximage.h,v 1.14 2004/02/21 22:30:06 jbednar Exp $
*/

#ifndef __MATRIX_IMAGE_H__
#define __MATRIX_IMAGE_H__

#include <cassert>
#include <algorithm>
#include <cmath>

// Uses roundabout paths to the current directory to avoid Cray C bug
#include "../src/matrix.h"
#include "../src/pixel.h"
#include "../src/colorlookup.h"
#include "../src/boundingbox.h"
#include "../src/image.h"

namespace  Plot {


/**
   Image based on a rectangular matrix with an optional border.
   Member functions take coordinates in the original matrix rather
   than (scaled and offset) pixel coordinates.
   
   The border was deliberately included in this class rather than
   simply having an operation that takes an existing image and adds a
   border.  The reason is that certain outlining operations supported
   by this class will deliberately overwrite part of the border so
   that they enclose the precise area included in the outline, without
   losing any enclosed area to the nonzero width of the outline itself.
*/
template<class PixelType=RGBPixel<>, class PixelMatrix= typename MatrixType<PixelType>::rectangular >  
class MatrixImage : public AARImage<PixelType,PixelMatrix> {
private:
  
  /** Copied here explicitly to avoid apparent bug in GCC 2.95 */
  typedef typename AARImage<PixelType,PixelMatrix>::PixelSubscript PixelSubscript;
  
  /** Copied here explicitly to avoid apparent bug in GCC 2.95 */
  typedef typename AARImage<PixelType,PixelMatrix>::CartesianCoordinate CartesianCoordinate;
  
  /** Index into a matrix; allowed to be negative */
  typedef int MatrixSubscript;
  
  /** Width of surrounding border */
  PixelSubscript border;
  
  /** Matrix height */
  MatrixSubscript height;
  
  /** Matrix width */
  MatrixSubscript width;
  
  /** Ratio between matrix and image size */
  double size_scale;
  
  /** Returns true if the given MatrixSubscripts are within the underlying matrix
      (not in the border) */
  inline bool inside(const MatrixSubscript r, const MatrixSubscript c) const {
    return (r>=0 && r<height &&
 	    c>=0 && c<width);
  }
  
  /** Returns the row MatrixSubscript of the matrix element plotted at the given pixel location */
  inline MatrixSubscript matrix_row(const MatrixSubscript r) const 
    {  return MatrixSubscript(floor((r-border)/size_scale));  }
  
  /** Returns the column MatrixSubscript of the matrix element plotted at the given pixel location */
  inline MatrixSubscript matrix_col(const PixelSubscript c) const 
    {  return MatrixSubscript(floor((c-border)/size_scale));  }
  
  /** Returns the top pixel row corresponding to the given matrix row */
  inline PixelSubscript pixel_row(const MatrixSubscript r) const 
    {  return PixelSubscript(r*size_scale)+border;  }
  
  /** Returns the left pixel column corresponding to the given matrix column */
  inline PixelSubscript pixel_col(const MatrixSubscript c) const 
    {  return PixelSubscript(c*size_scale)+border;  }
  
  /** Returns the Cartesian coordinate of the center of the specified matrix element */
  inline CartesianCoordinate x_from_matrixcol(const MatrixSubscript col)
    {  return CartesianCoordinate(pixel_col(col))+size_scale/2;  }
  
  /** Returns the Cartesian coordinate of the center of the specified matrix element */
  inline CartesianCoordinate y_from_matrixrow(const MatrixSubscript row)
    {  return CartesianCoordinate(nrows()-1-(pixel_row(row)+size_scale/2))+0.5;  }
  
  /** Sets the area corresponding to the given matrix element to the given color, if in bounds */
  inline void draw_element(const MatrixSubscript row, const MatrixSubscript col, PixelType p)
    {  draw_rectangle(pixel_row(row), pixel_col(col), pixel_row(row+1), pixel_col(col+1),p);  }
  
  template<class BoundingBoxType> bool onboundary
  (const BoundingBoxType& bb, unsigned thickness, PixelSubscript r, PixelSubscript c) const;
  
  
public:
  /** \name Constructors and Destructors */
  //@{
  
  /** Empty image constructor */
  MatrixImage() : border(0), height(0), width(0), size_scale(1.0)  {  }

  template <class Matrix>
  MatrixImage(const Matrix &A, double scale=1.0, int borderwidth=0,
	      const ColorLookup<PixelType>& L=RYWColorLookup<PixelType>(),
	      PixelType bordercolor=default_border );
  
  template <class Matrix>
  MatrixImage(const Matrix &H, const Matrix &S, const Matrix &V, double scale=1.0,
	      int borderwidth=0, PixelType bordercolor=default_border );
  //@}
  
  
  /** Calls AARImage<>::draw_bar for a line centered on the given matrix element */
  // Should length be scaled?
  void draw_bar_on_matrix_element(const MatrixSubscript row, const MatrixSubscript col,
				  double angle, double length, double width=1, PixelType p=default_fg)
    {  draw_bar(x_from_matrixcol(col), y_from_matrixrow(row), angle, length, width, p);  }
  
  
  /** Draws a line of the indicated thickness around the matrix area
      enclosed by the given BoundingBox.  If infinite_outline is true,
      all points not enclosed by the given box are drawn, not just
      those nearest the outline border.
  */
  template<class BoundingBoxType> inline void outline_boundary
  (const BoundingBoxType& bb, unsigned thickness, PixelType p=default_fg, bool infinite_outline=false);
  
};



/** Constructor taking a single matrix, to plot one dimension.
    L determines the mapping from a magnitude to a color */
template<class PixelType, class PixelMatrix>
template <class Matrix>
MatrixImage<PixelType,PixelMatrix>::MatrixImage
(const Matrix &A, double scale, int borderwidth, const ColorLookup<PixelType>& L,
 PixelType bordercolor)
  : AARImage<PixelType>(typename Matrix::size_type(scale*A.nrows()+2*borderwidth),
			typename Matrix::size_type(scale*A.ncols()+2*borderwidth)),
  border(borderwidth), height(A.nrows()), width(A.ncols()), size_scale(scale) {
  draw_border(borderwidth,bordercolor);
  typedef typename Matrix::size_type Subscript;
  for   (Subscript r=0; r<A.nrows(); r++)
    for (Subscript c=0; c<A.ncols(); c++) {
      const Bounded::LargeFloat val = mat::elem(A,r,c);
      draw_element(r,c,L(val));
    }
}



/** Constructor taking matrices for Hue, Saturation, and Value. */
template<class PixelType, class PixelMatrix>
template <class Matrix>
MatrixImage<PixelType,PixelMatrix>::MatrixImage
(const Matrix &H, const Matrix &S, const Matrix &V, double scale, int borderwidth,
 PixelType bordercolor)
  : AARImage<PixelType>(typename Matrix::size_type(scale*H.nrows()+2*borderwidth),
			typename Matrix::size_type(scale*H.ncols()+2*borderwidth)),
  border(borderwidth), height(H.nrows()), width(H.ncols()), size_scale(scale) {
  assert(H.nrows()==S.nrows()); assert(H.nrows()==V.nrows());
  assert(H.ncols()==S.ncols()); assert(H.ncols()==V.ncols());
  draw_border(borderwidth,bordercolor);
  typedef typename Matrix::size_type Subscript;
  for   (Subscript r=0; r<H.nrows(); r++)
    for (Subscript c=0; c<H.ncols(); c++)
      draw_element(r,c,HSVPixel<>(mat::elem(H,r,c),mat::elem(S,r,c),mat::elem(V,r,c)));
}



template<class PixelType, class PixelMatrix>
template<class BoundingBoxType>
void MatrixImage<PixelType,PixelMatrix>::outline_boundary
(const BoundingBoxType& bb, unsigned thickness, PixelType pixel, bool infinite_outline)
{
  PixelType p=pixel;

  /** Fully-qualified pathname to nrows() function to mollify GCC 2.5.2 */
  const PixelSubscript nr = AARImage<PixelType,PixelMatrix>::nrows();
  const PixelSubscript nc = AARImage<PixelType,PixelMatrix>::ncols();
    
  /** If a transparent color was passed in, compute an appropriate
      default, considering the pixels which will be overwritten.
      Should really use a different routine nearboundary to compute
      this, because what matters for contrast it not the pixels that
      are overwritten but the ones next to the border.
  */
  if (p.istransparent()) {
    PixelAverage<> pa;
    for   (PixelSubscript row=0; row<nr; row++)
      for (PixelSubscript col=0; col<nc; col++)
	if (onboundary(bb,thickness,row,col))
	  pa+=get_pixel(row,col);
    p=contrasting_color(pa());
  }

  for   (PixelSubscript row=0; row<nr; row++)
    for (PixelSubscript col=0; col<nc; col++)
      if ((infinite_outline && !bb.inside(matrix_row(row),matrix_col(col))) ||
	  (onboundary(bb,thickness,row,col)))
	draw_pixel(row,col,p);
}



/**
  Returns whether the given pixel is on the boundary (to within
  the given thickness), and thus needs to be set.  It is considered
  to be on the boundary if:
  
   (a) it is not in the interior of the intersection between
       the boundingbox and the main image area, and
	 
   (b) a pixel less than <thickness> pixels away _is_.

  The algorithm used could be slow for large thicknesses, but those
  should be unlikely.  If such thicknesses are common, it can be
  rewritten to pre-compute the condition for each element beforehand,
  then look it up (instead of recalculating it as many as
  (thickness*2+1)^2 times).
*/
template<class PixelType, class PixelMatrix>
template<class BoundingBoxType>
inline bool MatrixImage<PixelType,PixelMatrix>::onboundary
(const BoundingBoxType& bb, unsigned thickness, PixelSubscript r, PixelSubscript c) const
{
  if (inside(matrix_row(r),matrix_col(c)) && bb.inside(matrix_row(r),matrix_col(c)))
    return false;
  
  /** Fully-qualified pathname to nrows() function to mollify GCC 2.5.2 */
  const PixelSubscript nr = AARImage<PixelType,PixelMatrix>::nrows();
  const PixelSubscript nc = AARImage<PixelType,PixelMatrix>::ncols();

  for   (PixelSubscript sr=max<MatrixSubscript>(0,r-thickness);
	 MatrixSubscript(sr)<=min<MatrixSubscript>(nr,r+thickness); sr++)
    for (PixelSubscript sc=max<MatrixSubscript>(0,c-thickness);
	 MatrixSubscript(sc)<=min<MatrixSubscript>(nc,c+thickness); sc++) {
      const MatrixSubscript row=matrix_row(sr);
      const MatrixSubscript col=matrix_col(sc);
      if (inside(row,col) && bb.inside(row,col))
	return true;
    }
  
  return false;
}


       
} /* namespace Plot */
#endif /* __MATRIX_IMAGE_H__ */
