/** \file
  Msg namespace: General-purpose messaging handlers

  $Header: /u/nn/cvsroot/lissom/src/msg.h,v 1.12 2004/05/13 23:59:14 jbednar Exp $
*/

#ifndef __MSG_H__
#define __MSG_H__

#include <iostream>
#include <string>
#include <cassert>
using std::string;

/** Namespace for general-purpose messaging.  A namespace is used rather
   than a class so that it will be open, i.e. so that other files may
   add functions and handlers with the same interface.  */
namespace Msg {



/**
  General-purpose message handlers, for user notification or error
  reporting.  The ones defined in msg.h are just the most basic ones;
  others can provide error counting, manage multiprocessor output, pop
  up dialog boxes, log output, etc.  If a function or class accepts a
  template or other argument of type Msg::fnPtr; the appropriate
  handler can be passed in to obtain the appropriate behavior.

  These are grouped into a class to avoid having to put them into a
  .c file.
*/
//@{
typedef void (*fnPtr)     (const string& s, const bool terminate);
namespace Handler {
inline void ignore       (const string&  , const bool          =true) {  }
inline void cout_notify  (const string& s, const bool terminate=true)
  {  std::cout <<                s << (terminate? ".\n" : "");  }
inline void cerr_error   (const string& s, const bool terminate=true)
  {  std::cerr << "Error: "   << s << (terminate? ".\n" : "");  }
inline void cerr_warning (const string& s, const bool terminate=true)
  {  std::cerr << "Warning: " << s << (terminate? ".\n" : "");  }
}
//@}



/** Named message level, for the LevelHandler.  Defined in this namespace instead of
    within the LevelHandler to save typing when using these literals. */
typedef enum { Overwhelm=0, Verbose=1, Notify=2, Requested=3, Warning=4, Error=5 } MessageLevel;
const int NumMessageLevels=6;



/**
  Basic class to provide flexibility in handling informational
  messages, warnings, and error messages.

  Sample usage:

    A class can have a member variable of this type (e.g. 'message'),
    using register_handler to determine the behavior, and operator() to
    call the apropriate routine (e.g. 'message(Msg::Error,"Badly formed
    input")').
  
    The default just ignores messages, but a typical stdout-based instance would be:
    LevelHandler<> stdout_message =
      LevelHandler<>().register_handler(Notify,Handler::cout_notify).register_handler(Warning,Handler::cerr_warning).register_handler(Error,Handler::cerr_error);
 
  If you want to use different names for your messages, just
  pass in the appropriate enum and size as template arguments.

  The Default fnPtr was originally a template argument, but Cray
  C++ cannot handle inline functions as template arguments (and
  the C++ standard might also not allow that.)
*/
template <class Level=MessageLevel, int size=NumMessageLevels>
class LevelHandler;
template <class Level, int size>
class LevelHandler {
  typedef LevelHandler<Level,size> self;
  
public:
  LevelHandler& register_handler( Level m, fnPtr h)
    {  assert(m<size);  handler[m]=h;  return *this;  }

  void operator() (Level m, const string& s, const bool terminate=true) const
    {  handler[m](s,terminate);  }

  LevelHandler()
    {  for (int i=0; i<size; i++) handler[i] = Default;  }

  static self*& default_instance()
    {  return default_instance_;  }
  
  static void set_default_instance(self* h)
    {  default_instance_ = h;  }
  
private:
  fnPtr handler[size];
  static self* default_instance_;
  static fnPtr Default;
};


/* Static variables */
template <class Level, int size>
fnPtr LevelHandler<Level,size>::Default=Handler::ignore;


/** Instance providing a simple way for different classes to share a
    single handler.  */
template <class Level, int size>
LevelHandler<Level,size>* LevelHandler<Level,size>::default_instance_=0;

									
} /* namespace Msg */

#endif /* __MSG_H__ */

