/** \file
  NeuralRegion and InternalNeuralRegion classes
  
  $Header: /u/nn/cvsroot/lissom/src/neuralregion.h,v 1.58 2004/02/05 18:10:11 jbednar Exp $
*/

#ifndef __NEURALREGION_H__
#define __NEURALREGION_H__

#include <utility>
#include <string>
using std::string;

#include "matrix.h"
#include "generic_stdlib.h"
#include "matrixadapter.h"
#include "boundingbox.h"
#include "weightset.h"
#include "bounded.h"
#include "pointermap.h"
#include "sequencetransform.h"



/** Base class encompassing both sensory and cortical regions
    (input and hidden layers, in NN terminology) organized as a 2D matrix.
*/
class NeuralRegion : public HasTable<mat::MatrixInterface<Bounded::Magnitude> > {
public:
  /** \name Types */
  //@{

  /** Generic type for a real-valued magnitude on the scale 0 to 1.0.
      Larger or smaller values are acceptable (boundaries are not enforced)
      but are not always meaningful.
  */
  typedef Bounded::Magnitude Magnitude;

  /** Type to use for an activity level. */ 
  typedef Magnitude Activity;

  /** Type to use for distances across the surface of the matrix in any direction. */
  typedef double Length;

  /**
    Matrix for activity. Currently this is only available read-only so
    that a derived class can use some other representation internally
    as long as this representation is available for other regions to
    see.
  */
  typedef MatrixType<Activity>::rectangular ActivityMatrix;

  /** A single index for a matrix. */
  typedef ActivityMatrix::size_type Subscript;

  /** Coordinate of a specific unit in a matrix or the size of a matrix. */
  typedef std::pair<Subscript,Subscript> SubscriptPair;

  /** A map of a quantity from 0..1.0 for each neuron. */
  typedef mat::MatrixInterface<Magnitude> MapBaseType;

  /** An adapter from an ActivityMatrix to a MapBaseType; this is only
      one example of many such adapters accepted by MapBaseType. */
  typedef mat::MatrixInterfaceAdapter<ActivityMatrix,Magnitude> ActivityAdapter;
  
  /** BoundingBox on the surface of the NeuralRegion */
  typedef Boundary::AARBoundingBox<int,int> Bounds;

  /** Activation function */
  typedef SequenceTransform::OneD<ActivityMatrix,ActivityMatrix> ActivationFunction;

  /** Packages up specification of desired or actual input region.
      Should probably change this interface to use an AARectangle
      instead, if practical.
  */
  struct Dimensions {
    Dimensions(Subscript h, Subscript w, Length xo=0.5, Length yo=0.5)
      : height(h), width(w), xoffset(xo), yoffset(yo) {  }
    Subscript height;   /**< Vertical   length. */
    Subscript width;    /**< Horizontal length. */
    Length    xoffset;  /**< Horizontal offset relative to the origin. */
    Length    yoffset;  /**< Vertical   offset relative to the origin. */
  };
  //@}

  /** \name Constructors and destructors */
  //@{
  NeuralRegion(string name_i, Dimensions dims)
    : HasTable<mat::MatrixInterface<Bounded::Magnitude> >(),
      name_str(name_i), output(dims.height,dims.width), 
      bp(dims.height,dims.width),
      residual(dims.height,dims.width) { }
  
  NeuralRegion(const NeuralRegion& other) /**< Copy constructor */
    : HasTable<mat::MatrixInterface<Bounded::Magnitude> >(),
      name_str(other.name_str), output(other.output), 
      bp(other.bp), residual(other.residual) { }
  
  virtual ~NeuralRegion() { }
  //@}
  

  /** Activate the map given the current inputs. Must be implemented by derived classes. */
  virtual void activate(bool learn=false, bool settle=true, bool activatefn=true) =0; 

  
  /** \name Low-level interface -- avoid if possible */
  //@{
  
  /*   activity(Subscript i, Subscript j) may need to be made non-virtual for speed.  */
  /** Returns the current activity for the specified unit. */
  virtual Activity activity(Subscript i, Subscript j) const {  return output[i][j];  }
  /** Returns a const matrix of the current activity values. */

  virtual const ActivityMatrix& const_activity() const      {  return output;        }

  virtual const ActivityMatrix& backprojection() const      {  return bp;    }

  virtual const ActivityMatrix& residual_activity() const      {  return residual;    }

  //  virtual ActivityMatrix& activity()                        {  return output;        }

  //@}

  /** Returns the declared name of this region */
  const string& name() const {  return name_str;  }

  const int nrows() const {  return int(output.nrows());  }
  const int ncols() const {  return int(output.ncols());  }

  /** Returns true if this region is an internal region, i.e. if it
      accepts input directly from other regions.  Regions are assumed
      not to accept such input unless they declare otherwise. */
  virtual bool is_internal() const  {  return false;  }
  
  /** Returns true if this region might contain modifiable weights.
      Regions are assumed to be modifiable unless they declare
      otherwise. */
  virtual bool is_plastic() const  {  return true;  }

  /** Save the region's state to the given file.  The default
      implementation is a stub, for networks with no saveable state;
      others should implement some other form of saving.  This
      interface might need to be modified at some point.
      It would be made const except that some saving methods
      use an automatic verify that relies on write access to the
      region.
  */
  virtual void save_state(const string&) { }
  
  /** Load state from the given file.  The default implementation is a
      stub, for networks with no saveable state; others should
      implement some other form of saving.  Returns true if the load
      was successful.  This interface might need to be modified at
      some point; it might be possible to eliminate the iteration
      (first argument). */
  virtual bool restore_state(const int, const string&) {  return true;  } 

  /** If applicable, prune connections.  There may be some more
      general way to implement such functionality than having an explicit
      interface for routines as specific as this. */
  virtual void prune() {  }

  /** If applicable, lesion (disable) or unlesion (enable) the units
      within the specified Bounds.  The alive argument determines whether
      to disable or enable the units (false=disable).  Unless
      otherwise implemented by a specific region type, is a no-op. */
  virtual void lesion(const Bounds&, bool=false) {  }
  
  /** Returns the size of this region, in terms of the number of bytes
      taken to store its connections.  Required to be implemented by
      each type of region.  */
  virtual double size_connection_bytes()   const=0;
  
  /** Returns the size of this region, in terms of the number of
      connections which might hold unique values.  (This might be
      different from the total number of connections if e.g. some
      connections are shared between different units.)  */
  virtual double size_unique_connections() const=0;

  /* Backproject current activity. Must be implemented by derived classes. yschoe 
  virtual void backproject() = 0;
  */

private:
  const string name_str;

  /** \name Quasi-protected members -- should be used only by subtypes, but not necessarily on self. */
  //@{
public:
  /** Returns a BoundingBox enclosing the entire region */
  Bounds bounds() const {  return Bounds(0,0,output.ncols()-1,output.nrows()-1);  }
  //@}

// yschoe: make this temporarily public; was protected:
public:
  /** Matrix of current activity values to be propagated to other layers. */
  ActivityMatrix output;
  ActivityMatrix bp;
  ActivityMatrix residual;
};



/**
  Abstract base class for any NeuralRegion which accepts input from
  other NeuralRegions (unlike e.g. a Retina).
*/
class InternalNeuralRegion : public NeuralRegion {
public:
  /** \name Constructors and destructors */
  //@{
  InternalNeuralRegion(string name_i, Dimensions dims)
    : NeuralRegion(name_i,dims) { }
  
  virtual ~InternalNeuralRegion() { }
  //@}

  /**
    Matrix for weights; this does not have to be the actual type of
    the weights (if any) as they are actually stored internally, but
    it is the type used in interfaces that e.g. provide preset weights
    or plot a set of weights.
  */
  typedef ActivityMatrix WeightMatrix;

  /** Function which computes a starting or fixed set of weights. */
  typedef Generic::unary_virtual_function<double,WeightMatrix> WeightFunction;

  /** BoundingBox enclosing a weight matrix */
  typedef Boundary::BoundingBox<int,int> WeightBounds;

  /** A set of weights from a unit to or from a NeuralRegion */
  typedef WeightSet<WeightMatrix,WeightBounds,MapBaseType,mat::MatrixInterfaceAdapter<WeightMatrix,Magnitude> > Weights;

  /** BoundingBox for weights above a certain threshold */
  typedef WeightThreshold<int,int,double,WeightBounds,WeightMatrix> WeightOutlineBounds;

  
  /** Computes input dimensions that would be required to avoid border effects. */
  virtual Dimensions input_dimensions(WeightFunction& fn, Length size_scale=1.0) =0;

  virtual bool is_internal() const  {  return true;  }
  
  /** Declare that the given region gives input to this region, and define any
      connections needed. The connections will have the given name.
      Accepts fn (a unary function of size_scale) to initialize the weights.
      Any existing set of weights by that name will be overwritten.
  */
  virtual void add_input(const string& name, NeuralRegion& upstream_region,
			 WeightFunction& fn, Length size_scale) =0;

  /** Backproject current activity. Must be implemented by derived classes. yschoe */
  virtual void backproject() = 0;

  /** Backproject core routine: to be called by implementation classes'
      backproject().  This code is too detailed for such an abstract class,
      and should be moved elsewhere if possible. */
  virtual void backproject(const string& name, const double gammaaff=1.0) {

	// This function should be called by derived classes' 
	// backproject(noarg), which will know which map to call.
	// --- cout << " - " << name << endl;

	// 1. get named input (upstream's output). Cleaning up the 
	//    backprojection region should be done prior to calling this
	//    function, so it is omitted here.
	NeuralRegion* inp = dynamic_cast<NeuralRegion*>(get_input(name));
	if (!inp) {
	  // Messaging not currently possible from this routine
	  //cout << "Error: input region nonexistent" << endl;
		return;
	}

	// 2. Loop thorough all neurons in the current NeuralRegion bp 
	int nr = output.nrows(), nc = output.ncols();
	for (int i=0; i<nr; i++)
	  for (int j=0; j<nc; j++) {

	    double cur_output = bp[i][j];

	    // 2.1 get named weights and bounds (upstream->downstream)
	    const WeightMatrix  wmat = get_weights(name,i,j);
	    const WeightBounds* wbounds = get_weights_bounds(name,i,j);

	    // 2.2 get current activity (this map's output)
	    //    output is the current ActivityMatrix;
	    int xl = wbounds->aarectangle().corners().xl;
	    int yl = wbounds->aarectangle().corners().yl;
	    int xh = wbounds->aarectangle().corners().xh;
	    int yh = wbounds->aarectangle().corners().yh;

	    // 2.3 backproject loop
	    for (int x=xl; x<=xh; ++x)
	      for (int y=yl; y<=yh; ++y)
	        (inp->bp[x][y]) += cur_output * wmat[x-xl][y-yl] / gammaaff;

	    delete wbounds;

	  }
	// 2. end

  };

  
#ifdef OBJ_LISSOM
  
  /** If supported, calculate response to residual activity input. yschoe */
  virtual void resp_to_residual() {};

  /** resp_to_residual core routine: to be called by implementation classes'
      resp_to_residual().  Should be moved to another file if possible. */
  virtual void resp_to_residual(const string& name, const double gammaaff=1.0) {

	// This function should be called by derived classes' 
	// resp_to_residual(noarg), which will know which map to call.
	//
	// precondition: upstream "residual" must have been properly calculated
 	// assumption: result will be ADDED to the current "output"
	//	so "output" should have been cleaned if this is the first time.

	// 1. propagation starts at the current "residual"

	// 2. Loop thorough all neurons in the current NeuralRegion residual
	NeuralRegion* inp = dynamic_cast<NeuralRegion*>(get_input(name));

	int nr = output.nrows(), nc = output.ncols();
	for (int i=0; i<nr; i++)
	  for (int j=0; j<nc; j++) {

	    // 2.1 get named weights and bounds (upstream->downstream)
	    const WeightMatrix  wmat = get_weights(name,i,j);
	    const WeightBounds* wbounds = get_weights_bounds(name,i,j);

	    // 2.2 get current activity (this map's output)
	    //    output is the current ActivityMatrix;
	    int xl = wbounds->aarectangle().corners().xl;
	    int yl = wbounds->aarectangle().corners().yl;
	    int xh = wbounds->aarectangle().corners().xh;
	    int yh = wbounds->aarectangle().corners().yh;

	    // 2.3 backproject loop
	    double sum = 0.0;

	    for (int x=xl; x<=xh; ++x)
	      for (int y=yl; y<=yh; ++y)
		sum += (inp->residual[x][y]) 
			      * wmat[x-xl][y-yl];

	    output[i][j] += sum / gammaaff;

	    delete wbounds;

	  }
	// 2. end

  }

  /** If supported, train obj lissom weights and normalize */
  virtual void train_obj_weights() {};

  /** If supported, randomize lateral connection weights */
  virtual void randomize_lat_wts() {};
#endif /* OBJ_LISSOM */


  //@{ \name Weight access functions 

  /** Returns the specified weights for a unit */
  virtual const Weights weights(const string& name, int ui=0, int uj=0) const
  {  return Weights(new WeightMatrix(get_weights(name,ui,uj)),get_weights_bounds(name,ui,uj));  }


  /** Returns a bounding box returning inside() for units over the specified threshold.
      The box should be deleted when no longer needed.  This function should probably
      be moved outside the class, because it can be implemented in terms of weights().
  */
  virtual const WeightBounds* weights_outline
  (const string& name, int ui=0, int uj=0, WeightMatrix::value_type threshold=0) const
  {
    const WeightMatrix  wts = get_weights(name,ui,uj);
    const WeightBounds* bds = get_weights_bounds(name,ui,uj);
    const WeightBounds* otl = new WeightOutlineBounds(*bds,wts,threshold);
    return otl;
  }


//private: /* Eventually */
  /**
    Returns weights from the given unit from the group of weights with
    the given name.  The weights returned may e.g. be a constant set
    of weights for all units regardless, or they may be generated on
    the fly, so bear that in mind.  If a region does not have any
    weights with the specified name, it should return an empty (0x0)
    matrix.
  */
  virtual const WeightMatrix get_weights(const string& name, int ui=0, int uj=0) const =0;

  /**
    Allocates and returns a BoundingBox on the map for the weights of the 
    given unit; all active weights must lie within.  This box is not
    guaranteed to be the minimum such box; it generally stays the same
    even if units are pruned, and some types may only implement a
    rectangular box even if weights are internally constrained to
    e.g. a circular region.
  */
  virtual WeightBounds* get_weights_bounds(const string& name, int ui=0, int uj=0) const =0;

  //@}

  //@{ \name Input access functions
  /** Returns the specified input region for a unit */
  virtual NeuralRegion* get_input(const string& name) = 0;
  //@}

};


#endif /* __NEURALREGION_H__ */
