/** \file
  NeuralRegionManager class.
  
  $Header: /u/nn/cvsroot/lissom/src/neuralregionmanager.h,v 1.28 2004/02/05 18:20:01 jbednar Exp $
*/

#ifndef __NEURALREGIONMANAGER_H__
#define __NEURALREGIONMANAGER_H__

#include "neuralregion.h"
#include "matrixadapter.h"
#include "plotspecification.h"
#include "parametermap.h"
#include "ppm_draw.h"



/**  Manages housekeeping such as plotting and analysis for a NeuralRegion.  */
class NeuralRegionManager {
  typedef string bStr;
  typedef const string cStr;
  typedef InternalNeuralRegion::Weights Weights;
  typedef Scale::Linear<double> ScaleType;
  
public:
  /** Constructor.  If present, params_ should be allocated, as it
      will be deleted on destruction.

      If register_maps is true, all the maps defined by the region
      will be registered in alphabetical order to plot in the
      Activity plot_group, followed by a plot of the map Activity.
  */
  NeuralRegionManager(NeuralRegion* region, ParamMap& params_,
		      bool owns_params=true, const bool register_maps=true)
    : r(region), params(&params_,owns_params) {
    if (register_maps) {
      const NeuralRegion::Table& table = r->table();
      for (NeuralRegion::Table::const_iterator i=table.begin(); i!=table.end(); i++) {
	const string name = i->first;
	define_plot("Activity", name, name);	
      }
      region->table_set("Activity", new NeuralRegion::ActivityAdapter(region->const_activity()));
      define_plot("Activity", "Activity", "Activity");

      region->table_set("Backprojection", new NeuralRegion::ActivityAdapter(region->backprojection()));
      define_plot("Backprojection", "Backprojection", "Backprojection");

      region->table_set("Residual", new NeuralRegion::ActivityAdapter(region->residual_activity()));
      define_plot("Residual", "Residual", "Residual");
    }

    /* Declare useful parameters.  Should probably also mark them read-only. */
    ParameterMap<>& p = *dynamic_cast< ParameterMap<>* >(params.valueptr());
    p.set_local("plastic", (region->is_plastic()?  True : False ));
    p.set_local("internal",(region->is_internal()? True : False ));
  }
  
  virtual ~NeuralRegionManager() {  }

  /** Add the given map to the list of ones available, overwriting any
      old one with that name.
      
    The pointer should be allocated using new, since it will be
    deleted when this object is destroyed.
  */
  void map_add(const string& name, NeuralRegion::MapBaseType* map) {
    assert (int(map->nrows())==int(r->const_activity().nrows()));
    assert (int(map->ncols())==int(r->const_activity().ncols()));
    r->table_set(name,map);
  }

  /** Table holding lists of items to plot. */
  typedef Plot::PlotSpecification<NeuralRegion::MapBaseType,bStr,Scale::Linear<>,
				  NeuralRegion::Table> PlotSpec;
  typedef std::vector<PlotSpec*>                     PlotSpecList;
  typedef PointerMap<PlotSpecList>              PlotSpecListTable;

  /** Iterator for PlotSpecs */
  virtual PlotSpecList::const_iterator plotspeclist_begin(const string& group) const {
    const PlotSpecList* p=plotspecs.getptr(group);
    return (p? plotspecs.getptr(group)->begin() : PlotSpecList::const_iterator());
  }

  /** Iterator for PlotSpecs */
  virtual PlotSpecList::const_iterator plotspeclist_end(const string& group) const {
    const PlotSpecList* p=plotspecs.getptr(group);
    return (p? plotspecs.getptr(group)->end() : PlotSpecList::const_iterator());
  }

  /** Define a new plot with the given named maps.  Returns a pointer
      to the constructed specification. */
  inline PlotSpec* define_plot(const string& group, cStr& plotname,
			       cStr& strength, cStr& color="", cStr& confidence="")
    {  return define_plot(group,newPlotSpec(plotname,strength,color,confidence));  }


  /** Add maps for the given unit to the given MatrixTable, returning true on success. */
  inline bool add_maps_for_unit(PlotSpec::MatrixTable& maptable,
				const PlotSpec& spec,
				const int ui, const int uj) const 
    {  return add_maps_for_unit(maptable,spec.name(),spec.name(),spec,ui,uj);  }
    
  inline bool add_maps_for_unit(PlotSpec::MatrixTable& maptable,
				const string& weightname,
				const string& mapname,
				const PlotSpec& spec,
				const int ui, const int uj) const {
    
    const InternalNeuralRegion* p = internalregion();  assert(p);

    /* Error checking */
    if (ui>=p->nrows() || uj>=p->ncols()) {
      message(Msg::Error,"Unit (" +
	      String::stringrep(ui) + "," +
	      String::stringrep(uj) + ") is out of range of region " + p->name() );
      return false;
    }

    /* Add a map for the specified weights.  Currently does not supply
       any "reference count" arguments to the weight_scale command,
       since those are tough to determine in general. */
    const Scale::Linear<double> scale = weight_scale(spec, p->weights(weightname,ui,uj));
    maptable.set(mapname,p->weights(weightname,ui,uj).as_map(*(spec.reference_mapptr()),scale));

    return true;
  }


  /** Figure out the appropriate weight scale for the given unit.  If the reference
      count is greater than zero, it will be used for count-based
      scale types instead of the actual count of the number of active
      units.
  */
  Scale::Linear<double> weight_scale(const PlotSpec& spec, const Weights& weights, int reference_count=-1) const {
    double factor=1.0;
    double offset=0.0;
    
    const WtScale::Type type       = WtScale::Type(spec.lookup_param("ppm_weight_scale_type",       int(WtScale::Relative)));
    const double     	fixed_weight_multiplier  = spec.lookup_param("ppm_weight_fixed_multiplier", 1.0);
    const double     	fixed_weight_offset      = spec.lookup_param("ppm_weight_fixed_offset",     0.0);
    const double     	threshold                = spec.lookup_param("ppm_weight_threshold",        0.0);
    
    /* Simplest case */
    if (type == WtScale::Fixed) {
      factor = fixed_weight_multiplier;
      offset = fixed_weight_offset;
      }
    else {
      /* Compute scaling from max value, min value, number of connections, and/or num active */

      /* Initialize counts and totals to minimum or maximum possible, as appropriate */
      double max    = 0.0;
      double min    = 1.0;
      int    count  =   0;
      int    active =   0;

      typedef Weights::Matrix   Matrix;
      typedef Weights::Bounds   Bounds;
      typedef Matrix::size_type Subscript;
      
      const Matrix&   A  = weights.matrix();
      const Bounds&   b  = weights.bounds();
      const Subscript ro = b.aarectangle().corners().xl;
      const Subscript co = b.aarectangle().corners().yl;
      
      for   (Subscript rr=0; rr<A.nrows(); rr++)
	for (Subscript cc=0; cc<A.ncols(); cc++)
	  if (b.inside(rr+ro,cc+co)) {
	    count++;
	    const double value = mat::elem(A,rr,cc);
	    if (value > threshold) {
	      active++;
	      if (value > max) max = value;
	      if (value < min) min = value;
	    }
	  }
	  
      /* Override actual count if appropriate */
      if (reference_count>0) count = reference_count;

      switch (type) {
      case WtScale::Relative:     factor = ((max-min)>0 ? 1.0/(max-min) : 1.0);
	                          offset = -min;	            break;
      case WtScale::PeakRelative: factor = (max>0 ? 1.0/max : 1.0); break;
      case WtScale::Active:       factor = 1.0*active;   	    break; 
      case WtScale::Count:        factor = 1.0*count;               break;
      default:
	/* It would be nice to limit the number of these messages generated... */
	message(Msg::Error,"Unknown WtScale type: " + String::stringrep(int(type)));
	factor = 0; break;
      }
    }
    
    //cout << "plot: (" << factor << "," << offset << ")" << endl;
    return ScaleType(factor,offset);
  }

  
  /** Returns a pointer to the NeuralRegion managed by this object */
  inline NeuralRegion* region() {  return r;  }

  /** Returns a const pointer to the NeuralRegion managed by this object */
  inline const NeuralRegion* region() const {  return r;  }

  /** If the region managed is an InternalNeuralRegion, returns a
     pointer of that type to the NeuralRegion managed by this object
     (otherwise 0). */
  inline InternalNeuralRegion* internalregion()
    {  return dynamic_cast<InternalNeuralRegion*>(r);  }

  /** Const version of internalregion() */
  inline const InternalNeuralRegion* internalregion() const
    {  return dynamic_cast<const InternalNeuralRegion*>(r);  }


  /** Declare that the given region gives input to this region.
      The connections will have the given name.
      Accepts a unary function of size_scale to initialize the weights.
      Returns the PlotSpec defined for this input, if any, so that the
      parameters of the plot can be controlled externally.
  */
  PlotSpec* add_input(const string& name,
		      NeuralRegionManager& upstream_manager,
		      InternalNeuralRegion::WeightFunction& fn,
		      NeuralRegion::Length size_scale=1.0) {
    assert (region());
    assert (internalregion());
    internalregion()->add_input(name,*(upstream_manager.region()), fn, size_scale);
    PlotSpec* p=upstream_manager.newPlotSpec(name,name);
    assert (p);
    define_plot("Weights",p);
    return p;
  }

  /** Returns a default unit valid for this region */
  int default_ui() const {  return std::max(0,region()->nrows()/2-1);  }
  int default_uj() const {  return std::max(0,region()->ncols()/2-1);  }

  static void message(Msg::MessageLevel m, const string& s)
    {  PlotSpec::message(m,s);  }

  ParamMap&       param_map()       {  return *(params.valueptr());  }
  const ParamMap& param_map() const {  return *params;               }
  
private:  

  /** PlotSpec constructor accepting the names of one to three Matrices to plot. */
  PlotSpec* newPlotSpec(cStr& plotname, cStr& strength, cStr& color="", cStr& confidence="") const {
    // Can the dynamic_cast be eliminated?
    /* Specific type is needed here for set_local; requires that the
       call originally used a ParameterMap<>. */
    ParameterMap<>* p = dynamic_cast< ParameterMap<>* >(params.valueptr());
    assert (p);
    return new PlotSpec(r->name(),r->table(),plotname,*p,strength,color,confidence);
  }
    
  /** PlotSpec copy constructor, allowing some data to be overridden */
  PlotSpec* newPlotSpec(PlotSpec& o, cStr& plotname, cStr& strength, cStr& color="", cStr& confidence="") const {
    // Can the dynamic_cast be eliminated?
    ParameterMap<>* p= dynamic_cast<ParameterMap<>*>(&(o.param_map()));
    assert (p);
    return new PlotSpec(r->name(),r->table(),plotname,*p,strength,color,confidence);
  }

  /** Returns the specified plot group, creating it if it doesn't already exist */
  inline PlotSpecList& plotspec_group(const string& name) {
    PlotSpecList* p=plotspecs.getptr(name);
    if (!p) {
      p=new PlotSpecList;
      plotspecs.set(name,p);
    }
    return *p;
  }

  /** Add a new plot to the given plot group, using the given set of
      parameters.  The parameters must have been allocated using new,
      since they will be deleted when this object is destroyed. */
  inline PlotSpec* define_plot(const string& group,PlotSpec* spec) {
    Generic::insert_named(plotspec_group(group), spec->name(), spec);
    return spec;
  }

  /** Collection of groups of PlotSpecs */
  PlotSpecListTable plotspecs;

  /** Region which we are managing */
  NeuralRegion* r;

  /** User-settable parameters */
  OwningPointer<ParamMap> params;
};



#endif /* __NEURALREGIONMANAGER_H__ */

