/** \file
  ParamMap class.
  
  $Header: /u/nn/cvsroot/lissom/src/parammap.h,v 1.6 2000/12/07 23:21:19 jbednar Exp $
*/

#ifndef ___PARAMMAP_H___
#define ___PARAMMAP_H___

#include "polymorph.h"
#include "pointerlookup.h"
#include "msg.h"


/** Abstract base class providing Typeless-specific functionality for
    a PointerMap.  If error reporting is desired, the message()
    routine can be overridden by a derived class. */
class ParamMap : public PointerLookup<Typeless> {
public:
  typedef  ParamMap self;
  virtual ~ParamMap() { }
  
  /** Looks up the given parameter, setting the given location to its
      value if the types match.  If warn_unknown is true, an error is
      generated if the named parameter is not found.  If warn is true,
      other types of problems (e.g. type mismatch) generate errors.
  */
  template <class T>
  const T* get(const char* name, T*& x,
	       const bool warn=true, const bool warn_unknown=true) const {

    const Typeless*     p  = getptr(name);
    const Polymorph<T>* pp = dynamic_cast<const Polymorph<T>*>(p);
    
    if (warn_unknown && !pp)
      message(Msg::Error,"ParamMap: No parameter (or constant) `" +
	      string(name) + "' is defined");
    
    else if (warn && p && !pp)
      message(Msg::Error,"ParamMap: Parameter `" +
	      string(name) + "' does not have the expected type");
    
    return x = (pp ? pp->valueptr() : 0);
  }
  
  
  /** Wrapper for get() which substitutes the given default value if a
      value cannot be obtained from the map. */
  template <class T>
  T get_with_default(const char* name, const T& default_val,
		     const bool warn=true, const bool warn_unknown=false) const {
    T* p=0;
    get(name,p,warn,warn_unknown);
    return (p ? *p : default_val);
  }

  /** Returns a new child map (presumably linked to this map) of the
      same specific type as self; should be implemented by all derived
      classes.  The return type should always be ParamMap* to avoid
      problems with older compilers, e.g. Cray C++.  The child is
      allocated using new, and should be deleted when no longer
      needed.

      If owned is true, this object is responsible for the destruction
      of the child object when this child is destroyed.  (Usually a
      ParamMap will be associated with another object, to be destroyed
      when that object is destroyed, but passing true here can be
      useful for sub-maps owned by a single object.
  */
  virtual ParamMap* new_child(const string name_="", const bool owned=false)=0;
    
  /** Searches self and children for the specified child map,
      returning it if found or self otherwise.  The specification
      format is not determined, and the self may be returned in
      all cases if no list of children is kept.
  */
  virtual const ParamMap& get_child(const string& name_="") const=0;
  virtual       ParamMap& get_child(const string& name_="")=0;
    
private:
  virtual void message(Msg::MessageLevel, const string&) const { };
};


#endif /* ___PARAMMAP_H___ */
