/** \file
  Pixel classes.

  $Header: /u/nn/cvsroot/lissom/src/pixel.h,v 1.17 2002/10/08 04:21:43 jbednar Exp $
*/

#ifndef __PIXEL_H__
#define __PIXEL_H__

#include <cmath>
#include "mathconstants.h"

// Uses roundabout path to the current directory to avoid Cray C bug
#include "../src/boundednumber.h"


/** Namespace for image plotting functions. */
namespace  Plot {


/******************************************************************************/
/* Pixel classes                                                              */
/******************************************************************************/


/** Abstract base class for a square image pixel, possibly transparent. */
class Pixel {
public:
  /** Default constructor; pixel remains transparent. */
  Pixel() : transparent(true) {  }
  
  /** Constructor for a possibly non-transparent pixel. */
  Pixel(bool visible) : transparent(!visible) {  }
  
  /** Copy constructor. */
  Pixel(const Pixel& other) : transparent(other.transparent)  {  };

  /** Returns whether the pixel is transparent. */
  bool istransparent() {  return transparent;  }
  
  /** Makes this pixel visible */
  void show() {  transparent=false;  }
  
  /** Makes this pixel transparent */
  void hide() {  transparent=true;  }

  /** Integer Bounded number for a Pixel */
  typedef Bounded::Integer<unsigned char> IComponent;
  
  /** Floating-point Bounded number for a Pixel */ 
  typedef Bounded::Float<  float,1,0,Bounded::Crop<Bounded::LargeFloat> > FComponent;

  /** Circularly-wrapping Bounded number for the Hue */
  typedef Bounded::Float<  float,1,0,Bounded::Wrap<Bounded::LargeFloat> > HueComponent;

private:
  bool transparent; 
};


  
/* Forward declarations to allow HSVPixel and RGBPixel to convert to
   and from one another
*/
template<class>       class RGBPixel;
template<class,class> class HSVPixel;

 
  
/** Red,Green,Blue image Pixel. */
template<class Component=Pixel::IComponent>
class RGBPixel : public Pixel
{
  typedef RGBPixel self;
  
public:
  /**\name Constructors and destructors */
  //@{
  /** Default constructor; pixel remains transparent. */
  RGBPixel() : Pixel(), r(0), g(0), b(0) {  }
  
  /** Constructor taking a single greyscale value (required by some
      matrix types).

    The pixel remains transparent, although it's unclear (no pun
    intended) if that's appropriate.
  */
  RGBPixel(FComponent value) : Pixel(), r(value), g(value), b(value) {  }
  
  /** Constructor taking floating-point RGB triple */
  RGBPixel(FComponent red, FComponent green, FComponent blue, bool visible=true) : Pixel(visible),
    r(red), g(green), b(blue) {  }
  
  /** Copy constructor */
  RGBPixel(const self& other)
    : Pixel(other), r(other.r), g(other.g), b(other.b)  {  }
  
  /** Constructor taking HSV pixel */
  template<class OComponent, class OHComponent>
  RGBPixel(const HSVPixel<OComponent,OHComponent>& O);
  //@}
  
  /**\name Access functions for the components of the pixel */
  //@{
  Bounded::Magnitude red()   const {  return r.mag();  } /**< Returns the red   component on a scale [0.0,1.0] */
  Bounded::Magnitude green() const {  return g.mag();  } /**< Returns the green component on a scale [0.0,1.0] */
  Bounded::Magnitude blue()  const {  return b.mag();  } /**< Returns the blue  component on a scale [0.0,1.0] */
  //@}
  
  /** Unary addition */
  self operator+= (const self &B) 
    {  r=r+B.r; g=g+B.g; b=b+B.b;  }

  /** Multiplicative scaling */
  self& operator*=(const Bounded::LargeFloat m)
    {  r*=m; g*=m; b*=m; return *this; }

  /**\name Functions written in terms of the public access functions */
  //@{
  
  /** Writes a readable representation to a stream */
  std::ostream& put(std::ostream& s) const
  {  return s << "(" << red() << "," << green() << "," << blue() << ")";  }
  
  /** Binary addition of RGBPixel values */
  self operator+(const self& o) {
    self tmp;
    tmp.r=r+o.r;
    tmp.g=g+o.g;
    tmp.b=b+o.b;
    return tmp;
  }

  /** Writes raw data to a binary file.
      You must ensure that the max_val is of the type you want the output to use. */
  template<class OutputType>
  void write_binary(std::ostream& s, OutputType max_val) const {
    s.put(OutputType(max_val*red()));
    s.put(OutputType(max_val*green()));
    s.put(OutputType(max_val*blue()));
  }
  //@}
  
private:
  Component r,g,b;
};



/** Stream output */ 
template<class Component>
std::ostream& operator<<(std::ostream& s, const RGBPixel<Component>& o)
  {  return o.put(s);  }

 

/** Hue,Saturation,Value image Pixel. */
template<class Component=Pixel::FComponent, class HComponent=Pixel::HueComponent>
class HSVPixel : public Pixel
{
public:
  /**\name Constructors and destructors */
  //@{
  /** Default constructor; pixel remains transparent. */
  HSVPixel() : Pixel(), h(0), s(0), v(0) {  }
  
  /** Constructor taking HSV triple */
  HSVPixel(HComponent hue, Component saturation, Component value)
    : Pixel(true), h(hue), s(saturation), v(value)  {  }
  
  /** Copy constructor */
  HSVPixel(const HSVPixel& other)
    : Pixel(other), h(other.h), s(other.s), v(other.v)  {  }
  
  /** Constructor taking RGB pixel */
  template<class OComponent>  
  HSVPixel(const RGBPixel<OComponent>& other);
  //@}
  
  /**\name Access functions for the components of the pixel */
  //@{
  Bounded::Magnitude hue()        const {  return h.mag();  } /**< Returns the hue        component on a scale [0.0,1.0] */
  Bounded::Magnitude saturation() const {  return s.mag();  } /**< Returns the saturation component on a scale [0.0,1.0] */
  Bounded::Magnitude value()      const {  return v.mag();  } /**< Returns the value      component on a scale [0.0,1.0] */
  //@}
  
  /** Stream output */
  std::ostream& put(std::ostream& s)
  {  return s << "(" << hue() << "," << saturation() << "," << value() << ")";  }

  /** This parameter can be set to a nonzero value if you want hues to
      go in the opposite direction */
  static int reversed_color_order;
  
private:
  HComponent h;
  Component  s,v;
};


/* Static variables */
template<class Component, class HComponent>
int HSVPixel<Component,HComponent>::reversed_color_order=false;


/** Stream output */ 
template<class Component>
std::ostream& operator<<(std::ostream& s, const HSVPixel<Component>& o)
{  return o.put(s);  }


/** Constructor for an RGBPixel from an HSVPixel.

  Algorithm is from _Computer_Graphics_, 2nd Ed., Foley et al, 1990,
  except Hue modified to lie in [0,1] instead of [0,360].
*/
template<class Component>
template<class OComponent, class OHComponent>
RGBPixel<Component>::RGBPixel(const HSVPixel<OComponent,OHComponent>& other) : Pixel(other)
{
  int    j;
  double rd, gd, bd;
  double f, p, q, T;
  double hue;
  
  const double h=(HSVPixel<OComponent,OHComponent>::reversed_color_order?
		  1-other.hue() : other.hue());
  const double s=other.saturation();
  const double v=other.value();
  
  /* convert HSV back to RGB */
  if (s==0.0) { rd = v;  gd = v;  bd = v; }
  else {
    /* hue ranges [0,6) */
    hue = h * 6.0;
    
    j = (int) floor(hue);
    if (j<0) j=0;          /* either h or floor seem to go neg on some sys */
    f = hue - j;
    p = v * (1-s);
    q = v * (1 - (s*f));
    T = v * (1 - (s*(1 - f)));
    
    switch (j) {
    case 0:  rd = v;  gd = T;  bd = p;  break;
    case 1:  rd = q;  gd = v;  bd = p;  break;
    case 2:  rd = p;  gd = v;  bd = T;  break;
    case 3:  rd = p;  gd = q;  bd = v;  break;
    case 4:  rd = T;  gd = p;  bd = v;  break;
    case 5:  rd = v;  gd = p;  bd = q;  break;
    default: rd = v;  gd = T;  bd = p;  break;  /* never happen */
    }
  }
  
  r = Component(Bounded::LargeInt(floor( rd*r.Max + 0.5 )));
  g = Component(Bounded::LargeInt(floor( gd*g.Max + 0.5 )));
  b = Component(Bounded::LargeInt(floor( bd*b.Max + 0.5 )));
}



/** Constructor for an HSVPixel from an RGBPixel.

  Algorithm is from _Computer_Graphics_, 2nd Ed., Foley et al, 1990,
  except Hue modified to lie in [0,1] instead of [0,360].
  S, and V are also specified in the range [0,1].
*/
template<class Component, class HComponent>
template<class OComponent>
HSVPixel<Component,HComponent>::HSVPixel(const RGBPixel<OComponent>& other) : Pixel(other)
{
  double r = other.red();
  double g = other.green();
  double b = other.blue();

  double hh=0;
  
  double maxval = std::max(std::max(r,g),b);
  double minval = std::min(std::min(r,g),b);

  /* Value */
  v = maxval;

  /* Saturation */
  s = (maxval == 0.0) ? 0.0 : ((maxval-minval)/maxval);
  
  /* Hue */
  if (s.mag() == 0.0)
    hh = 0;
  else {
    double delt = maxval - minval;
    
    if (r == maxval)
      hh =       (g - b)/delt; /* Between yellow and magenta */
    
    else if (g == maxval)
      hh = 2.0 + (b - r)/delt; /* Between cyan and yellow    */
    
    else if (b  == maxval)
      hh = 4.0 + (r - g)/delt; /* Between magenta and cyan   */
    
    /* Convert to degrees (disabled) */
    /* *h *= 60; if (*h < 0.0) *h += 360; */
    
    /* Scale to Max */
    hh *= (60.0/360.0); if (hh < 0.0) hh = hh+h.Max;
  }
  h = (reversed_color_order? 1.0-hh : hh);
}



/** Averages a series of pixels in HSV space.

  Each hue,saturation pair forms a vector, averaged with others using
  vector addition, while the values are averaged arithmetically.
  PixelType must be an HSVPixel, or at least one supporting that
  interface.
*/
template<class HSVPixelType=HSVPixel<> >
class PixelAverage {
private:
  double valtot, colorx, colory, count;

public:
  PixelAverage() : valtot(0), colorx(0), colory(0), count(0)  { }

  /** Adds the given pixel to the internal totals */
  void operator+=(HSVPixelType p) {
    const double angle = 2*(Math::pi)*p.hue();
    const double mag   = p.saturation();
    colorx += mag * cos(angle);
    colory += mag * sin(angle);
    valtot += p.value();
    count++;
  }
  
  /** Returns the average of the pixels seen so far. */
  HSVPixelType operator() () const {
    const Bounded::Magnitude hue = (colorx? atan2(colory,colorx)/2/(Math::pi) : 0.0);
    const Bounded::Magnitude sat = (count? sqrt(colorx*colorx+colory*colory)/count : 0.0);
    const Bounded::Magnitude val = (count? valtot/count : 0.0);
    return HSVPixel<>(hue,sat,val);
  }
};

    

/** Returns a color which contrasts with the given pixel.

  Presumably one should pass in the average color predominating in
  the area where the new color will be used.  Currently uses only
  the value informtation of the given pixel; picking white for a
  dark color, black for a light color, and blue for ones in between.
  HSVPixelType must be an HSVPixel, or at least one supporting that
  interface.
*/
template<class HSVPixelType>  
RGBPixel<>  contrasting_color( HSVPixelType prevailing_color) {
  const Bounded::Magnitude  val = prevailing_color.value();
  if      (val >= 0.75) return RGBPixel<>(0,0,0); /* bright background */
  else if (val <  0.25) return RGBPixel<>(1,1,1); /* dark background   */
  else                  return RGBPixel<>(0,0,1); /* medium background */
}

  
} /* namespace Plot */
#endif /* __PIXEL_H__ */
