/** \file
  PlotSpecification class.

  $Header: /u/nn/cvsroot/lissom/src/plotspecification.h,v 1.37 2004/05/13 23:59:17 jbednar Exp $
*/

#ifndef __PLOT_SPECIFICATION_H__
#define __PLOT_SPECIFICATION_H__

#include <utility>
#include <string>
using std::string;

// Uses roundabout paths to the current directory to avoid Cray C bug
#include "../src/matriximagefactory.h"
#include "../src/pointermap.h"
#include "../src/allocatingpointer.h"
#include "../src/tnt_gnuplot.h"
#include "../src/msg.h"
#include "../src/parammap.h"
#include "../src/stringutils.h"
#include "../src/matrixhistogram.h"
#include "../src/histogramimage.h"
#include "../src/scale.h"


/* Helper functions, without a real home */ 


/** Helper for getting the values for a color parameter; those are not
    currently supported directly by the ParamMap.  Color should be
    an RGB type. */
template<class Color>
Color lookup_color
(const ParamMap& params, const string& basename, const Color& default_val)
{
  /* Assumes there will be three parameters named with the given
     suffixes, one for each color component. */
  const string R_param = basename+"_R";
  const string G_param = basename+"_G";
  const string B_param = basename+"_B";
  const double R = params.get_with_default(R_param.c_str(), default_val.red());
  const double G = params.get_with_default(G_param.c_str(), default_val.green());
  const double B = params.get_with_default(B_param.c_str(), default_val.blue());
  return Color(R,G,B);
}



/** Helper for getting the values for a rectangle parameter; those are
    not currently supported directly by the ParamMap.  */
template<class SubRegion>
SubRegion lookup_subregion
(const ParamMap& params, const SubRegion& default_val)
{
  const int s_c  = params.get_with_default("start_c", -1);
  const int s_r  = params.get_with_default("start_r", -1);
  const int e_c  = params.get_with_default("end_c",   -1);
  const int e_r  = params.get_with_default("end_r",   -1);

  if (s_c<0 || s_r<0 || e_c<0 || e_r<0)
    return default_val;
  return   SubRegion(-s_c,-s_r,-e_c,-e_r);
}



namespace  Plot {


  
/** Allocates and returns a new ParamMap child of the given
    ParamMap-derived-type parent for use with PlotSpecification.
    Sub-maps are created and filled with values to be used by
    PlotSpecification, as determined by the other arguments here. */
template<class ParamMapType>
ParamMapType* PlotSpec_params(const string& child_name,
			      ParamMapType& parent,
			      const string& strength_name="",
			      const string& color_name="",
			      const string& confidence_name="")
{
  /* The dynamic_cast is required only because new_child has to return
     the base type only (to work with some compilers, e.g. Cray C++) */
  ParamMapType* p  = dynamic_cast<ParamMapType*>(parent.new_child(child_name));
  PlotSpec_child_map(*p,"Strength",strength_name);
  PlotSpec_child_map(*p,"Color",color_name);
  PlotSpec_child_map(*p,"Confidence",confidence_name);
  PlotSpec_child_map(*p,"Histogram","",false);
  return p;
}



/** Helper for PlotSpec_params: allocates an owned child map with
    appropriate parameters set */
template<class ParamMapType>
ParamMapType* PlotSpec_child_map(ParamMapType& parent, const string& childname, const string& matrixname="",
				 bool setmatrixname=true)
{
  ParamMapType* c = dynamic_cast<ParamMapType*>(parent.new_child(childname,true)); assert(c);
  if (setmatrixname) c->set_local("name",matrixname);
  return c;
}


/**
  Complete specification for a particular type of plot, using named
  matrices chosen from a table of matrices passed in.  Allows for
  parameter-based control of plot types.

  At present, two plot devices are supported: PPM image files and
  gnuplot PostScript files.  More can be added here as needed.
*/
template <class Matrix, class String=string, class ScaleType=Scale::Linear<>,
	  class MatrixTableType=PointerMap<Matrix>,
	  class ParameterMapType=ParamMap >
class PlotSpecification {
  typedef const String cStr;
  typedef ParameterMapType ParamMapType;
  typedef Histo::OneDHistogram<> Histogram;
  typedef Plot::RGBPixel<> ColorType;
  
public:
  typedef MatrixTableType  MatrixTable;

private:
  
  /** Constructor accepting the names of one to three Matrices to
      plot.  Params must be allocated using new, as it will be deleted
      on destruction */
  PlotSpecification(cStr& objectname, const MatrixTable& maptable, cStr& name, ParamMapType* params_)
    : plotname(name), objname(objectname), params(params_,true), maps(&maptable) { }

public:
  /** Specification for a rectangular subregion of the NeuralRegion. */
  typedef Boundary::AAArbitraryRectangle<double,int> SubRegion;

  /**
    Returns a map of the size used in the map table, as a reference
    for creating maps to match.  Assumes that one will only want to do
    that for regions already having some other plot; if this
    assumption is false, the map table class can be changed to be
    constructed with a fixed size, to ensure that all maps within are
    that size, and to return a dummy map of that size when requested.
  */
  const Matrix* reference_mapptr() const {
    assert (maps->begin()!=maps->end());
    return maps->begin()->second;
  }
  
  /** Convenience constructor: puts together a nested parameter map for typical case. */
  template <class DerivedFromParamMap>
  PlotSpecification(cStr& objectname, const MatrixTable& maptable, cStr& name, DerivedFromParamMap& parentparams,
		    const string& strength_name="", const string& color_name="", const string& confidence_name="")
    : plotname(name), objname(objectname),
      params(PlotSpec_params(name,parentparams,strength_name,color_name,confidence_name),true),
      maps(&maptable) { }

  ~PlotSpecification() {  }
  
  /** Returns the declared name for this particular plot */
  const string name()       const {  return plotname;  }

  /** Returns the name for the entity which is being plotted, as opposed to the
      name of the particular plot (i.e.\ name()). */
  const string objectname() const {  return objname;  }

  bool                 save_matrix  (const string filebase, const MatrixTable* tempmaps=0,
				     const SubRegion* overriding_region=0) const;
  bool                 gnuplot_image(const string filebase, const MatrixTable* tempmaps=0,
				     const SubRegion* overriding_region=0) const;
  Plot::MatrixImage<>* bitmap_image(const double master_scale=-1, const SubRegion* reference=0, const MatrixTable* tempmaps=0,
				    const SubRegion* overriding_region=0,
				    const int border_width=-1, const string overriding_colorspec="") const;
  Plot::AARImage<>*    bitmap_histogram(const double master_scale, const ParamMap& defaults, const SubRegion* reference,
					const MatrixTable* tempmaps=0, const string& name="Histogram") const;

  
  static void message(Msg::MessageLevel m, const string& s, const bool terminate=True) {
    // Written in this awkward fashion to avoid bug (ICE) in GCC 2.95
    const bool nonnull = (mhp!=0 && (*mhp)); 
    if (nonnull) (**mhp)(m,s,terminate);
  }

  /** If a parameter map is available, looks up the given parameter in it, substituting
      the default value if no parameter found. */
  template <class T>
  T lookup_param(const string& name, const T& default_val) const 
    {  return lookup_param(*params,name.c_str(),default_val);  }
  
  /* Cascading lookup command, with up to two levels of backup maps.
     The routine retrieves the parameter from each of the given maps,
     in order, returning the first value found that differs from the
     given default. If none were found or none differed, the default
     value itself is returned.

     This routine could possibly be used (within the ParameterMap
     class routine get_with_default) as the basis for a
     general-purpose defaults system, with each ParameterMap knowing
     the name of its set of defaults, to which get_with_default turns
     when only a default value is found (e.g. if never set
     explicitly).
  */
  template <class T>
  T lookup_param(const ParamMap& params1, const string& name, const T& default_val,
		 const ParamMap* params2=0, const ParamMap* params3=0) const {
    const T p1=params1.get_with_default(name.c_str(),default_val);
    if (p1!=default_val || !params2) return p1;

    const T p2=params2->get_with_default(name.c_str(),default_val);
    if (p2!=default_val || !params3) return p2;
    
    return     params3->get_with_default(name.c_str(),default_val);
  }
  

  ParamMapType&       param_map()       {  return *(params.valueptr());  }
  const ParamMapType& param_map() const {  return *params;               }
  
private:
  Plot::AARImage<>* histogram_image(const double master_scale,
				    const SubRegion* reference,
				    const ParamMap&  histoparams,
				    const string colorspec,
				    const Histogram& histo) const;
    
  double compute_size_scale(const double master_scale, const double specified_scale,
			    const SubRegion& paramregion, const SubRegion* reference) const;

  /** Looks up the given name in the tempmaps given, if any, and then in the permanent maps */
  const Matrix* map_getptr(const string& name, const MatrixTable* tempmaps=0, bool warn=true) const {
    if (name=="") return 0;
    const Matrix* p  = (tempmaps ? tempmaps->getptr(name) : 0);
    const Matrix* pp = (p? p : maps->getptr(name) );
    if (warn && !pp && name!=lastwarning && name!=previouswarning) {
      message(Msg::Warning, "Could not find map matrix named '" + name
	      + "' when plotting " + plotname + " for " + objname );
      previouswarning=lastwarning;
      lastwarning=name;
    }
    return pp;
  }
  

  /* Convenience macro which looks up the map name under the given
     paramname in the given parameter map */
  const Matrix* map_getptr(const ParamMap& params, const MatrixTable* tempmaps=0,
			   bool warn=true, const string& paramname="name") const
    {  return map_getptr(lookup_param(params, paramname, string()),tempmaps, warn);  }
  
  String plotname, objname;
  
  OwningPointer<ParamMapType> params;
  const MatrixTable* maps;
  
  /** Keep track of previous two warnings to avoid repeating them.  Two should
      be sufficient to cover Color and Confidence; presumably Strength will
      never have an error that's not urgent.  */
  mutable string lastwarning,previouswarning;

  static Msg::LevelHandler<>** mhp;
};


/* Static variables */
template <class Matrix,class String,class ScaleType,class MatrixTableType,class ParameterMapType>
Msg::LevelHandler<>**
PlotSpecification<Matrix,String,ScaleType,MatrixTableType,ParameterMapType>::
mhp=&(Msg::LevelHandler<>::default_instance());



/**
  Create a 3D plot using gnuplot.  Currently only one dimension is
  used, either the strength (if present) or the color (but as a
  height, not a color).  The other dimensions could be plotted in 3D
  if gnuplot supported combined plots (e.g. coloring in each grid
  element in a surface plot), but instead something like Matlab would
  be needed for that.  They could also be combined side-by-side in
  gnuplot.
*/
template <class Matrix,class String,class ScaleType,class MatrixTableType,class ParameterMapType>
bool
PlotSpecification<Matrix,String,ScaleType,MatrixTableType,ParameterMapType>::
gnuplot_image(const string filebase, const MatrixTable* tempmaps,
	      const SubRegion* overriding_region) const
{
  const string  outname = (filebase=="" ? "" : filebase+"_"+objectname()+"_"+name()+".ps");
  const string  title   = objectname() + " " + name();
  const Matrix* s = map_getptr(params->get_child("Strength"), tempmaps);
  const Matrix* c = map_getptr(params->get_child("Color"),    tempmaps);
  const SubRegion paramregion = lookup_subregion(*params,SubRegion());
  const SubRegion subregion   = (overriding_region? *overriding_region : paramregion);
  typedef typename Matrix::value_type T;
  typedef mat::MatrixInterface<T> BaseMatrixType;
  typedef mat::MatrixAdapter<BaseMatrixType> MatrixWindow;
  
  const string messages = (s?  mat::gnuplot(MatrixWindow(*s,1.0,subregion),title,outname) :
			   (c? mat::gnuplot(MatrixWindow(*c,1.0,subregion),title,outname) : string()));

  if (messages != "") {
    message(Msg::Warning, "Gnuplot wrapper for " + title + ": " + messages);
    return false;
  }
  
  return true;
}



/** Function object returning the difference between two given objects of
    identical type, scaled such that if each is in the range 0,1.0,
    the result will also be in the range 0,1.0. */
template <class T>
struct normalized_difference_fnobj : public std::binary_function<T, T, T> {
  T operator()(const T& x, const T& y) const { return 0.5+0.5*(x-y); }
};


/**
  Save the matrix as an ASCII .mat file.  Currently only one dimension is
  used, either the strength (if present) or the color.  
*/
template <class Matrix,class String,class ScaleType,class MatrixTableType,class ParameterMapType>
bool
PlotSpecification<Matrix,String,ScaleType,MatrixTableType,ParameterMapType>::
save_matrix(const string filebase, const MatrixTable* tempmaps,
	    const SubRegion* overriding_region) const
{
  /* Mostly copied from gnuplot_image */
  const string  outname = (filebase=="" ? "" : filebase+".matrix");
  const string  title   = objectname() + " " + name();
  const Matrix* s = map_getptr(params->get_child("Strength"), tempmaps);
  const Matrix* c = map_getptr(params->get_child("Color"),    tempmaps);
  const SubRegion paramregion = lookup_subregion(*params,SubRegion());
  const SubRegion subregion   = (overriding_region? *overriding_region : paramregion);
  typedef typename Matrix::value_type T;
  typedef mat::MatrixInterface<T> BaseMatrixType;
  typedef mat::MatrixAdapter<BaseMatrixType> MatrixWindow;

  if (!c && !s){
    message(Msg::Warning, "Nothing to plot for " + name() );
    return true;
  }

  std::ofstream matrixfile(outname.c_str());
  if (!matrixfile) {
    message(Msg::Warning, "Could not open matrix file " + outname);
    return false;
  }
  MatrixWindow submatrix((s? *s : *c),1.0,subregion);
  
  /* Compute value with respect to reference, if any */
  typedef normalized_difference_fnobj<T>                               RefFunType;
  typedef mat::MatrixInterfaceBinaryAdapter<BaseMatrixType,RefFunType> BinFunAdapter;
  typedef mat::MatrixInterfaceAdapter<BaseMatrixType>                  MatAdapter;
  const BaseMatrixType*   refptr       = map_getptr("__StrengthReference", tempmaps,false);
  const BinFunAdapter*    strength_ref = 0;
  if (s && refptr) {
    strength_ref = new BinFunAdapter(*s,*refptr,RefFunType());
    assert (strength_ref);
    s = strength_ref;
  }

  /* Octave-compatible header */
  matrixfile
    << "# Created from " << title << std::endl
    << "# name: "        << name() << std::endl
    << "# type: matrix"  << std::endl
    << "# rows: "        << submatrix.nrows() << std::endl
    << "# columns: "     << submatrix.ncols() << std::endl;

  stream_output(MatrixWindow((s? *s : *c),1.0,subregion),matrixfile);

  return true;
}



/** If appropriate, automatically computes a size_scale that makes
    region plots similar in size.  Favors integer size scales,
    although if the user has set non-integer scales, the results may
    also be non-integer even though the users' values are only
    multiplied by integer factors. */
template <class Matrix,class String,class ScaleType,class MatrixTableType,class ParameterMapType>
double
PlotSpecification<Matrix,String,ScaleType,MatrixTableType,ParameterMapType>::
compute_size_scale(const double master_scale, const double specified_scale,
		   const SubRegion& paramregion, const SubRegion* reference) const
{
  if (!reference) return specified_scale;

  if (master_scale>0) {
    const double height  = (!paramregion.isinfinite()? paramregion.height() : reference_mapptr()->nrows());
    const double width   = (!paramregion.isinfinite()? paramregion.width()  : reference_mapptr()->ncols());
    const int    hfactor = int(master_scale*reference->height()/height);
    const int    wfactor = int(master_scale*reference->width()/width);
    const int    factor  = std::min(hfactor,wfactor);
    if (factor>0) return factor;
  }
  
  return specified_scale;
}



/** Allocates and returns a bitmap image of the given region using the
    given set of params.  For now, assumes parameter names start with
    ppm_.  If there is a matrix with the special name
    "__StrengthReference" the elements of the strength matrix is
    computed as values relative to that reference.
*/
template <class Matrix,class String,class ScaleType,class MatrixTableType,class ParameterMapType>
Plot::MatrixImage<>*
PlotSpecification<Matrix,String,ScaleType,MatrixTableType,ParameterMapType>::
bitmap_image(const double master_scale, const SubRegion* reference,
	     const MatrixTable* tempmaps, const SubRegion* overriding_region,
	     const int border_width, const string overriding_colorspec) const
{
  /* Get values for parameters (or use specified defaults if none) */
  const Tristate  paper_based = lookup_param(*params,"ppm_paper_based_colors", False);
  const double    subplot_hue = lookup_param(*params,"ppm_subplot_hue",        0.0);
  const string    colorspec   =(overriding_colorspec!=""? overriding_colorspec :
				lookup_param(*params,"colorspec",              string("RYW")));
  const ColorType bordercolor = lookup_color(*params,"ppm_bg",  AARImage<>::default_border);
  const int       border      = (border_width>=0 ? border_width : lookup_param(*params,"ppm_border", 1));
  const ParamMap& strength    = params->get_child("Strength");
  const ParamMap& color       = params->get_child("Color");
  const ParamMap& confidence  = params->get_child("Confidence");
  const SubRegion paramregion = lookup_subregion(*params,SubRegion());
  const SubRegion subregion   = (overriding_region? *overriding_region : paramregion);
  const double    size_scale  = compute_size_scale(master_scale,lookup_param(*params,"size_scale", 1.0),
						   paramregion,reference);

  Plot::ColorLookupFactory<>::result_type* clu =
    Plot::ColorLookupFactory<>(paper_based,subplot_hue)(colorspec);
  typedef typename Matrix::value_type T;
  typedef mat::MatrixInterface<T> BaseMatrixType; 
  const Plot::MatrixImageFactory<BaseMatrixType,T,ScaleType>
    factory(paper_based,size_scale,border,subregion,clu,bordercolor);

  
  /* Compute value with respect to reference, if any */
  typedef normalized_difference_fnobj<T>                               RefFunType;
  typedef mat::MatrixInterfaceBinaryAdapter<BaseMatrixType,RefFunType> BinFunAdapter;
  typedef mat::MatrixInterfaceAdapter<BaseMatrixType>                  MatAdapter;
  const BaseMatrixType*   strengthptr  = map_getptr(strength, tempmaps);
  const BaseMatrixType*   refptr       = map_getptr("__StrengthReference", tempmaps,false);
  const BinFunAdapter*    strength_ref = 0;
  if (strengthptr && refptr) {
    strength_ref = new BinFunAdapter(*strengthptr,*refptr,RefFunType());
    assert (strength_ref);
    strengthptr = strength_ref;
  }

  /* Create image */
  Plot::MatrixImage<>* img =
    factory(strengthptr,
	    ScaleType( lookup_param(strength,   "value_scale", 1.0),
		       lookup_param(strength,   "value_offset",0.0)),
	    map_getptr(color, tempmaps),
	    ScaleType( lookup_param(color,      "value_scale", 1.0),
		       lookup_param(color,      "value_offset",0.0)),
	    map_getptr(confidence, tempmaps),
	    ScaleType( lookup_param(confidence, "value_scale", 1.0),
		       lookup_param(confidence, "value_offset",0.0)) );
  
  delete clu;
  delete strength_ref;
  
  /* Overlay line plots here? */
  //if (show_overlay)
  
  return img;
}


/** Allocates and returns a bitmap histogram of the given region using the given set of params.  */
template <class Matrix,class String,class ScaleType,class MatrixTableType,class ParameterMapType>
Plot::AARImage<>*
PlotSpecification<Matrix,String,ScaleType,MatrixTableType,ParameterMapType>::
bitmap_histogram(const double master_scale, const ParamMap& defaults,
		 const SubRegion* reference, const MatrixTable* tempmaps,
		 const string& name) const
{
  const ParamMap& histoparams = params->get_child(name);
  const bool      enabled     = lookup_param(histoparams,"ppm_histograms",True);
  const int       num_bins    = lookup_param(histoparams,"num_bins",    24  );
  const bool      cyclic      = lookup_param(histoparams,"cyclic",      True);
  const string    hcolorspec  = lookup_param(histoparams,"histo_colorspec",string(""));
  const string    ncolorspec  = lookup_param(histoparams,"colorspec",   string("Hue"));
  const string    colorspec   =(hcolorspec!=""? hcolorspec : ncolorspec);
  const Matrix*   str_mat     = map_getptr(params->get_child("Strength"), tempmaps); 
  const string    spec_bins   = lookup_param(histoparams,"name",string(),&defaults);
  const string    color_name  = lookup_param(params->get_child("Color"),"name",string());
  ScaleType scale(num_bins);
  typedef mat::MatrixInterfaceConstant<float>  ConstantMatrix;
  if (!enabled) return 0;

  /* Three possible types of plots:
     
     1. Fully specified case -- use specified <bin map> for the bin numbers.
  
        The <value map> is the Strength map, or else 1.0 if there is
	no Strength map.  If the specified <bin map> does not exist
	for this plot, no histogram plot is generated.  No error is
	raised in this case, to make it easy to specify a <bin map>
	that is defined only for some of the available plots.

	Should consider removing &defaults and histo_colorspec; that
	convoluted logic may no longer be necessary and is only used
	for this case.  Presumably, the other two cases now account
	for the conditions where one would have previously set
	::PlotGroup::histogram::name or histo_colorspec globally, and
	for bizarre cases the individual ...::Histogram::name and
	...::Histogram::colorspec parameters should allow maximum
	flexibility to customize individual plots.  The advantage of
	the other two cases over &defaults and histo_colorspec is that
	they help ensure that histograms match the main plot as
	closely as possible, at least by default.  Plus, they are much
	simpler and more natural to reason about.
  */
  if (spec_bins!="") {
    const Matrix* bin_mat = map_getptr(spec_bins, tempmaps, false);
    if (!bin_mat) return 0;
    if (str_mat)
      return histogram_image(master_scale, reference, histoparams, colorspec,
			     Histogram(Histo::matrix_histogram(*bin_mat,*str_mat, num_bins,scale,cyclic)));
    else {
      ConstantMatrix const_one(1.0,bin_mat->nrows(),bin_mat->ncols());
      return histogram_image(master_scale, reference, histoparams, colorspec,
			     Histogram(Histo::matrix_histogram(*bin_mat,const_one, num_bins,scale,cyclic)));
    }
  }

  /* 2. If a color subplot exists, take that to be the <bin map>.

        In this case, the histogram is assumed to be cyclic and
	the colorspec is assumed to be Hue.
  */
  else if (color_name!="") {
    const Matrix* bin_mat = map_getptr(color_name, tempmaps);
    if (!bin_mat) return 0;
    if (str_mat)
      return histogram_image(master_scale, reference, histoparams, "Hue",
			     Histogram(Histo::matrix_histogram(*bin_mat,*str_mat, num_bins,scale,true)));
    else {
      ConstantMatrix const_one(1.0,bin_mat->nrows(),bin_mat->ncols());
      return histogram_image(master_scale, reference, histoparams, "Hue",
			     Histogram(Histo::matrix_histogram(*bin_mat,const_one, num_bins,scale,true)));
    }
  }
  /* 3. Default to histogram of Strength matrix, assuming cyclic=false and
        the normal colorspec of the plot. */
  else {
    if (!str_mat) return 0; /* Probably means there's no regular plot either */
    ConstantMatrix const_one(1.0,str_mat->nrows(),str_mat->ncols());
    return histogram_image(master_scale, reference, histoparams, ncolorspec,
			   Histogram(Histo::matrix_histogram(*str_mat,const_one, num_bins,scale,false)));
  }
}



template <class Matrix,class String,class ScaleType,class MatrixTableType,class ParameterMapType>
Plot::AARImage<>* 
PlotSpecification<Matrix,String,ScaleType,MatrixTableType,ParameterMapType>::
histogram_image(const double master_scale, const SubRegion* reference,
		const ParamMap& histoparams, const string colorspec,
		const Histogram& histo) const
{
  const ColorType bordercolor   = lookup_color(histoparams,"ppm_bg", AARImage<>::default_border);
  const double    range         = lookup_param(histoparams,"range_end",   0.25);
  const bool      vertical      = lookup_param(histoparams,"vertical",    True);
  const double    aspect_ratio  = lookup_param(histoparams,"aspect_ratio",0.25);
  const int       border        = lookup_param(histoparams,"ppm_border",  1   );
  const double    subplot_hue   = lookup_param(histoparams,"ppm_subplot_hue", 0.0);
  const Tristate  paper_based   = lookup_param(histoparams,"ppm_paper_based_colors", False);
  const ColorType histobg       = lookup_color(histoparams,"ppm_line",    ColorType());
  const bool      outlinehist	= lookup_param(histoparams,"histogram_outline", False);
  const double    psize_scale   = lookup_param(*params,"size_scale", 1.0);
  const SubRegion fakesubregion = SubRegion(0,0,
					    histo.num_bins()*(vertical? aspect_ratio : 1.0),
					    histo.num_bins()*(vertical? 1.0 : aspect_ratio));
  const double    size_scale    = compute_size_scale(master_scale,psize_scale,fakesubregion,reference);
  const string    title         = objectname() + "::" + name();
  
  OwningPointer<Plot::ColorLookupFactory<>::result_type> clu
    (Plot::ColorLookupFactory<>(paper_based,subplot_hue)(colorspec),true);

  return new
    Plot::OneDHistogramImage<>(histo,vertical,size_scale,aspect_ratio,
			       range,border,*clu,bordercolor,histobg,bordercolor,outlinehist,title);
}


} /* namespace Plot */
#endif /* __PLOT_SPECIFICATION_H__ */
