/** \file
  Input routines for PNM, PPM, PGM, and PBM files.

  $Header: /u/nn/cvsroot/lissom/src/pnm_io.h,v 1.18 2002/03/23 20:02:50 jbednar Exp $
*/

#ifndef __PNM_IO_H__
#define __PNM_IO_H__

#include <string>
using std::string;
#include <fstream>
#include <utility> /* For pair type */
#ifdef _CRAYMPP
#include <pair.h>
#endif

/* Optional support for .pbm.gz and .pgm.gz files */
#ifdef USE_ZLIB
#include "gzstream.h"
#define IFSTREAM igzstream
#else
#define IFSTREAM std::ifstream
#endif


/**
  Swallows comments from the given stream, where a comment
  is defined as a line whose first non-whitespace character
  is a pound sign (#).
*/
inline std::istream& eatcomments(std::istream& is)
{
  char c;
  is >> c;
  while (c=='#') {
    while (is.get()!='\n'); /* Swallow rest of line */
    is >> c;
  }
  is.putback(c);
  
  return is;
}


/**
  Read the specified PBM or PGM file into the given matrix in
  column-major order.  Any 2D array type which supports a
  (width,height) constructor should work, e.g. TNT's Matrix.

  Returns false if there was a problem reading the file.
*/
template <class Array2D>
bool pgm_read(const string& filename, Array2D& image)
{
  IFSTREAM file(filename.c_str());
  if (!file) return false;
  
  string filetype; file >> filetype >> eatcomments;
  const bool is_greyscale = (filetype=="P2" || filetype=="P5");
  const bool is_bitmap    = (filetype=="P1" || filetype=="P4");
  if (!(is_greyscale || is_bitmap)) return false;
  
  int width;      file >> width  >> eatcomments;
  int height;     file >> height >> eatcomments;
  double maxval=1;
  if (!is_bitmap) file >> maxval;
      
  if (!maxval) return false;
  
  image=Array2D(height,width); /* Resize */

  /* ASCII text bitmap */
  if (filetype=="P1") {
    file >> eatcomments; /* Discard comments */
    for (int i=0; i<height; i++) {
      for (int j=0; j<width; j++) {
	int val;
	file >> val;
	image[i][j] = (val? 0.0 : 1.0);
      }
      file >> eatcomments;
    }
  }

  /* ASCII text grayscale */
  if (filetype=="P2") {
    file >> eatcomments; /* Discard comments */
    for (int i=0; i<height; i++) {
      for (int j=0; j<width; j++) {
	int val;
	file >> val;
	image[i][j] = val/maxval;
      }
      file >> eatcomments;
    }
  }

  /* Raw binary bitmap. */
  else if (filetype=="P4") {
    //file.get(); /* Discard newline */
    for (int i=0; i<height; i++) {
      int b=8; /* Bit position */
      char bitbuffer=0;
      for (int j=0; j<width; j++, b++) {
	if (b>7) {
	  b=0;
	  file.get(bitbuffer);
	}
	const unsigned char mask = 128>>b;
	const bool bit           = ((unsigned char)bitbuffer) & mask;
	image[i][j] = bit? 0.0 : 1.0;
      }
    }
  }

  /* Raw binary grayscale */
  else if (filetype=="P5") {
    file.get(); /* Discard newline */
    for (int i=0; i<height; i++) {
      for (int j=0; j<width; j++) {
	char val;
	file.get(val);
	image[i][j] = ((unsigned char)val)/maxval;
      }
    }
  }

  return true;
}



/**
  Utility routine which returns the size of the given PNM file
  as a (width,height) if found and readable, or else (0,0).
*/
inline std::pair<int,int> pnm_size(const string& filename)
{
  IFSTREAM file(filename.c_str());
  if (!file) return std::make_pair(0,0);
  
  string filetype; file >> filetype >> eatcomments;
  if (filetype[0]!='P') return std::make_pair(0,0);

  int width;     file >> width  >> eatcomments;
  int height;    file >> height >> eatcomments;
  return std::make_pair(width,height);
}


/**
  Utility routine which returns true if the file declares that
  white (i.e. 255/255) pixels are to be considered transparent.
  
  The PNM specs do not provide for transparency, but as a convention I
  assume that if the file has a comment (all on one line) of the form:
  
  #Transparent: 255

  then then I should consider it so.  The comment must appear
  within the first two non-comment lines to be detected, there must
  be no space after the #, and there must be one space after the :.

  Only a value of 255 (pure white) is supported at present.  The
  reason is that pgm_read returns floats, not discrete integers, so
  it's more complicated to check any pixel value other than pure white
  or black.  (One would have to check a range half a pixel-value-step
  larger and smaller in each direction.)  In any case, in a
  non-colormapped grayscale picture, only black and white are likely
  to be useful as transparent.  (Otherwise there would be a big step
  in grayscale gradients somewhere.)
*/
inline bool pgm_white_is_transparent(const string& filename)
{
  IFSTREAM file(filename.c_str());
  if (!file) return false;

  int non_comment_lines=0;
  while (!file.eof()) {
    char c;
    file >> c;

    /* Examine comment line */
    if (c=='#') {
      string line;
      c=file.get();
      while (c!='\n') {
	/* Ignores trailing whitespace: */
	if (line.length()<=15) line+=c;
	c=file.get();
      }
      if (line=="Transparent: 255")
	return true;
    }
    
    /* Swallow non_comment_line */
    else {
      non_comment_lines++;
      if (non_comment_lines>2) return false;
      while (file.get()!='\n' && !file.eof());
    }
  }

  return false;
}

#endif
