/** \file
  Map of pointers to objects allocated externally but owned by the map.

  $Header: /u/nn/cvsroot/lissom/src/pointermap.h,v 1.11 2004/08/24 13:55:35 jbednar Exp $
*/

#ifndef __POINTERMAP_H__
#define __POINTERMAP_H__

#include <cassert>
#include <map>
#include <string>
using std::string;
#include <algorithm>
#include <functional>

#include "generic_stdlib.h"
#include "pointerlookup.h"


/**
  This is provided for use with objects which want to maintain a
  simple database of named objects.  By default, the objects are
  assumed to be allocated on the heap and their memory management is
  handled by the database; if instead you wish to manage the memory
  yourself (e.g. for objects existing in more than one map) supply
  false for the owns_objects template parameter.
*/
template<class T, class K=string, bool owns_objects=true>
class PointerMap : public PointerLookup<T,K> {
public:
  /** Type of this class */
  typedef PointerMap<T,K,owns_objects> self;
  
  /** Key type of our local map. */
  typedef K  key_type;

  /** Data element type of our local map. */
  typedef T* data_type;

  /** Element type of our local map. */
  typedef std::pair<const key_type, data_type> value_type;

  /** Type of our local map. */
  typedef std::map< key_type, data_type > map_type;

  /** Map's iterator. */
  typedef typename map_type::iterator iterator;

  /** Map's const_iterator. */
  typedef typename map_type::const_iterator const_iterator;  

  /** Map's size_type */
  typedef typename map_type::size_type size_type;

  
  PointerMap() { }
  virtual ~PointerMap() {  erase();  }

  /** Copy construction and assignment are unimplemented; implementing
      them wouldn't be particularly difficult but it would require
      assuming a clone operator for the datatype. */
  PointerMap(const self& o)       { assert(0); abort(); }
  self& operator=( const self& o) { assert(0); abort(); }
  
  /** Pass most things right through to the std::map */
  iterator       begin()       {  return m.begin();  } /**< Underlying map's begin().*/
  const_iterator begin() const {  return m.begin();  } /**< Underlying map's begin().*/
  iterator       end()         {  return m.end();    } /**< Underlying map's end().*/
  const_iterator end()   const {  return m.end();    } /**< Underlying map's end().*/
  /** Insert into underlying map */
  std::pair<iterator,bool> insert(const value_type& x)  {  return m.insert(x);  }  

  /**
    Instead of directly using the map's [], which has a tricky
    semantics, or find, which is cumbersome, map access is handled
    through named get and set member functions.  It is illegal to
    supply an invalid pointer to set, since this class will handle
    the deallocation of the object pointed to.
  */
  void set(const key_type& k, const data_type d) {
    iterator existing = m.find(k);
    if (existing == end())
      insert(Generic::make_pair_leftconst(k,d));
    else {
      if (owns_objects) delete (*existing).second;
      (*existing).second=d;
    }
  }

  
  /**
    Similar to map's [], but if the key is not present just returns a
    null pointer without adding a new entry.
  */
  T* getptr(const key_type& k) {
    const_iterator existing = m.find(k);
    return (existing == end() ? 0 : (*existing).second);
  }

  /** Const version of getptr */
  const T* getptr(const key_type& k) const {
    const_iterator existing = m.find(k);
    return (existing == end() ? 0 : (*existing).second);
  }

    
  //  virtual T* clone() const { return new PointerMap(*this); };

  /** Remove all items from the map */
  void erase() {
    if (owns_objects) 
      Generic::delete_contents(m);
    else
      m.erase(m.begin(),m.end());
  }

  /** Remove the item with the given key, if any, from the map */
  void erase(const key_type& k) {
    T* ptr = getptr(k);
    if (ptr) {
      if (owns_objects) delete ptr;
      m.erase(k);
    }
  }

  /** Size of the underlying map */
  size_type size() {  return m.size();  }
  
protected:
  /** Underlying container. */
  map_type m;
};




/** Mixin class for something that owns a simple publicly-accessible
    table of named items. */
template <class BaseType, class TableType=PointerMap<BaseType> >
class HasTable {
public:
  typedef TableType Table;
  typedef typename Table::key_type key_type;
  
  /** Add the given item to the table of available items, overwriting
      any old one with that name. The pointer should be allocated
      using new, since it will be deleted when this object is
      destroyed (depending on the underlying TableType.) */ 
  void table_set(const key_type& name, BaseType* i) {
    assert (i);  assert (name!=key_type());
    table_.set(name,i);
  }

  const Table& table() const {  return table_;  }

private:
  /** Table of named items which can be requested for this object */
  Table table_;
};


#endif /* __POINTERMAP_H__ */
