/** \file
  Polymorph class.
  
  $Header: /u/nn/cvsroot/lissom/src/polymorph.h,v 1.13 2001/04/07 08:15:30 jbednar Exp $
*/

#ifndef ___POLYMORPH_H___
#define ___POLYMORPH_H___

#include <cstring>
#include <string>
using std::string;
#include <typeinfo>
#include <cassert>

#include "../src/stringutils.h"
#include "../src/stringparser.h"
#include "../src/allocatingpointer.h"
#include "../src/tristate.h"
#include "../src/typeless.h"


#ifdef ANSI_COMPATIBLE
/* Optional: defines LONG_DOUBLE_UNAVAILABLE where needed */
#include "../src/ind_types.h"
#endif



/* Forward declaration */
template <class T>  class Polymorph;


/** Wrapper to hide the type of an enclosed variable.
    This serves as a base type to Polymorph, to allow
    specializations of Polymorph to override some of
    this functionality.
*/
template <class T>
class PolymorphBase : public Typeless {
public:
  typedef T value_type;
  typedef  PolymorphBase<value_type> self;
  virtual ~PolymorphBase() { }

  PolymorphBase(const value_type  val) : p(val) {  };
  PolymorphBase(      value_type* ptr) : p(ptr) {  };

  virtual bool operator<  (const Typeless&   o) const;
  virtual bool operator>  (const Typeless&   o) const;

  bool operator<  (const self&       o) const {  return (*p < *(o.p));  }
  bool operator>  (const self&       o) const {  return (*p > *(o.p));  }

  bool operator<  (const value_type& o) const {  return (*p <   o   );  }
  bool operator>  (const value_type& o) const {  return (*p >   o   );  }

  self&     operator= (const value_type& o) {  p=o;  return *this;  }
  Typeless& set(const Typeless&   o);
  
  const value_type& operator* () const {  return *p;  }
  
  /** Returns a pointer to the underlying value; use with
      caution. */
  value_type* valueptr() const {  return p.valueptr();  }

  virtual string stringrep(String::StreamFormat format=String::StreamFormat()) const
    {  return String::stringrep(*p,format);  }
  
  virtual string typestring() const
    {  return (name_str? string(name_str) : string(typeid(T).name()));  }

  /** By default, typestring() returns the value of typeid().name()
      for a given type, which is implementation-dependent; for
      templates such as strings it is often very ugly.  If the
      typestring is shown to the user, you can use this call to
      override the default with something prettier or possibly
      something domain-specific. */
  static void set_typestring(const char* str) {  name_str=str;  }

  /** Returns the result of parsing the given string with the given parser.
      The actual type of the result is a Polymorph<value_type>, which is
      a derived type of this one.  (Otherwise each derived specialization
      would need to replicate this routine, changing only the returned
      value's type.)
   */
  virtual Typeless* parse(const StringParser& parser, const string& s) const {
    value_type result;
    (void)parser; (void)s;
    parser.parse(s,result);
    return new Polymorph<value_type>(result); 
  }

  /** Constructs an identical copy of self and returns it.  Due to
      Cray C++'s variance from the C++ standard, cannot use actual
      type as the return value (or it wouldn't match the protype in
      Typeless.) */
  virtual Typeless* clone() const { return new self(*this); };

  virtual Typeless* duplicate() const { return new self(*p); };

protected:
  AllocatingPointer<value_type> p;
  
  static const char* name_str;
};



/* Static variables */
template <class T>
const char* PolymorphBase<T>::name_str=0;



/** Polymorphic comparison operator */
template <class T>
bool PolymorphBase<T>::operator<  (const Typeless&   o) const
{
  /* First try to match exact type */
  const self* oself = dynamic_cast<const self*>(&o);
  if (oself) return (*this < **oself);
  
  /* Fall back to numeric comparison if both are numbers */
  if (isnumeric() && o.isnumeric())
    return (numericvalue() < o.numericvalue());
  
  assert(0); /* Unsupported comparison */
  exit(-1);
  return true;
}



/** Polymorphic comparison operator */
template <class T>
bool PolymorphBase<T>::operator>  (const Typeless&   o) const
{
  /* First try to match exact type */
  const self* oself = dynamic_cast<const self*>(&o);
  if (oself) return (*this > **oself);
  
  /* Fall back to numeric comparison if both are numbers */
  if (isnumeric() && o.isnumeric())
    return (numericvalue() > o.numericvalue());
  
  assert(0); /* Unsupported comparison */
  exit(-1);
  return true;
}



/**
  Polymorphic assignment operator.  This version accepts identical
  types only.  One could imagine supporting arbitrary types by
  printing the o value to a string and then parsing it into our type,
  but that would be pretty kludgy.  In any case, the aborting
  assertion here should probably be eliminated somehow.
*/
template <class T>
Typeless& PolymorphBase<T>::set(const Typeless&   o)
{
  const self* oself = dynamic_cast<const self*>(&o);
  if (oself) p = *(oself->p);
  else { assert(0); } // Unsupported assignment; should be an exception instead
  
  return *this;
}



/** Generic implementation of a polymorphic object;
    no support for numericvalue(). */
template <class T>
class Polymorph : public PolymorphBase<T> {
public:
  typedef T value_type;
  typedef  Polymorph<value_type> self;
  virtual ~Polymorph() { }
  Polymorph(const value_type  val) : PolymorphBase<value_type>(val) {  };
  Polymorph(      value_type* ptr) : PolymorphBase<value_type>(ptr) {  };
  virtual Typeless* clone()     const { return new self(*this); };
  virtual Typeless* duplicate() const { return new self(*p);    };  
};


/** Returns a polymorph of the appropriate type for the given pointer argument.
    This can be used to avoid having to specify the specific template
    argument to determine the Polymorph type.
*/
template <class T>
Polymorph<T> make_Polymorph(T* ptr)
{  return Polymorph<T>(ptr);  }



/** Returns a polymorph of the appropriate type with an internal copy
    of the given non-pointer argument. */
template <class T>
Polymorph<T> allocating_Polymorph(const T& val)
{  return Polymorph<T>(val);  }


/** Specialization for numeric types.

  This implementation adds numericvalue() support to the given type.
  It would be cleaner to have numericvalue() as a template function,
  but then it couldn't be virtual, and also we need to specify that
  only numeric types can use this specialization.  Thus here we
  explicitly specialize Polymorph for the typical numeric types.

  If you use other numeric types, just add the appropriate
  "NumericPolymorph(type);" line before you use them (e.g. just after
  including this file) and numericvalue() should magically start to
  work for them.

  Note that the prettyprint implementation assumes that you will want
  all values which happen to be equal to Tristate's Uninitialized to
  print as "Uninitialized" for all numeric types, which is currently
  useful but may not always be so.
*/
#define NumericPolymorph(type)\
template <> class Polymorph<type> : public PolymorphBase<type> {\
public:\
  typedef type value_type;\
  typedef  Polymorph<value_type> self;\
  virtual ~Polymorph() { }\
  Polymorph(value_type  val) : PolymorphBase<value_type>(val) {  };\
  Polymorph(value_type* ptr) : PolymorphBase<value_type>(ptr) {  };\
  virtual Typeless* clone()     const { return new self(*this); };\
  virtual Typeless* duplicate() const { return new self(*p);    };\
  virtual bool  isnumeric() const {  return true;  }\
  virtual Typeless::exprtype numericvalue() const\
    {  return Typeless::exprtype(*p);  }\
  virtual string prettyprint(String::StreamFormat format=String::StreamFormat()) const\
    {  return (*p==Uninitialized ? std::string("Uninitialized") : stringrep(format));  }\
}

NumericPolymorph(int);
NumericPolymorph(double);
NumericPolymorph(Tristate);



/** Specialization for string types */
template <> class Polymorph<string> : public PolymorphBase<string> {
public:
  typedef string value_type;
  typedef  Polymorph<value_type> self;
  virtual ~Polymorph() { }
  Polymorph(value_type  val) : PolymorphBase<value_type>(val) {  };
  Polymorph(value_type* ptr) : PolymorphBase<value_type>(ptr) {  };
  virtual Typeless* clone()     const { return new self(*this); };
  virtual Typeless* duplicate() const { return new self(*p);    };
  virtual string prettyprint(String::StreamFormat format=String::StreamFormat()) const\
    {  return ("\"" + stringrep(format) + "\"");  }
};


#endif /* ___POLYMORPH_H___ */
