/** \file
  One-dimensional radial functions.

  $Header: /u/nn/cvsroot/lissom/src/radialfunction.h,v 1.12 2003/03/29 21:45:17 jbednar Exp $
*/

#ifndef __RADIAL_FUNCTION_H__
#define __RADIAL_FUNCTION_H__

#include <cmath>
#include <string>
using std::string;
  
#include "shuffledrand.h"
#include "mathconstants.h"


/**
  This family of generic routines return the value of a mathematical
  function parameterized by a sigma (half-width), evaluated at a given
  radius.  For convenience, both parameters are supplied as squares,
  since the underlying functions are almost always specified in terms
  of the squares.
  
  The matrix routine will evaluate one of these functions (or another
  you supply with the same parameter list and semantics) at discrete
  values over a certain range, and return the result as a matrix.
  
  A similar set of PlanarFunctions could also be provided, taking
  two half-width parameters and two position parameters.
*/
namespace RadialFunction {
  using Math::pi;

  /** Returns a constant value regardless of radius. */
  template<class T>
  inline T Constant( T, T )
    {  return 1.0;  }

  /** Returns a uniformly-random value in the range [0,1.0) regardless of radius. */
  template<class T>
  inline T Random( T, T )
  {  return shuffled_rand();  }

  /** Returns the value of a Gaussian at the given radius. */
  template<class T>
  inline T Gaussian( T x_sq, T sig_sq)
    {  return exp( -x_sq/(2*pi*sig_sq) );  }

  /** Returns the value of a Cauchy distribution at the given radius.
      1/(s*PI*(1+x^2/s^2))
  */
  template<class T>
  inline T Cauchy(T x_sq, T s_sq)
    {  return 1/(sqrt(s_sq)*pi*(1+x_sq/s_sq));  }

  /** Returns the Laplacian of a Gaussian, which is an operator from
  cite{marr:vision} and, in turn, from cite{marr:prslb80}.  It does
  not currently have the intended effect, perhaps because of
  normalization.
  */
  template<class T>
  inline T LoG( T x_sq, T sig_sq) { 
    const T exponent = -x_sq/(2*sig_sq);
    return 1/(pi*sig_sq*sig_sq)*(1.0+exponent) * exp(exponent);
  }

  
  /** Generates a square matrix of values given a function, params,
      and bounds.  The size of the matrix is guaranteed to be at least
      1x1, which for a normalized matrix with max_radius==0 will be
      the identity function for convolution. */
  template<class matrix_type, class radial_function, class radius_type>
    matrix_type matrix
    (/* Function pointer or function object of the form T fn(T,T) */
     radial_function radial_fn,
     radius_type radius,
     radius_type max_radius_,
     string& errors,
     bool circular=true,
     bool normalize=true
     )
  {
    typedef typename matrix_type::value_type value_type;
    typedef typename matrix_type::size_type  size_type;
    
    const radius_type radius_sq      = radius*radius;
    /* Uses floating-point max_radius for circular radius, integer for other tasks */
    const radius_type max_radius_sq  = max_radius_*max_radius_;
    const size_type   max_radius     = size_type(floor((0.5+(max_radius_))));
    const size_type   kernelwidth    = std::max(size_type(1),size_type(max_radius*2+1));
    matrix_type kernel(kernelwidth,kernelwidth);
    
    /* Cropping is counterproductive for very small radii */
    if (max_radius<=1) circular=false;
    
    /* Create kernel matrix */
    value_type kernelsum=0;
    /*   ki and kj cannot be size_type since that might be unsigned */
    int ki = -max_radius;
    for (size_type i=0; i < kernelwidth; ki++,i++) {
      int kj = -max_radius;
      for (size_type j=0; j < kernelwidth; kj++,j++) {
  	const size_type	ki_sq = ki*ki;
  	const size_type	kj_sq = kj*kj;
  	const size_type	r_sq  = ki_sq + kj_sq;
  	
  	if (circular && (r_sq > max_radius_sq))
  	  kernel[i][j] = 0;
  	else {
  	  const value_type val = (radial_fn)(r_sq,radius_sq);
  	  kernelsum += val;
  	  kernel[i][j] = val;
  	}
      }
    }
    //cout << "Unnormalized sum of blurring kernel values is " <<kernelsum;
    
    /* Normalize kernel matrix */
    if (normalize) {
      if (kernelsum==0)
  	errors += "Can't normalize a kernel that sums to zero";
      else
  	kernel *= 1/kernelsum;
    }
    
    return kernel;
  }

} /* namespace */


#endif /* __RADIAL_FUNCTION_H__ */
