/** \file
  Retina class and associated functions.

  $Header: /u/nn/cvsroot/lissom/src/retina.h,v 1.38 2002/06/19 04:21:42 yschoe Exp $
*/

#ifndef __RETINA_H__
#define __RETINA_H__

#include <assert.h>
#include <algorithm>

#include "genericalgs.h"
#include "robj.h"
#include "neuralregion.h"
#include "matrix.h"
#include "allocatingpointer.h"



/******************************************************************************/
/* General-purpose utility functions                                          */
/******************************************************************************/

/**
  Generic Retinal_Obj draw function into a matrix.

  The integer eye matrix coordinates (r,c) are first mapped into
  real-valued world coordinates (x,y).  This requires flipping the
  vertical dimension (since arrays are numbered from the top down yet
  the Y axis increases from the bottom up) and sampling at the center
  of the pixel covering (x,y) to (x+1,y+1), which requires an offset
  of (0.5,0.5).  Nothing is drawn if the given object pointer is
  null; such a condition is generally but not always an error.
*/
template <class Matrix>
void draw(Matrix& area, const Retinal_Obj* obj,
	  const typename Matrix::value_type xoff=0.5,
	  const typename Matrix::value_type yoff=0.5,
	  const typename Matrix::value_type size_scale=1.0)
{
  if (!obj) return;
  
  typedef typename Matrix::value_type T;
  typedef typename Matrix::size_type  Subscript;

  obj->update();
  const T s=1/size_scale;
  const T xo=xoff*s;
  const T yo=yoff*s;
  for (Subscript r=0; r<area.nrows(); r++) {
    const T y   = (area.nrows()-r-1)*s+yo; 
    for (Subscript c=0; c<area.ncols(); c++) {
      const T x   = c*s+xo;
      const T act = (obj? obj->activation(x,y) : 0);
      area[r][c]=act;
    }
  }
  //cout << gnuplot(area, "Area after draw") << endl;
}



/******************************************************************************/
/* Retina classes                                                             */
/******************************************************************************/


/** Basic retina implemented as a transparent window into the world. */
class Retina : public NeuralRegion {
public:
  Retina(string name_i, Dimensions dims,
	 ActivationFunction* actfn_=0,
	 Activity size_scale_=1.0)
    : NeuralRegion(name_i, dims), world_model(0), actfn(actfn_),
      xo(dims.xoffset), yo(dims.yoffset), size_scale(size_scale_) { }
  
  virtual ~Retina() { };

  virtual void activate(bool=false,bool=false,bool=false) {
    draw(output, world_model, xo, yo, size_scale);
    if (actfn) (*actfn)(output,output);
  }

  /** Make the retina point to a particular view of the world. */
  void set_input(const Retinal_Obj* const obj) {  world_model=obj;  };

  virtual bool is_plastic() const  {  return false;  }
  
  double size_connection_bytes()   const {  return 0;  }
  double size_unique_connections() const {  return 0;  }

  /* virtual void backproject() { } */

protected:
  /** Most-recently-selected world view */
  const Retinal_Obj* world_model;
  /** Activation function, generally at most thresholding. */
  const OwningPointer<ActivationFunction> actfn;
  /** Offset relative to the origin; this allows drawing to be aligned properly
      even when retina has been expanded as a buffer for later regions. */
  const Activity xo,yo; /**< See xo. */
  /** Size scale relative to ther previous region; this allows drawing to be
      sized properly even when retina is of a different magnification than
      later regions. */
  const Activity size_scale;
};


#endif
