/** \file
  Retinal_Object and derived classes: draw 2D greyscale objects in a plane.

  $Header: /u/nn/cvsroot/lissom/src/retinalobjs.h,v 1.65 2002/01/19 07:01:50 jbednar Exp $
*/

#ifndef __RETINALOBJS_H__
#define __RETINALOBJS_H__

#include <vector>
#include <algorithm>
#include <string>
using std::string;

#include "matrix.h"

#include "genericalgs.h"
#include "valuegen.h"
#include "valuegenmap.h"
#include "pnm_io.h"
#include "boundingbox.h"
#include "robj.h"
#include "tristate.h"
#include "mathconstants.h"


/******************************************************************************/
/*  Global defines                                                            */
/******************************************************************************/

#ifndef RADIANS_TO_DEGREES
#define RADIANS_TO_DEGREES(theta) (180.0*(theta)/(Math::pi))
#endif

/** Prefer Generic::wrap() to this macro. */
#ifndef CONSTRAIN_ANGLE
#define CONSTRAIN_ANGLE(angle)                                    \
  (((angle)>=0) ?                                                  \
   fmod( (angle),                                  (Math::pi) ) :   \
   fmod( (angle) + 2*(Math::pi)*(-1*floor((angle)/((Math::pi)*2))+1),  (Math::pi))     )
#endif



/******************************************************************************/
/* Retinal_Object class                                                       */
/******************************************************************************/
class Retinal_Composite;

/**
  Semi-abstract base class: Retinal_Obj with a name, intensity scale,
  intensity offset, center X and Y position, and angle which can be
  enabled and disabled on command.  The interface is very fat,
  supplying many operations that have proven useful in practice, but
  which are not strictly required.

  Although it is abstract and cannot be instantiated, this class is
  actually fairly concrete, supplying implementations for nearly all
  operations.  A derived class need only provide activ() and clone()
  to be complete.  Most will also override update() to handle values
  needed to be cached for activ() to use quickly, and some may
  override other virtual functions.
  
  The parameters are primarily ValueGenerator types; these are all
  passed in a map for convenience in object creation and to make it
  easier to manage next() and reset() of those variables.  The current
  values of those variables may be cached internally by a call to
  update().  To ensure the coherence of this cache, update() should be
  called before calling activation(), after any ValueGenerator might
  have changed its value.  Ordinarily this is achieved by calling
  update() just before a series of calls to activation().
  
  No assignment operator is defined yet.
*/
class Retinal_Object : public Retinal_Obj {
protected:
  /** Concrete type of bounding box to use. */
  typedef AARBoundingBox<Coordinate> BBox;

public:
  /** Underlying type for a ValueGenerator variable. */
  typedef double Variable;
  /** Type of ValueGenerator to use for variables */
  typedef ValueGenerator<Variable> VarGen;
  /** Type of ValueGenerator map to use to hold variables */
  typedef ValueGeneratorMap<Variable> VarMap;

  /** \name    Constructors and destructors */
  //@{
  Retinal_Object(const string& name_val, const VarMap& init_vars=VarMap())
    : bounding_box(), active(true), name_(name_val),
      default_act(0.0), parentptr(0) {
    vars.merge(init_vars);
    set_var_default("scale",1.0);
  }

  Retinal_Object(const Retinal_Object& b) /* Copy constructor */
    : Retinal_Obj(b), bounding_box(b.bounding_box), vars(b.vars),  
      active(b.active), name_(b.name_),
      default_act(b.default_act), parentptr(0) { }

  virtual ~Retinal_Object() { }
  //@}
  
  
  /** \name    Primary Interface

    If a derived class overrides one of these virtual functions,
    it should call this base-class version when it is done.
  */
  //@{
  virtual bool next()  {  return vars.next();  }
  virtual void reset() {  vars.reset();  }
  virtual bool update() const  {
    /* Cache the result of sub-computations */    
    theta       = get_var("theta");
    cx          = get_var("cx");
    cy          = get_var("cy");
    scale       = get_var("scale");
    offset      = get_var("offset");
    default_act = default_activ();
    
    return true;
  }

  /** Returns the activation at a given set of coordinates.

    Must call update() just before this, though multiple
    activation() calls in a row are fine.
  */
  virtual Activity activation(Coordinate x, Coordinate y) const {
    return offset + scale *
      ((!active || !bounding_box.inside(x,y)) ? default_act : activ(x,y)); 
  }
  //@}


  /** \name    Other functions which may be useful */
  //@{      

  virtual string stringrep() const;
  
  /** Return an identical copy of this object allocated from the heap. */
  virtual Retinal_Object* clone() const =0;

  /** Replace all Variables with those in the specified object. */
  virtual void relink (const Retinal_Object* master)
    {  if (master) vars.relink(&(master->vars));  }

  /** Query whether the object is currently displayed. */
  bool get_active() const { return active; }
  /** Set whether the object is currently displayed. */
  void set_active(bool whether_active) {  active=whether_active;  }

  /** Query name string. */
  const string& name()     const                {  return name_;   }
  const string& get_name() const                {  return name_;   } // Deprecated
  /** Set name string. */
  void          set_name(const string& newname) {  name_=newname;  }

  /** Query a parameter. */
  const Variable get_var(const string& varname) const {  return vars.get(varname);  }

  /** Deprecated: get a pointer to the actual parameter storage location. */
  Variable* get_varptr(const string& varname) {  return vars.getptr(varname);  }

  /** Returns a string representation of the specified VarGen, if found.
      Deprecated, because it relies on obtaining the underlying VarGen
      type from the VarMap, but useful for debugging. */
  string vargen_stringrep(const string& varname) const {
    VarGen* v = vars.getvalgenptr(varname);
    return v? v->stringrep() : "";
  }

  /** If this object stores a pointer to its parent, return it. */
  virtual const Retinal_Composite* get_parent() const {  return parentptr;  }

  /** Store a pointer to our parent. */
  virtual void                     set_parent(const Retinal_Composite* _parentptr)
    {  parentptr=_parentptr;  }


  /** Query current angle for this object. */
  virtual Angle angle() const {  return get_var("theta");  }

  /** Find by name; see Retinal_Composite::find() for more info. */
  virtual Retinal_Object* find(const string& othername)
    {  return (othername=="" || othername==name_ ? this : 0);  }

  /** For use in complicated constructors: set a lot of variables at once. */
  void merge_vars(const VarMap& init_vars) {  vars.merge(init_vars);  }

  /** Returns true if the given coordinate is inside the active area
      of this object, favoring the most precise response possible,
      regardless of performance.  A rectangular bounding box will
      suffice, but if an object can give a more precise answer,
      e.g. if it really has a circular or irregular shape instead of a
      box, it should give the more precise answer here.  This call
      should be used only when such precision is needed; the regular
      rectangular bounding box can be used in other cases.
  */
  virtual bool inside(Coordinate x, Coordinate y) const
    {  return bounding_box.inside(x,y);  }
  //@}


  /** \name    Quasi-protected members.

    These members would otherwise be marked 'protected', but a
    Retinal_Composite object needs to be able to access them.  A
    composite object is not necessarily (or even usually) the same
    type as the objects it contains, and an obscure C++ rule
    stipulates that a class can only access protected members of
    objects of its own specific type.  Thus even though these members
    are nominally public, only derived classes should use them.
  */
  
  //@{
  /** As activation() is to activ(), this is to default_activ(). */ 
  virtual Activity default_activation() const 
    {  return get_var("offset") + get_var("scale") * default_act;  }

  /** Bounding box enclosing this object in the plane */
  mutable BBox bounding_box;
  //@}


protected:
  friend class DefaultActivityAccumulator;

  /** Set a parameter only if not yet set; accepts only a variable, not a generator */
  const void set_var_default(const string& varname, Variable val)
    {  vars.set_default(varname,val);  }

private:
  VarMap vars; 
  bool active;
  string name_;
  mutable Variable default_act,theta,cx,cy,scale,offset;
  const Retinal_Composite* parentptr;

  /**
    Unscaled activation.
    
    For most objects this routine constitutes the input-drawing
    performance bottleneck.
  */
  virtual Activity activ(Coordinate x, Coordinate y) const =0;

  /**
    Activation to use when object is inactive or position is out of bounds.
    
    Could replace with measurement of activity along bounding-box
    borders, though the current implementation may actually be more
    general.
  */
  virtual Activity default_activ() const { return 0.0; }
  //virtual Activity default_activ() const { return 0.5; } /* To debug non-Composite bounding boxes */

  /** Assignment operator; required by compiler but is not usable */
  void operator= (const Retinal_Object &) { abort(); } 
};




#ifndef NO_VALGEN_STRINGS
/** Helper class to handle string constructors for Retinal_Objects */
class RetinalObjectStringArgs {
public:
  /** Map for default values */
  typedef std::map<string,string> DefaultsMap;
  /** Allows object to declare the parameters it expects */
  typedef std::vector<string> ParamList;

  /**
    The constructor requires the actual string arguments,
    a ValueGeneratorFactory to be used to construct the variables,
    a map (possibly empty) of default values, and 
    a pointer to the object to whose values we should link, if any.
  */
  RetinalObjectStringArgs( StringArgs arglist,
			   VGenFactory<Retinal_Object::Variable>& vgf,
			   DefaultsMap dm=DefaultsMap(),
			   Retinal_Object* m=0)
    : args(arglist), defaults(dm), master(m), f(vgf), parser(args.p) { }
  
  /**
    Returns a set of ValueGenerators constructed by parsing the list
    of string arguments using the list of expected parameters and the
    map of defaults.  This should be all that is needed for simple
    argument lists consisting only of ValueGenerators.
  */
  Retinal_Object::VarMap vars(const ParamList& params);


  /** \name    Extended interface
      Not all applications need these functions. */
  //@{
  
  /**
    Safe lookup function for DefaultsMap.
    Defaults to empty string when not found; this usually evaluates to
    zero and/or a warning when parsed.
  */
  string get_default(const string& name) const {
    std::map<string,string>::const_iterator existing = defaults.find(name);
    return (existing ==  defaults.end() ? string() : (*existing).second);
  }

  /** Overrides defaults in the map passed in at construction. */
  void set_default(const string& name, const string& value)
    {  defaults[name]=value;  }

  /** Signals an error using the StringParser's error facility. */
  void error(const string& s) const {  parser.error(s);  }

  /** Looks up and returns a parse of the default with the given name. */
  const string parsed_get_default(const string& name) const
    {  return parser.parse(get_default(name));  }

  /** Consumes the next argument using the given name to find the default. */
  const string parsed_next(const string& name)
    {  return parser.parse(args.next(get_default(name))); }

  /** Simple numeric parsing. */
  Retinal_Object::Variable& parse(const string& s, Retinal_Object::Variable& x) const
    {  return parser.parse(s,x);  }
  //@}

  
private:
  /** Set the given ValueGenerator to the given value, optionally linking to one instead. */
  inline void set_linked(const string& name, const string& value,
			 Retinal_Object::VarMap& vars) {
    vars.set(name, f.create(StringArgs(parser,parser.parse(value)),
			    (!master ? 0 : master->get_varptr(name)) ) );
  }

  friend class RetinalObjectFactory; /* For new_Retinal_Clone */
  
  StringArgs args;
  DefaultsMap defaults;
  Retinal_Object* master;
  VGenFactory<Retinal_Object::Variable>& f;
  const StringParser& parser;
};
#endif /* NO_VALGEN_STRINGS */



/******************************************************************************/
/* Retinal_Composite classes                                                  */
/******************************************************************************/

/**
  Base class for composite objects, which are Retinal_Objects which 
  may be composed of other Retinal_Objects.

  Additional operations are supported for composite objects; other
  derived Retinal_Objects should rarely need to define their own
  operations other than those overriding or completing the ones in
  Retinal_Object.
*/
class Retinal_Composite : public Retinal_Object {
public:
  /** Possible values for an accum_type */
  enum AccumulationType { Max, Min, Add, OneHot, Replace };

  /** \name    Constructors and destructors */
  //@{
  Retinal_Composite( const string& name_val, AccumulationType accum_type_i=Max,
		     const VarMap& init_vars=VarMap())
    : Retinal_Object(name_val,init_vars),
      accum_type(accum_type_i), dominant_child(0) {
    set_var_default("size_scale",1.0);
  }
    
#ifndef NO_VALGEN_STRINGS
  RetinalObjectStringArgs::ParamList paramlist();
  Retinal_Composite( const string& name_val, RetinalObjectStringArgs& sa)
    : Retinal_Object(name_val), dominant_child(0) {
    sa.set_default("size_scale","1.0"); /* Special default */    
    merge_vars(sa.vars(paramlist()));
    accum_type = Retinal_Composite::AccumulationType(get_var("accum_type"));
  }
#endif

  Retinal_Composite(const Retinal_Composite& b) /* Copy constructor */
    : Retinal_Object(b), children(b.children), accum_type(b.accum_type),
      dominant_child(0) {
    std::transform(ISEQ(children),children.begin(),std::mem_fun(&Retinal_Object::clone));
  }
  
  virtual ~Retinal_Composite() {  remove_all();  };
  //@}


  /** \name    Primary Interface */
  //@{
  virtual bool next();
  virtual void reset();
  virtual bool update() const;
  //@}


  /** \name    Composite-only interface */
  //@{


  /** Most Composites don't have any width or height, and instead just
      ask their parents.  With any luck, they will have a parent that
      knows the width and height, and overrides this implementation.
  */
  virtual Coordinate nominal_height() const {
    const Variable size_scale = get_var("size_scale");
    return (get_parent()? get_parent()->nominal_height()/size_scale : 0);
  }
  
  virtual Coordinate nominal_width()  const {
    const Variable size_scale = get_var("size_scale");
    return (get_parent()? get_parent()->nominal_width()/size_scale : 0);
  }
  
  /** Incorporate the given retinal object (must be allocated from the heap) */
  virtual void   add(Retinal_Object* obj) {
    children.push_back(obj);
    if (obj) obj->set_parent(this);
  }

  /** Whether there are currently any children */
  virtual bool   is_empty() const {  return (children.empty()); }

  /**
    Returns the first object with a matching name from a preorder
    search on self and children.  The empty string matches the first
    object found, whatever it is.
  */
  virtual Retinal_Object* find(const string& othername) {
    if (Retinal_Object::find(othername)) return this;
    for (iterator i=children.begin(); i!=children.end(); i++) {
      Retinal_Object* child_ptr = (*i)->find(othername);
      if (child_ptr) return child_ptr;
    }
    return 0;
  }

  /** Remove all children with the specified name. */
  virtual void   remove(const string& childname="") {
    for (iterator i=children.begin(); i!=children.end(); i++)
      if (childname=="" || ((*i)->name() == childname)) {
	if (dominant_child==*i) dominant_child=0;
	delete *i;          /* Remove object pointed to */
	children.erase(i);  /* Compact remaining objects */
	i--;                /* Adjust iterator to account for the missing elem */
      }
  }

  /** Remove all children; separate routine for use with for_each. */
  virtual void   remove_all() {  Generic::delete_contents(children);  dominant_child=0; }

  /** Rarely needed: Returns child which has the highest activation at
     that location, or self if childless. */
  virtual const Retinal_Object& mostactive(Coordinate x, Coordinate y) const;


  /** Deprecated: Returns child specified by number. */
  Retinal_Object* operator[] (unsigned int idx) const { return children[idx]; }
  size_t size() const { return children.size(); }
  //@}

  
  /** \name    Other functions which may be useful */
  //@{
  virtual string stringrep() const;
  virtual Retinal_Object* clone() const
    { return new Retinal_Composite(*this); };
  bool inside(Coordinate x, Coordinate y) const;
  //@}
  
protected:
  /** Container with pointers to the child objects managed by this instance. */
  typedef std::vector<Retinal_Object*>      ChildrenContainer;
  /** Iterator over this instance's children. */
  typedef ChildrenContainer::iterator       iterator;
  /** Iterator over this instance's children. */
  typedef ChildrenContainer::const_iterator const_iterator;
  /** The child objects managed by this instance. */
  ChildrenContainer children;

  /** How to combine the responses from multiple children. */
  AccumulationType accum_type;
  /** Starting value for an accumulation. */
  Activity accum_base() const;
  /** Cached values for activ(). */ 
  mutable Variable cosmt,sinmt,cx,cy,div_size;
  /** If accum_type is OneHot, the currently-dominant child. */
  mutable Retinal_Object* dominant_child;

private:
  virtual Activity activ(Coordinate x, Coordinate y) const;
  virtual Activity default_activ() const;
};



/**
  Same as Retinal_Composite, but can enforce minimum and/or maximum
  distances between children's centers and do other housekeeping
  chores.

  The constructor interface is just to allow the options to be 
  changed externally after creation; it's not particularly
  pretty.
*/
class Retinal_ManagedComposite : public Retinal_Composite {
public:
  Retinal_ManagedComposite
  ( const string& name_val,
    Tristate* min_distance_enforce, double* min_distance,
    Tristate* max_distance_enforce, double* max_distance,
    AccumulationType accum_type_i=Max, const VarMap& init_vars=VarMap())
    : Retinal_Composite(name_val,accum_type_i,init_vars),
      min_dist_enforce(min_distance_enforce),  min_dist(min_distance),
      max_dist_enforce(max_distance_enforce),  max_dist(max_distance) { }

  Retinal_ManagedComposite(const Retinal_ManagedComposite& b)
    : Retinal_Composite(b), /* Copy constructor */
      min_dist_enforce(b.min_dist_enforce),    min_dist(b.min_dist),
      max_dist_enforce(b.max_dist_enforce),    max_dist(b.max_dist) { }

  virtual bool next();

  virtual Retinal_Object* clone() const
    { return new Retinal_ManagedComposite(*this); };
  
private:
  Tristate* min_dist_enforce; double* min_dist;
  Tristate* max_dist_enforce; double* max_dist;
  
  bool distance_valid(const Retinal_Object& obj1, const Retinal_Object& obj2);
  bool accumulate_managed_next(bool val, Retinal_Object* r);
};



/**
  Specialized version of Retinal_ManagedComposite which can
  save some computation for activ() by knowing it is always 
  at center (0,0), angle 0, and size 1.0.  Also can store and
  return a nominal width and height, though it doesn't use
  them internally.

  Typically used for an entire retina's contents.
*/
class Retinal_AnchoredManagedComposite : public Retinal_ManagedComposite {
public:
  Retinal_AnchoredManagedComposite
  ( const string& name_val,
    Tristate* min_distance_enforce, double* min_distance,
    Tristate* max_distance_enforce, double* max_distance,
    AccumulationType accum_type_i=Add, const VarMap& init_vars=VarMap(),
    Coordinate height_=0, Coordinate width_=0)
    : Retinal_ManagedComposite(name_val, 
			       min_distance_enforce,  min_distance,
			       max_distance_enforce,  max_distance,
			       accum_type_i, init_vars),
    height(height_), width(width_) { }
  
  virtual string stringrep() const;

  virtual Retinal_Object* clone() const
    { return new Retinal_AnchoredManagedComposite(*this); };
  
  virtual Angle angle() const {
    /* Returns angle of last child, since own angle is always zero */
    return ( (children.begin()==children.end()) ? 0 :
	     children.back()->angle() );
  }

  virtual Coordinate nominal_height() const {  return height;  }
  virtual Coordinate nominal_width()  const {  return width;   }

  void resize(const Coordinate height_, const Coordinate width_)
    {  height=height_; width=width_;  }
  
private:
  virtual Activity activ(Coordinate x, Coordinate y) const;
  Coordinate height;
  Coordinate width;
};



/******************************************************************************/
/* Derived Retinal_Object classes                                             */
/******************************************************************************/

/** Retinal_Object for random noise across the entire retina. */
template<class RandomGen=Distributions::uniform>
class Retinal_RandomNoise : public Retinal_Object {
public:
  Retinal_RandomNoise( const string& name_val, const VarMap& init_vars=VarMap())
    : Retinal_Object(name_val, init_vars)  { }

#ifndef NO_VALGEN_STRINGS  
  RetinalObjectStringArgs::ParamList paramlist() {
    RetinalObjectStringArgs::ParamList p;
    p.push_back("scale");
    p.push_back("offset");
    return p;
  }
  Retinal_RandomNoise( const string& name_val, RetinalObjectStringArgs& sa)
    : Retinal_Object(name_val,sa.vars(paramlist())) { }
#endif

  virtual Activity activation(Coordinate, Coordinate) const {
    const Activity rad=scale*0.5;
    return RandomGen()(offset+rad,rad);
  }

  virtual bool update() const  {
    scale  = get_var("scale");
    offset = get_var("offset");

    return Retinal_Object::update();
  }

  virtual Retinal_Object* clone() const
    {  return new Retinal_RandomNoise(*this);  }
  
private:
  mutable Variable scale,offset;
  virtual Activity activ(Coordinate, Coordinate) const { return 0; } // Unused 
};

typedef Retinal_RandomNoise<> Retinal_UniformRandomNoise;


/** Retinal_Object for radially-symmetric Gaussian function. */
class Retinal_CircularGaussian : public Retinal_Object {
public:
  Retinal_CircularGaussian( const string& name_val, const VarMap& init_vars=VarMap(),
			    const Coordinate bound_multiplier=2.5) /* Empirical default */
    : Retinal_Object(name_val,init_vars), bound_mult(bound_multiplier) {  }

#ifndef NO_VALGEN_STRINGS  
  RetinalObjectStringArgs::ParamList paramlist();
  Retinal_CircularGaussian( const string& name_val, RetinalObjectStringArgs& sa,
			    const Coordinate bound_multiplier=2.5) /* Empirical default */
    : Retinal_Object(name_val,sa.vars(paramlist())), bound_mult(bound_multiplier) { }
#endif

  Retinal_CircularGaussian(const Retinal_CircularGaussian& b) /* Copy constructor */
    : Retinal_Object(b), bound_mult(b.bound_mult) {  }

  virtual bool update() const;
  
  virtual Retinal_Object* clone() const
    { return new Retinal_CircularGaussian(*this); };

private:
  const Coordinate bound_mult;
  mutable Variable cx,cy,div_sigmasq;
  virtual Activity activ(Coordinate x, Coordinate y) const;
};



/** Retinal_Object for an ellipsoidal Gaussian function. */
class Retinal_Gaussian : public Retinal_Object {
public:
  Retinal_Gaussian( const string& name_val, const VarMap& init_vars=VarMap(),
		    const Coordinate bound_multiplier=2.5) /* Empirical default */
    : Retinal_Object(name_val,init_vars), bound_mult(bound_multiplier) {  }

#ifndef NO_VALGEN_STRINGS  
  RetinalObjectStringArgs::ParamList paramlist();
  Retinal_Gaussian( const string& name_val, RetinalObjectStringArgs& sa,
		    const Coordinate bound_multiplier=2.5) /* Empirical default */
    : Retinal_Object(name_val,sa.vars(paramlist())), bound_mult(bound_multiplier) { }
#endif

  Retinal_Gaussian(const Retinal_Gaussian& b) /* Copy constructor */
    : Retinal_Object(b), bound_mult(b.bound_mult) {  }

  virtual bool update() const;
  
  virtual Retinal_Object* clone() const
    {  return new Retinal_Gaussian(*this);  };
  
private:
  const Coordinate bound_mult;
  mutable Variable cx,cy,cost,sint,div_xsigma,div_ysigma;
  virtual Activity activ(Coordinate x, Coordinate y) const;
};



/** Retinal_Object for a single-color square or rectangle. */
class Retinal_Rectangle : public Retinal_Object {
public:
  Retinal_Rectangle( const string& name_val, const VarMap& init_vars=VarMap())
    : Retinal_Object(name_val,init_vars) {  }

#ifndef NO_VALGEN_STRINGS  
  RetinalObjectStringArgs::ParamList paramlist();
  Retinal_Rectangle( const string& name_val, RetinalObjectStringArgs& sa)
    : Retinal_Object(name_val,sa.vars(paramlist())) { }
#endif

  virtual bool update() const;
  
  virtual Retinal_Object* clone() const
    {  return new Retinal_Rectangle(*this);  };
  
private:
  virtual Activity activ(Coordinate, Coordinate) const 
    {  return 1.0;  } /* Brightness is constant within the bounding box */
};



/** Retinal_Object for sine-wave or cosine-wave grating. */
class Retinal_SineGrating : public Retinal_Object {
public:
  Retinal_SineGrating( const string& name_val, const VarMap& init_vars=VarMap())
    : Retinal_Object(name_val,init_vars) {  }

#ifndef NO_VALGEN_STRINGS
  RetinalObjectStringArgs::ParamList paramlist();
  Retinal_SineGrating( const string& name_val, RetinalObjectStringArgs& sa)
    : Retinal_Object(name_val,sa.vars(paramlist())) {  }
#endif

  virtual bool update() const;

  virtual Retinal_Object* clone() const
    { return new Retinal_SineGrating(*this); };
  
private:
  mutable Variable cost,sint,phase,freq,cx,cy;
  virtual Activity activ(Coordinate x, Coordinate y) const;
};


/**
  Retinal_Object for Gabor function, which is a sine grating
  masked by a Gaussian.
  Could have this inherit from both Gaussian and SineGrating,
  but the complication is probably not worth it.
*/
class Retinal_Gabor : public Retinal_Object {
public:
  Retinal_Gabor( const string& name_val, const VarMap& init_vars=VarMap(),
		 const Coordinate bound_multiplier=2.5) /* Empirical default */
    : Retinal_Object(name_val,init_vars), bound_mult(bound_multiplier) {  }
  
#ifndef NO_VALGEN_STRINGS
  RetinalObjectStringArgs::ParamList paramlist();
  Retinal_Gabor( const string& name_val, RetinalObjectStringArgs& sa,
		 const Coordinate bound_multiplier=2.5) /* Empirical default */
    : Retinal_Object(name_val), bound_mult(bound_multiplier) {
    /* Special defaults */
    sa.set_default("ysigma", sa.get_default("xsigma")); //* Defaults to circular pattern
    sa.set_default("freq",   "1.0");  //* Defaults to show three lobes for typical Gabor sigma
    sa.set_default("phase",  "0");    //* Defaults to fix sine wave in center
    merge_vars(sa.vars(paramlist()));
  }
#endif
      
  Retinal_Gabor(const Retinal_Gabor& b) /* Copy constructor */
    : Retinal_Object(b), bound_mult(b.bound_mult) {  }

  virtual bool update() const;
  
  virtual Retinal_Object* clone() const
    { return new Retinal_Gabor(*this); };
  
private:
  const Coordinate bound_mult;
  mutable Variable cx,cy,cost,sint,div_xsigmasq,div_ysigmasq,freq,phase;
  virtual Activity activ(Coordinate x, Coordinate y) const;
};


/** Retinal_Object for a line with Gaussian falloff around a constant-intensity area. */
class Retinal_FuzzyLine : public Retinal_Object {
public:
  Retinal_FuzzyLine( const string& name_val, const VarMap& init_vars=VarMap() )
    : Retinal_Object(name_val,init_vars) {  }

#ifndef NO_VALGEN_STRINGS
  RetinalObjectStringArgs::ParamList paramlist();
  Retinal_FuzzyLine( const string& name_val, RetinalObjectStringArgs& sa)
    : Retinal_Object(name_val,sa.vars(paramlist())) {  }
#endif

  virtual bool update() const;

  virtual Retinal_Object* clone() const
    { return new Retinal_FuzzyLine(*this); };
  
private:
  mutable Variable cx,cy,centerw,cost,sint,div_ysigmasq;
  virtual Activity activ(Coordinate x, Coordinate y) const;
};


/** Retinal_Object for a circular disc with Gaussian falloff around a
    constant-intensity area. */
class Retinal_FuzzyDisc : public Retinal_Object {
public:
  Retinal_FuzzyDisc( const string& name_val, const VarMap& init_vars=VarMap() )
    : Retinal_Object(name_val,init_vars) {  }

#ifndef NO_VALGEN_STRINGS
  RetinalObjectStringArgs::ParamList paramlist();
  Retinal_FuzzyDisc( const string& name_val, RetinalObjectStringArgs& sa)
    : Retinal_Object(name_val,sa.vars(paramlist())) {  }
#endif

  virtual bool update() const;

  virtual Retinal_Object* clone() const
    { return new Retinal_FuzzyDisc(*this); };
  
private:
  mutable Variable cx,cy,centerw,div_ysigmasq;
  virtual Activity activ(Coordinate x, Coordinate y) const;
};


/** Retinal_Object for a circular torus with Gaussian falloff around a
    constant-intensity area. */
class Retinal_FuzzyRing : public Retinal_Object {
public:
  Retinal_FuzzyRing( const string& name_val, const VarMap& init_vars=VarMap() )
    : Retinal_Object(name_val,init_vars) {  }

#ifndef NO_VALGEN_STRINGS
  RetinalObjectStringArgs::ParamList paramlist();
  Retinal_FuzzyRing( const string& name_val, RetinalObjectStringArgs& sa)
    : Retinal_Object(name_val,sa.vars(paramlist())) {  }
#endif

  virtual bool update() const;

  virtual Retinal_Object* clone() const
    { return new Retinal_FuzzyRing(*this); };
  
private:
  mutable Variable cx,cy,centerw,div_ysigmasq,radius;
  virtual Activity activ(Coordinate x, Coordinate y) const;
};


/** Retinal_Object for a greyscale PGM bitmap read from a file. */
class Retinal_PGM : public Retinal_Object {
public:
  Retinal_PGM(const string& name_val, const string& filename_base,
	      const VarMap& init_vars=VarMap() )
    : Retinal_Object(name_val,init_vars), basename(filename_base), ascale(1.0),
      aoffset(0.0), transparent_level(-1)
    {  set_active(pgm_read(basename,image)); border=mat::edge_average(image);  }

#ifndef NO_VALGEN_STRINGS
  RetinalObjectStringArgs::ParamList paramlist();
  Retinal_PGM( const string& name_val, RetinalObjectStringArgs& sa,
	       Retinal_Composite* parent);
#endif

  Retinal_PGM(const Retinal_PGM& b) /* Copy constructor */
    : Retinal_Object(b), basename(b.basename), image(b.image),
      border(b.border), transparent_level(b.transparent_level) { }

  virtual bool update() const;

  virtual Retinal_Object* clone() const
    {  return new Retinal_PGM(*this);  };

  virtual string stringrep() const {  return Retinal_Object::stringrep()+ " " + basename;  }
  
  virtual bool inside(Coordinate x, Coordinate y) const;

private:
  
  string basename;
  typedef MatrixType<Activity>::rectangular image_type;
  image_type image;
  mutable Variable border,cx,cy,cosmt,sinmt,div_size,xoff,yoff,ascale,aoffset;
  Activity transparent_level;
  virtual Activity activ(Coordinate x, Coordinate y) const;
  virtual Activity default_activ() const {  return border;  }
  //virtual Activity default_activ() const {  return 0.5;  } /* To debug PGM bounding boxes */
  Activity pixel_value(Coordinate x, Coordinate y, bool& inbounds) const;
};



/******************************************************************************/
/* RetinalObjectFactory class                                                 */
/******************************************************************************/

/**
  This class can create most or all of the Retinal_Object types
  contained in retinalobjs.h from a (fairly) simple string specification.
  If you need to support more types defined in other files, derive a
  new class from this one and have it call this create() for anything 
  your class doesn't handle.
*/
#ifndef NO_VALGEN_STRINGS
class RetinalObjectFactory {
public:
  virtual ~RetinalObjectFactory() { }

  /** Allocate and return an object based on the given state, defaults, and arguments. */
  virtual Retinal_Object* create(const string& name,
				 RetinalObjectStringArgs sa,
				 Retinal_Composite* basecomposite,
				 Retinal_Composite* parent) const {
    const string typenam = sa.parsed_next("type");
    
    if      (typenam=="Input_CircularGaussian"     )  return new Retinal_CircularGaussian     (name,sa);
    else if (typenam=="Input_Composite"            )  return new Retinal_Composite            (name,sa);
    else if (typenam=="Input_FuzzyDisc"            )  return new Retinal_FuzzyDisc            (name,sa);
    else if (typenam=="Input_FuzzyLine"            )  return new Retinal_FuzzyLine            (name,sa);
    else if (typenam=="Input_FuzzyRing"            )  return new Retinal_FuzzyRing            (name,sa);
    else if (typenam=="Input_Gabor"                )  return new Retinal_Gabor                (name,sa);
    else if (typenam=="Input_Gaussian"             )  return new Retinal_Gaussian             (name,sa);
    else if (typenam=="Input_Rectangle"            )  return new Retinal_Rectangle            (name,sa);
    else if (typenam=="Input_PGM"                  )  return new Retinal_PGM                  (name,sa,parent);    
    else if (typenam=="Input_SineGrating"          )  return new Retinal_SineGrating          (name,sa);
    else if (typenam=="Input_UniformRandomNoise"   )  return new Retinal_UniformRandomNoise   (name,sa);
    else if (typenam=="Input_Clone"                )  return new_Retinal_Clone                (name,sa,basecomposite);
    else {
      sa.error("Unknown retinal object type: `"+typenam+"'");
      return 0;
    }
  }

private:
  /** Retinal_Clone is not a real type, just shorthand for this function, which
    looks up the given object and returns an identical copy of it. */
  Retinal_Object* new_Retinal_Clone( const string& name_val,
				     RetinalObjectStringArgs& sa,
				     Retinal_Composite* basecomposite) const {
    const string oldname = sa.parsed_next("clone_name");
    Retinal_Object* oldobj = basecomposite->find(oldname);
    if (!oldobj) {
      sa.error("No object named `" + oldname + "' found to clone");
      return 0;
    }

    /* Make a new object and override certain fields as necessary */
    Retinal_Object* newobj = oldobj->clone();
    if (name_val!="")  newobj->set_name(name_val);
    if (sa.master) newobj->relink(sa.master);
    /* Process named parameters */
    newobj->merge_vars(sa.vars(RetinalObjectStringArgs::ParamList()));
    newobj->reset();
    
    return newobj;
  }
};
#endif /* NO_VALGEN_STRINGS */


#endif /* __RETINALOBJS_H__ */
