/** \file
  Retinal_Obj class: Abstract base class for an object which has a
  discrete state and which can return an activation value for any
  given floating-point position in the plane.

  $Header: /u/nn/cvsroot/lissom/src/robj.h,v 1.4 2001/04/01 07:06:00 jbednar Exp $
*/

#ifndef __ROBJ_H__
#define __ROBJ_H__

#include <iostream>
#include <string>
using std::string;

#include "vgen.h"

/**
  Abstract base class for an object which has a discrete state and
  which can return an activation value for any given floating-point
  position in the plane.  Derived classes implement a hierarchy of
  objects of various shapes which produce greyscale activation
  patterns suitable e.g. for presentation to a network that processes
  images.
*/
class Retinal_Obj : public ValueGen {
public:
  /** Type to use for an activity level. */
  typedef double Activity;
  /** Type to use for an angle, in radians. */
  typedef double Angle;
  /** Type to use for a distance or location on the surface of the retina. */  
  typedef double Coordinate;
  
  /** Advance to the next state (whatever that means for this object). */
  virtual bool next() =0;

  /** Reset to the starting state. */
  virtual void reset() =0;

  /** Must be called just before calling activation() one or more times */
  virtual bool update() const =0;

  /** Activation level for the given Cartesian position */
  virtual Activity activation(Coordinate x, Coordinate y) const =0;

  /**
    Returns the orientation of this object in radians.
    This value is intended to be used only for convenience
    for tasks like constructing default filenames.  Since
    not all Retinal_Objs are guaranteed to even have a
    meaningful orientation (e.g. circular objects or unbounded objects),
    bear in mind that the value will not always be useful.
  */
  virtual Angle angle() const=0;

  /** (Partial) representation of this object's state as a string */
  virtual string stringrep() const =0;

  /**
    Virtual interface to the operator<<, since << cannot be virtual.
    The string representation may someday be replaced with one that
    specifies the object entirely, in a form that can be read in
    using operator>>, rather than just printing the current state.
    Thus calling code should not rely upon this particular output
    format.
  */
  virtual std::ostream& put(std::ostream& s) const {
    return s << this->stringrep();
  }
};


/** Stream output for a Retinal_Obj */
inline std::ostream& operator<<(std::ostream& s, const Retinal_Obj& r) {  return r.put(s); }


#endif /* __ROBJ_H__ */
