/** \file
  Improved random number generators based on drand48; should be replaced with
  a real random number generating class that can support independent streams
  of values.

  $Header: /u/nn/cvsroot/lissom/src/shuffledrand.h,v 1.8 2001/11/27 05:51:16 jbednar Exp $
*/

#ifndef __SHUFFLEDRAND_H__
#define __SHUFFLEDRAND_H__

#include <cassert>
#include <cmath>
#include <stdlib.h>

#ifdef ANSI_COMPATIBLE
#define	NO_RAND48_AVAIL
#endif



/**
  Routine inspired by examples in "Numerical Recipes" with an extra
  shuffling of random numbers to avoid sequential correlations. Used
  e.g. in generating inputs to avoid potential artifacts.
  Returns value in open-ended range [0.0, 1.0).

  There is technically no reason why this routine would need to be a
  template, since only type double and (possibly) float are likely
  to be useful.  (It relies on drand48, which uses only 48 bits and
  thus wouldn't be sufficient for a 64-bit type.)  However, the Cray
  T3E compiler (as of PE3.2) got confused in certain obscure cases
  when this routine was coded as a regular function in a .c file.
*/
template<class T>
class ShuffledRandom {
public:
  
  /** Optional; may be called to reset the internal state */
  static void reset() {
    for (int i=1; i<=197; i++) v[i]= plain_rand();
    y = plain_rand();
    is_initialized = true;
  }

  /** Returns the next number in the series */
  static T next() {
    if (!is_initialized) reset();
    int i = (int)floor(1.0 + 197.0 * y); /* Compute a random index into v */
    assert(!(i > 197 || i < 1));
    y = v[i];                            /* Pick the number at the index  */
    v[i] = plain_rand();                    /* Restock with new random number*/
    return(y);
  }

  /** Default seed for shuffled_reset */
  static int default_seed;

  /** Simple random value 0..1.0, on which shuffledrandom is built. */
  static T plain_rand();

  /** Set seed for plain_rand() */
  static void plain_rand_seed(long int seedval);

private:
  static T y;
  static T v[198];
  static bool is_initialized;

  /** Because of this class's dependence on plain_rand, which does not
      allow independent streams of random numbers, all operations and
      variables are static.  To avoid any impression that different
      instantiations might be independent, instantiation is prevented
      by making the constructor private. */
  ShuffledRandom() {};
};


/** Static variables */
template<class T> T    ShuffledRandom<T>::y;
template<class T> T    ShuffledRandom<T>::v[198];
template<class T> bool ShuffledRandom<T>::is_initialized=false;
template<class T> int  ShuffledRandom<T>::default_seed=87654321;


template<class T> T
ShuffledRandom<T>::plain_rand()
{
#ifdef NO_RAND48_AVAIL
  /**
     This is an ok (but not great) substitution as long as
     int is at least 32 bits, but use drand48() if it is available.
     A separate random number generator altogether would need
     to be written to work well on 16-bit or smaller machines.
  */
  return (rand()/(RAND_MAX+1.0));
#else
  return drand48();
#endif
}



template<class T> void
ShuffledRandom<T>::plain_rand_seed(long int seedval)
{
#ifdef NO_RAND48_AVAIL
  /* Should match the definition of plain_rand() */
  srand(seedval);
#else
  srand48(seedval);
#endif
}



/** Shortcut for typical usage of shuffled. */
inline double shuffled_rand()
{  return ShuffledRandom<double>::next();  }


/** Reset the stream of values returned by shuffled. */
inline void   shuffled_rand_reset(long int new_seed=ShuffledRandom<double>::default_seed)
{
  ShuffledRandom<double>::plain_rand_seed(new_seed);
  ShuffledRandom<double>::reset();
  ShuffledRandom<double>::next();
}


#endif /* __SHUFFLEDRAND_H__ */
