/** \file
  Very basic parser to support object construction from a string specification.

  $Header: /u/nn/cvsroot/lissom/src/stringparser.h,v 1.29 2004/08/24 13:56:05 jbednar Exp $
*/

#ifndef __STRINGPARSER_H__
#define __STRINGPARSER_H__

#include <string>
using std::string;
#include <sstream>
#include <vector>

#include "tristate.h"

/**
  This class provides a basic way for values of various types to be
  specified in a string.  As implemented, it supports only the basic
  reading of integers and doubles as numeric constants, copies
  strings verbatim, ignores pointers, and breaks lines into words
  without taking quotes, comments, etc. into account.  

  The pointer version of parsing for a datatype [parse(s,T*&x)] has
  the same semantics as the regular version, except that it is
  possible that the pointer itself might be changed by the expression.
  (This implementation does not have any way to actually specify a
  pointer in a string, but derived classes can provide a mechanism for
  that.)  The supplied pointer must already be valid, since any
  regular value found will be placed into the pointee.  Any client
  that would like to support such pointer expressions should use the
  pointer form instead of the regular form (i.e. double* instead of
  double).
  
  For convenience in expressions, the parsed value is returned in both
  the return type and in the supplied variable.  (Of course, the
  variable is required to be in the argument list so that the correct
  return type can be deduced.)

  Often this class will be overridden with a more capable class, which
  can e.g. accept expressions from which the actual value may be
  computed based on known values of variables, implement quoting, etc.
  Derived classes may raise exceptions or print error messages during
  the parsing, but they must always return a legal value of the
  indicated type.

  Unfortunately, C++ does not implement virtual templates, or else the
  basicparse() routine could have just been called parse() and some of
  the others eliminated.  Had that been done, however, no derived
  class could override the operations since they would not have been
  virtual.  So, alas, the types must be limited to those enumerated
  here.
*/

class StringParser {
public:
  StringParser() { };
  virtual ~StringParser() { }

  /** Type for the result of parsing a string into a list of strings */
  typedef std::vector<string>           arglist;
  /** Iterator for an arglist. */
  typedef std::vector<string>::iterator arglist_iterator;
  /** Shorter alias for arglist_iterator. */
  typedef std::vector<string>::iterator argptr;
  typedef std::vector<string>::const_iterator const_argptr;
  
  StringParser(const StringParser&) { } /**< Copy constructor */

  /** Return a duplicate of the current object */
  virtual StringParser* clone() const { return new StringParser(*this); };

  /** \name Family of parse functions differing only by their return datatypes */
  //@{ 

  /** Parse into a double. */
  virtual double&   parse(const string& s, double&   x) const {  basicparse(s,x);  return x;  }
  /** Parse into an int. */
  virtual int&      parse(const string& s, int&      x) const {  basicparse(s,x);  return x;  }
  /** Parse into a Tristate. */
  virtual Tristate& parse(const string& s, Tristate& x) const {  basicparse(s,x);  return x;  }
  /** Parse into a string. */
  virtual string&   parse(const string& s, string&   x) const {  x=s;              return x;  }

  
  /** Parse into a pointer to a double. */
  virtual double*&   parse(const string& s, double*&   x) const {  basicparse(s,*x); return x;  }
  /** Parse into a pointer to an int. */
  virtual int*&      parse(const string& s, int*&      x) const {  basicparse(s,*x); return x;  }
  /** Parse into a pointer to a Tristate. */
  virtual Tristate*& parse(const string& s, Tristate*& x) const {  basicparse(s,*x); return x;  }
  /** Parse into a pointer to a string. */
  virtual string*&   parse(const string& s, string*&   x) const {  *x=s;             return x;  }


  /** Parse into a vector of strings by breaking at word boundaries. */
  virtual arglist&  parse(const string& s, arglist&  x) const {
    std::istringstream ist(s);
    x.resize(0);
    string w;
    while (ist>>w)
      x.push_back(w);
    return x;
  }
  //@}

  /** Short version of parse() for simple string substitution*/
  virtual string parse(const string& s) const {  string dummy;  return parse(s,dummy);  }
  
  /** This implementation ignores errors (since it doesn't generate
     any), but derived classes may redefine this function if they wish
     (for instance, to write messages to the console). */
  virtual void error(const string&) const {  }
  
private:

  /**
    This was once a template, but the template was replaced with a
    macro to avoid an "Internal compiler error" in GCC 2.95.  To
    restore the template, just replace the first line of the define
    with:

    \verbatim
    template <class T>
    \endverbatim

    and remove the backslashes and the macro calls.
  */
//template <class T>
#define BASICPARSE_DEF(T) 						\
    static T& basicparse(const string& s, T& x) 			\
    {  std::istringstream ss(s); ss >> x; return x;  }

  BASICPARSE_DEF(double);
  BASICPARSE_DEF(int);
  BASICPARSE_DEF(Tristate);
};




/**
  Simple string-based function argument parser using StringParser.  An
  argument of this type is good for a function which constructs
  an object from a string specification.  If you're starting with a
  simple string, you can use the string constructor below; otherwise
  (e.g. if starting with a pre-segmented vector of strings) use the
  iterator-based constructor.  An object from this class can be passed
  around as a reference argument to various other classes, each of
  whom can consume some of the tokens until all are gone.

  To consume a token, supply to next() a default value of the type you
  would like returned; this value will be returned if no arguments
  remain.  Even if no default is meaningful, supply one anyway but
  first check that the arglist is not empty(), which will ensure that
  the default is not used.
*/
class StringArgs {
  StringParser::arglist a; // Only used if we are constructed from string
  StringParser::argptr beg;
  StringParser::argptr end;
  
public:
  /** Provided as a public service to those needing a parser; avoid if possible */
  const StringParser& p; 

  StringArgs(const StringParser& parser, StringParser::argptr b, StringParser::argptr e)
    :  beg(b), end(e), p(parser)  { }
  
  StringArgs(const StringParser& parser, const string& s) : p(parser)  {
    p.parse(s,a);
    beg=a.begin();
    end=a.end();
  }

  /** Copy constructor. */
  StringArgs(const StringArgs& other)
    : a(other.a), beg(other.beg), end(other.end), p(*other.p.clone()) {
      /*
	Copy constructor is tricky -- if the other was using an internal
	container, need to construct equivalent iterators for our own
	copy.
      */ 
      if (!other.a.empty()) {
	beg=a.begin()+((StringParser::const_argptr)(other.beg)-(StringParser::const_argptr)(other.a.begin()));
	end=a.begin()+((StringParser::const_argptr)(other.end)-(StringParser::const_argptr)(other.a.begin()));
      }
    }

  /** Returns true if no arguments remain. */
  bool empty() const {  return (beg==end);  }
  
  /** Returns true if exactly one argument remains. */
  bool oneremaining() const {  return (beg+1==end);  }
  
  /** Family of parses based on type of the given default.  Returns
      the first argument (or the default) without modifying the state. */
  template<class T>
  T top(const T& default_val) const {
    if (empty())
      return default_val;
    else {
      T result;
      p.parse(*(beg), result);
      return result;
    }
  }

  /** Consumes the current argument. */
  void skip() {  if (!empty()) beg++;  }
  
  /** Same as top() but consumes the current argument. */
  template<class T>
  T next(const T& default_val) {
    if (empty())
      return default_val;
    else {
      T result;
      p.parse(*(beg++), result);
      return result;
    }
  }

  /** Version returning a pointer instead of a value. */
  template<class T>
  T* next(const T& default_val, T*& ptr) {
    if (empty())
      *ptr = default_val;
    else
      p.parse(*(beg++), ptr);
    return ptr;
  }

  /** Consumes only if equal to given value, and returns true if so. */
  template<class T>
  bool nextis(const T& test_val) {
    if (!empty()) {
      T result;
      p.parse(*beg, result);
      if (test_val == result) {
	beg++;
	return true;
      }
    }
    return false;
  }

  /** Returns true if the next argument after this one is equal the given value;
      rarely needed. */
  template<class T>
  bool lookaheadone(const T& test_val) const {
    if (empty() || (beg+1)==end)
      return false;

    T result;
    p.parse(*(beg+1), result);
    return (test_val == result);
  }

  /** Consumes the current argument as a string and returns a new
      parser parsing the substructure of it, i.e. arguments within
      that string. */
  StringArgs recursiveparser() {  return recursiveparser(next(string("")));  }

  /** Returns a new parser parsing the given string. */
  StringArgs recursiveparser(const string s) const {  return StringArgs(p, s);  }

  /** Returns a string representation of the remaining arguments. */
  string stringrep() const {
    if (beg==end) return "";
    string result = *beg;
    for (StringParser::argptr i=beg+1; i!=end; i++)
      result = result + " '" + *i + "' ";
    return result;
  }
};


#endif /* __STRINGPARSER_H__ */
