/** \file
  General-purpose string utility functions

  $Header: /u/nn/cvsroot/lissom/src/stringutils.h,v 1.23 2004/08/24 13:56:05 jbednar Exp $
*/

#ifndef __STRINGUTILS_H__
#define __STRINGUTILS_H__

#include <sstream>
#include "stringparser.h"

/** General-purpose string utility functions. */
namespace String {
  

/** Returns true if the specified globstr, which may include a "*"
    wildcard at the beginning and/or end (only!), matches the teststr.
    Eventually it would be much better to replace it with a call to a
    full globbing or regular expression library, but this
    implementation was simpler for now.  Callers should not assume
    that strings containing some wildcards that do not now match will
    continue not matching in future.  If the teststr starts with a
    period, the result will be true only if it matches exactly.
*/
template <class String>
bool glob_match(const String& globstr, const String& teststr)
{
  /* Brute force: examines all of the supported cases individually */
  if (globstr.empty()) return teststr.empty();
  
  const bool startglob = (globstr[0] =='*');
  const bool endglob   = (globstr[globstr.length()-1]=='*');

  if ((teststr[0]=='.') && startglob) return false;
  
  if (!startglob && !endglob) return globstr==teststr;
  
  const String globtext  = String(globstr, (startglob? 1 : 0), globstr.length()-startglob-endglob);

  if (!startglob && endglob)
    return (teststr.find(globtext)==0);

  if (startglob && !endglob)
    return (teststr.rfind(globtext)==teststr.length()-globtext.length());

  /** startglob && endglob */
  return (globtext=="" || teststr.find(globtext)<teststr.length());
}



/** Returns true if the two strings of any type match up to trailing digits. */
template <class String>
bool non_numeric_basename_matches(const String& a, const String& b)
{
  return (String(a.data(),a.find_last_not_of("0123456789")) ==
	  String(b.data(),b.find_last_not_of("0123456789")) );
}



/** Converts any trailing digits to the given type and returns them;
    result is undefined if there are no digits at the end. */
template <class T, class String>
T numeric_extension(const String& a)
{
  String b=a;
  b.erase(0,b.find_last_not_of("0123456789")+1);
  if (b.empty())
    return T();
  
  T result;
  const StringParser sp;
  return sp.parse(b,result);
}



/** Returns the integer portion of the given string, which may
    represent either an integer or a floating-point number. */
template <class String>
int integer_portion(const String& a)
{
  int pivot = int(a.rfind("."));
  if (pivot<0 || pivot>=int(a.length())) pivot=a.length();
  String b(a,0,pivot);
  int result;
  const StringParser sp;
  return sp.parse(b,result);
}



/** Returns the fractional portion of the given string, which may
    represent either an integer or a floating-point number.  If
    the string has no decimal point, zero is returned.
*/
template <class String>
int fractional_portion(const String& a)
{
  int pivot = int(a.rfind("."));
  if (pivot<0 || pivot+1>=int(a.length())) return 0;
  String b(a,pivot+1,a.length()-pivot-1);
  int result;
  const StringParser sp;
  return sp.parse(b,result);
}



/**
  Returns the length of the prefix of the given string which consists
  entirely of characters allowed in C-language identifiers. (First
  character must be alphabetical, and rest must be alphanumeric or an
  underscore.)  An extra character nonalnum can be supplied which
  will also be accepted; if e.g. ':' namespace or class-qualified
  names will be accepted.  If an explicit nonalnum is supplied,
  it will also be accepted in the first position of the identifier.
*/
inline unsigned C_identifier_length(const char* s, const char nonalnum='_')
{
  if (!s || (!isalpha(*s) && (nonalnum=='_' || *s!=nonalnum)))
    return 0;
  const char * rem;
  for (rem=s; (isalnum(*rem) || *rem == '_' || *rem == nonalnum); rem++);
  return rem-s;
}



template <class String>
inline unsigned C_identifier_length(const String& s, const char nonalnum='_')
{  return C_identifier_length(s.c_str(),nonalnum);  }



/** Strips off one set of surrounding single or double quotes, if any are present */
template <class String>
String strip_quotes(const String& s)
{
  unsigned start = 0;
  unsigned end   = s.length()-1;

  if ( (s[start] == '"'  && s[end] == '"' ) ||
       (s[start] == '\'' && s[end] == '\'') ) {
    start++;
    end--;
  }
  return String(s,start,end-start+1);
}



/** Returns the result of replacing all instances of the given
    fromtext with the given totext. */
template <class String>
inline string replace_all(const String& s, const String& fromtext, const String& totext)
{
  unsigned pos;
  unsigned start=0;
  string p=s;
  while ((pos=p.find(fromtext,start))<p.length()) {
    p.replace(pos,fromtext.length(),totext);
    start=pos+totext.length();
  }
  return p;
}



/** Convenience structure to wrap up and pass around a set of
    arguments for stream output, e.g. for stringrep().  Somewhat
    similar to Stroustrup's Form class from section 21.4.6.3
    of the C++ Programming Language, 3rd Ed.; might want to adapt
    that class someday instead.
*/
class StreamFormat {
public:
  
  virtual ~StreamFormat() { }

  /** Constructor taking explicit arguments.  The width and fill apply
      to all types; they can e.g. be used to padd a numeric value with
      a specified number of leading zeros.  The digitsafterdecimal is
      intended only for floating-point values; it specifies the
      precise number of digits to be used after the decimal (which by
      default depends instead on the actual value). */
  StreamFormat(const int width_=-1, const char fill_=' ', const int digitsafterdecimal_=-1,
	       const bool rightadjust_=true)
    : width(width_), fill(fill_), digitsafterdecimal(digitsafterdecimal_),
      rightadjust(rightadjust_) { }

  /** Constructor taking flags in a string format, which is handy for
      implementing formatted string substitution.  One could e.g. accept
      ${05.2value} or ${value:%05.2}, pull out the 05.2, and pass it to
      the flags argument here.

      Currently accepts an optional leading fill character (anything
      but [1-9], an integer width, and (preceded by a decimal point,
      if present) an optional integer number of fractional digits to
      be used for floating point values.
  */
  StreamFormat(const string& flags)
    : width(-1), fill(' '), digitsafterdecimal(-1), rightadjust(true) {
    if (flags!="") {
      const bool   hasfill     = (!isdigit(flags[0]) || flags[0] == '0');
      const string rest        = string(flags.begin()+(hasfill? 1 : 0),flags.end());
      const bool   hasfraction = (rest.find(".")<rest.length());
      
      width = String::integer_portion(rest);
      if (hasfill)     fill = flags[0];
      if (hasfraction) digitsafterdecimal = String::fractional_portion(rest);
    }
  }

  /** Virtual interface to the operator<<, since << cannot be virtual. */
  virtual std::ostream& put(std::ostream &s) const {
    /* Uses old-style stream setf interface for compatibility with
       more C++ implementations */
    if (width>=0)                s.width(width);
    if (digitsafterdecimal>=0) { s.precision(digitsafterdecimal); s.setf(std::ios::fixed,std::ios::floatfield); }
    if (fill!=' ')               s.fill(fill);
    if (!rightadjust)            s.setf(std::ios::left,std::ios::adjustfield);
    return s;
  }

private:
  int  width;
  char fill;
  int  digitsafterdecimal;
  bool rightadjust;
};



/** Instruct the given stream to use the specified formatting */
inline std::ostream& operator<<(std::ostream &s, const StreamFormat& a) {  return a.put(s);  }

  

/**
  Convert the given item (such as an integer or a float) into a
  string.  Specializations are required for objects that don't support
  the << operator or for whom the << output is inappropriate.  If
  desired, output can be controlled using the StreamFormat argument.
*/
template<class T>
string stringrep(const T& item, const StreamFormat format=StreamFormat())
{
  std::ostringstream os;
  os << format << item;
  return string(os.str());
}


} /* namespace String */

#endif /* __STRINGUTILS_H__ */
