/** \file
  Generic gnuplot line plotting for most vector and matrix classes.
  Loosely adapted from the Matrix class by Laurent.Deniau@cern.ch.
  The filename is currently a misnomer; the routines are no longer
  restricted to TNT matrices.
  
  $Header: /u/nn/cvsroot/lissom/src/tnt_gnuplot.h,v 1.18 2003/01/16 04:11:58 jbednar Exp $

  These routines should work with just about any vector class
  implementing operator[], e.g. std::vector.  Most matrix classes
  (including MTL and TNT) should also work, as long as they implement
  operator[][] and have nrows() and ncols() functions.  (For TNT one
  would need a "#define num_cols ncols" and "#define num_rows nrows"
  before including the headers.)  

  \verbatim
  Usage: [vec::|mat::]gnuplot(<vector-or-matrix>, [<outputfilename>, [<gnuplot-code>, [<axis-coords>]]]);
  \endverbatim

  When no outputfilename is specified, a temporary file is generated
  and then viewed onscreen.  The gnuplot-code can be arbitrary Gnuplot
  commands, separated by newlines; an occurences of the word "Matrix"
  in the commands are replaced by the name of the temporary file to
  which the given vector or matrix is saved.  The routines return an
  error/warning message as a string, or the empty string if no error.

  Examples using the TNT matrix and vector classes:

  <pre>
    Vector<double> A(4,    "3  4  8  1");
    Matrix<double> B(4, 4, "3  4  8  1 "
                           "7  8  5  1 "
                           "1  2  6  4 "
                           "2  4  9  7 ");

    // Plot with default axis coords
    vec::gnuplot(A);
    mat::gnuplot(B);

    // Plot with X-axis coords specified by 5x+10
    vec::gnuplot(A,"","",10,5);
    mat::gnuplot(B,"","",10,5,4,2);  // (and Y-axis coords by 2y+4)

    // Plot with specified arbitrary axis coords
    vec::gnuplot(A,"","",Vector<double>(4,"10 15 20 25"));
    mat::gnuplot(B,"","",Vector<double>(4,"10 15 20 25"), Vector<double>(4,"4  6  8 10"));
  </pre>

  Note: if you are used to pixel-based image plots, these plots may be
  somewhat misleading because the integer bounds of the array are used
  as the floating-point bounds for the plot, with gnuplot linearly
  interpolating lines between the bounds.  This means that each row
  (and column, for 3D plots) will appear to have one fewer value than
  they really do.  The alternative would be to pad each dimension with
  an extra data point somehow, but this would look strange if the
  wrong value were chosen, and would probably be even more confusing
  in any case.
*/

#ifndef __GNUPLOT_H__
#define __GNUPLOT_H__

#include <iostream>
#include <fstream>
#include <cstdio>
#include <string>
using std::string;
#include <vector>
#include "matrix.h"

#include "ind_types.h"



/**
  Namespace for generic routines operating on vectors; this helps
  avoid overloading ambiguities.
*/
namespace vec {


/*
  Most of the types used are generic to keep the routines as general
  as possible, but the shortcut routines which generate their own axis
  labels use these definitions to construct them.
*/
/** Type for coordinate on an axis */
typedef double                 AxisCoord;
/** Type of a list of axis coordinates. */
typedef std::vector<AxisCoord> AxisVector;
/** Type of the index used by an AxisVector */ 
typedef AxisVector::size_type  AxisSubscript;


 
/******************************************************************************/
/* Private routines                                                           */
/******************************************************************************/


/** Cleaner wrapper for ugly C tempnam() function. */
/* If you have a tempnam() function on your system, do a
   #define TEMPNAME tempnam
   Otherwise write such function or edit this one. 
*/
inline string temporaryname(const string prefix)
{
  char* matrixfilechars = TEMPNAME(0,prefix.c_str());
  string matrixfilename(matrixfilechars);
  free(matrixfilechars);
  return matrixfilename;
}

 

/** Actually calls gnuplot on a given file and saves or displays the output. */
inline string call_gnuplot(const string& title,
			   const string& matrixfilename,
			   const string& outputfilename,
			   const string& gpscript_command,
			   bool parametric)
{
  const bool   interactive=(outputfilename=="");
  const bool   view_as_ps=true; // Leaves outputfile in tmp, though
  const bool   create_ps=(!interactive || view_as_ps);
  const string ps_view_command="ghostview";

  const string gpscript_save_to_ps =
    "set terminal postscript eps color solid 'Times-Roman' 14\n"
    "set output '";

  const string gpscript_display = "pause -1\n";

  const string scriptfilename = temporaryname("gnupl");

  const string outfilename = (outputfilename!="" ? outputfilename :
			      temporaryname("gnupl"));
    
  /* Create scriptfile */
  {
    std::ofstream scriptfile(scriptfilename.c_str());
    if (!scriptfile) return "Cannot open temporary script file";
    
    // write script to temporary file  
    if (create_ps)  scriptfile << gpscript_save_to_ps << outfilename << "'\n";
    if (parametric) scriptfile << "set parametric\n";
    
    scriptfile << "set title '" << title << "'\n";
    
    unsigned pos;
    string p = gpscript_command;
    const string filenamemarker="Matrix";
    while ((pos=p.find(filenamemarker))<p.length())
      p.replace(pos,filenamemarker.length(),matrixfilename);
    
    scriptfile << p;
    if (!create_ps) scriptfile << gpscript_display;
    
    if (scriptfile.bad()) return "Error writing script file";
  }
  
  // call gnuplot
  string cmd = "gnuplot " + scriptfilename;
  system(cmd.c_str());

  // remove temporary files
  if (remove(scriptfilename.c_str()) || remove(matrixfilename.c_str()))
    return "Cannot remove temporary files";

  if (interactive && view_as_ps) {
    string viewcmd = ps_view_command + " " + outfilename + "&";
    system(viewcmd.c_str());
  }
  
  return "";
}




/******************************************************************************/
/* 1D vector plots                                                            */
/******************************************************************************/


/** General routine for plotting vectors using gnuplot. */
template <class VectorA, class VectorX>
string gnuplot(const VectorA& A,
		    const string&  title,
		    const string&  outputfilename,
		    const string&  gpscript,
		    const VectorX& X)
{
  typedef typename VectorA::size_type  Subscript;
  
  if (Subscript(X.size())!=Subscript(A.size()))
                    return "Skipping; axis vector incompatible with vector size";
  if (!A.size())    return "Skipping empty vector";
  if (A.size()==1)  return "Skipping single-element vector";

  /* Create matrix file */
  {
    string matrixfilename = temporaryname("gnupl");
    std::ofstream matrixfile(matrixfilename.c_str());
    if (!matrixfile) return "Cannot open temporary file";
    
    // write matrix to temporary file
    for (Subscript i=0; i<A.size(); i++) {
      matrixfile << X[i] << " ";
      matrixfile << A[i] << " ";
      matrixfile << std::endl;
    }
    
    if (matrixfile.bad()) return "Error writing matrix file";
  }
  
  return call_gnuplot
    ( title,matrixfilename,outputfilename,

	gpscript!="" ? gpscript :
	"set xlabel 'Element'\n"
	"set ylabel 'Value'\n"
	"plot 'Matrix' notitle with lines\n",
	
	false);
}



/** Shortcut version for plotting a vector using gnuplot; generates linearly spaced labels */
template <class Vector>
string gnuplot(const Vector& A,
		    const string& title="",
		    const string& outputfilename="",
		    const string& gpscript="",
		    AxisCoord xo=0, AxisCoord xm=1)
{
  AxisVector X(A.size(),0.0);
  for (AxisSubscript i=0; i<AxisSubscript(A.size()); i++) X[i]=xo+i*xm;
  return gnuplot(A,title,outputfilename,gpscript,X);
}


} /* namespace vec */



/******************************************************************************/
/* 2D matrix plots                                                            */
/******************************************************************************/

namespace mat {
  using vec::call_gnuplot;
  using vec::temporaryname;
  using vec::AxisCoord;
  using vec::AxisVector;
  using vec::AxisSubscript;

 
/** General routine for plotting matrices using gnuplot. */
template <class Matrix, class Vector>
string gnuplot(const Matrix& A,
	       const string& title,
	       const string& outputfilename,
	       const string& gpscript,
	       const Vector& R,
	       const Vector& C)
{
  typedef typename Matrix::size_type  Subscript;
  
  if (int(C.size())!=int(A.ncols())) return "Skipping; column axis vector incompatible with matrix size";
  if (int(R.size())!=int(A.nrows())) return "Skipping; row axis vector incompatible with matrix size";
  if (A.ncols()==0 || A.nrows()==0)  return "Skipping empty matrix";
  if (A.ncols()==1 && A.nrows()==1)  return "Skipping single-element matrix";

  const string matrixfilename = temporaryname("gnupl");
  
  /* Create matrix file */
  {
    std::ofstream matrixfile(matrixfilename.c_str());
    if (!matrixfile) return "Cannot open temporary file";
    
    // write matrix to temporary file
    for (Subscript i=0; i<A.nrows(); i++) {
      for (Subscript j=0; j<A.ncols(); j++) {
	matrixfile << R[i] << " ";
	matrixfile << C[j] << " ";
	matrixfile << mat::elem(A,i,j) << std::endl;
      }
      matrixfile << std::endl;
    }
    
    if (matrixfile.bad()) return "Error writing matrix file";
  }
  
  return call_gnuplot
    ( title,matrixfilename,outputfilename,

      gpscript != "" ? gpscript :
      "set xlabel 'R'\n"
      "set ylabel 'C'\n"
      "set zlabel 'Value'\n"
      "set hidden3d\n"
      "splot 'Matrix' notitle with lines\n",
      
      true);
}



/** Shortcut version for plotting a matrix using gnuplot; generates linearly spaced labels. */
template <class Matrix>
string gnuplot(const Matrix& A,
		    const string& title="",
		    const string& outputfilename="",
		    const string& gpscript="",
		    AxisCoord Ro=0, AxisCoord Rm=1,
		    AxisCoord Co=0, AxisCoord Cm=1)
{
  AxisVector R(A.nrows(),0.0);
  for (AxisSubscript i=0; i<AxisSubscript(A.nrows()); i++) R[i]=Ro+i*Rm;

  AxisVector C(A.ncols(),0.0);
  for (AxisSubscript j=0; j<AxisSubscript(A.ncols()); j++) C[j]=Co+j*Cm;
  
  return gnuplot(A,title,outputfilename,gpscript,R,C);
}



} /* namespace mat */


#endif
