/** \file
  Generators for various distributions, including random.

  $Header: /u/nn/cvsroot/lissom/src/valuegen.h,v 1.37 2002/07/13 07:38:25 jbednar Exp $
*/

#ifndef __VALUEGEN_H__
#define __VALUEGEN_H__

#ifndef NO_VALGEN_STRINGS
/* Optional: support for reading expressions from strings
   and showing state as a string */
#include "stringutils.h"
#include "msg.h"
#endif

#include "vgen.h"
#include "allocatingpointer.h"
#include "shuffledrand.h"
#include "mathconstants.h"

/******************************************************************************/
/* General-purpose random-value generators for various distributions          */
/******************************************************************************/


namespace Distributions {

/**
  Returns a random number from a uniform distribution
  with a given radius (that is, range/2) and mean.  The radius is
  inclusive on the lower side and exclusive on the upper.
*/
struct uniform {
  double operator() ( double mean=0.5, double radius=0.5)
    {  return mean + 2.0*radius*(shuffled_rand() - 0.5); }
};



/**
  Returns a random number from the Gaussian distribution
  with the given sigma and mean.  
  Uses the Box-Muller method from Numerical Recipes in C (Ch 7).
*/
struct normal {
  double operator() ( double mean=0.0, double radius=1.0)
    {  return mean + radius*(sqrt((-2.0) * log(shuffled_rand())) * cos(2*(Math::pi) * shuffled_rand()));  }
};

}


/******************************************************************************/
/* ValueGenerator objects                                                     */
/******************************************************************************/
/**
  This heads a class hierarchy of objects which primarily contain a
  variable of the given type, but which also contain operations for
  automatically generating a new value from a given distribution or by
  some other means.  Semantics for a value "a" of this type and a
  value "b" of the numeric type "T" which it contains:

  <pre>
    ValueGenerator<T>  a;
    T b;
    T* c;
    ValueGenerator<T>* d;

    a = b;           // Assign the value b to a
    b = a.value();   // Assign the value of a to b
                     // (can omit .value() if conversion operator defined)
    a.next();        // Choose a new value for a; returns false if unsuccessful
    a.reset();       // Reset a to its initial value, if any
    c = a.valueptr();// Returns address of data in a
    c = a.genptr();  // Returns address of a
    d = a.clone();   // d points to an identical but independent copy of a
                     // (i.e., d may outlive a.  This helps support
		     // polymorphism, since a may be of a subtype of d, yet
		     // the duplicate is of the subtype, not d's type.
  </pre>

  Various derived types are used to support different distributions
  by reimplementing "next()" and "reset()"; their interface should differ
  only by their constructors and not by their usage.

  The operator relink() is supported for the rare instances when one might
  wish to change a stored external value pointer to point to a different 
  location.  If an internal storage location is actually being used, the
  operation has no effect.  

  This base class is a basic version that acts just like an ordinary
  variable or reference.
*/
template<class T=double>
class ValueGenerator : public ValueGen {
  typedef ValueGenerator<T> self;
  AllocatingPointer<T> v;
  
public:
  /** Constructor for an internally-allocated variable */
  ValueGenerator(T  value=T()) : v(value)  { }

  /** Constructor for a reference to an external value */
  ValueGenerator(T* value)     : v(value) { }
  
#ifndef NO_VALGEN_STRINGS
  ValueGenerator(StringArgs& args) {  args.next(T(),v.valueptrref());  }

  /** Returns the value as a scalar.  Derived classes will usually
      want to provide a more specific representation, namely
      what the StringArgs constructor will accept for that class.
      
      At present it cannot be guaranteed that the value returned is
      actually what was used to create the object; perhaps the
      StringArgs passed to the constructor should be saved instead.
      If so, the StringArgs constructor above would have to be changed
      to include a string specifying the ValueGenerator type.
  */
  virtual string stringrep() const
    {  return String::stringrep(value());  }
#endif

  virtual ~ValueGenerator() { }

  /** Returns the current underlying value. */
  T operator*() const {  return *v;  }

  /** Returns the current underlying value. */
  T value() const { return *v; }

  /** Returns a pointer to the underlying value; use with caution. */
  T* valueptr() const {  return v.valueptr();  }

  /** \name    Operations that may be overridden by derived classes */
  //@{
  virtual void relink(T* ptr) {  v.relink(ptr);  }
  
  virtual bool next()  { return true; }
  virtual void reset() { }

  /** Constructs an identical copy of self and returns it. */
  virtual ValueGenerator<T>* clone() const { return new self(*this); };
  //@}
};


/** Short name for typical usage. */
typedef ValueGenerator<double> DGen;



/** Choose value from a given random distribution. */
template<class T=double, class RandomGen=Distributions::uniform>
class ValueGenerator_Random : public ValueGenerator<T> {
  typedef ValueGenerator_Random<T,RandomGen> self;
  T v; /**< Internal storage for generated value; ignored after a relink() */
  AllocatingPointer<T> m,r;

public:
  /** Constructor defaults to no randomness (which acts just like a regular
      ValueGenerator, except that relinking always replaces the pointer).
  */
  ValueGenerator_Random(T mean, T radius=0)
    : ValueGenerator<T>(&v), v(mean),   m(mean),  r(radius) { }

  /** Constructor using external mean and radius */  
  ValueGenerator_Random(T* mean, T* radius)
    : ValueGenerator<T>(&v), v(*mean), m(mean), r(radius) { }

#ifndef NO_VALGEN_STRINGS
  ValueGenerator_Random(StringArgs& args) : ValueGenerator<T>(&v) {
    args.next(T(),m.valueptrref()); v = *m;
    args.next(T(),r.valueptrref());
  }

  /** Note that the value returned does not indicate the actual
      RandomGen, which may differ from the defaults. */
  virtual string stringrep() const
    {  return "Random " + String::stringrep(*m) + " " + String::stringrep(*r);  }
#endif

  /** Copy constructor */
  ValueGenerator_Random(const self& other)
    : ValueGenerator<T>(&v), v(other.v), m(other.m), r(other.r)  { }

  virtual bool next()
    {  v = RandomGen()( *m, *r ); return true; }

  virtual void reset() { v=*m; } // Resets to mean value
  
  virtual ValueGenerator<T>* clone() const { return new ValueGenerator_Random(*this); };
};



/**
  Distribute value around a master value located elsewhere,
  using a supplied uncorrelation factor plus upper and lower bounds.

  If using a uniform distribution, the value may be cropped to the
  given bounds, as it is by default, but cropping should be avoided
  for other distributions since the cropping would either change the
  distribution or would require generating a number (possibly
  infinite) of different values from the distribution before finding
  one in the right range.
*/
template<class T=double, class RandomGen=Distributions::uniform, bool crop=true>
class ValueGenerator_Correlate : public ValueGenerator<T> {
  typedef ValueGenerator_Correlate<T,RandomGen,crop> self;
  T v; /**< Internal storage for generated value; ignored after a relink() */
  AllocatingPointer<T> lb,ub,m,uncorr;

public:
  /** master_value_ptr and uncorrelation_ptr should be const */
  ValueGenerator_Correlate(T* master_value_ptr, T* uncorrelation_ptr,
			   T lower_bound, T upper_bound)
    : ValueGenerator<T>(&v), v(*master_value_ptr),
    lb(lower_bound), ub(upper_bound),
    m(master_value_ptr), uncorr(uncorrelation_ptr) { }

#ifndef NO_VALGEN_STRINGS
  /** Construct from a string.

    If present, the ref argument is used in place of the first argument
    expected in the string, and the string is parsed from the beginning
    for the next argument.
  */
  ValueGenerator_Correlate(StringArgs& args, T* ref=0)
    : ValueGenerator<T>(&v) {
    if (ref) m=ref;
    else     args.next(v,m.valueptrref());
    args.next(v,uncorr.valueptrref());
    args.next(T(),lb.valueptrref());
    args.next(T(),ub.valueptrref());
    
    v = *m;  
  }
  
  virtual string stringrep() const {
    return "Correlate " +
      String::stringrep(*m) + " " +
      String::stringrep(*uncorr) + " " +
      String::stringrep(*lb) + " " +
      String::stringrep(*ub);
  }
#endif
  
  ValueGenerator_Correlate(const self& other)
    : ValueGenerator<T>(&v), v(other.v), lb(other.lb), ub(other.ub),
      m(other.m), uncorr( other.uncorr)  { } /**< Copy constructor */

  virtual bool next() {
    const T uncorr_radius = (*ub-*lb)*(*uncorr);

    /* This conditional will presumably be optimized away */
    if (crop) {
      const T low    = std::max(*lb,*m-uncorr_radius);
      const T high   = std::min(*ub,*m+uncorr_radius);
      const T radius = (high-low)/2;
      const T mean   = low+radius;
      v = RandomGen()(mean, radius);
    }
    else
      v = RandomGen()(*m, uncorr_radius);
    return true;
  }

  virtual void reset() { v=*m; } // Resets to master value

  virtual ValueGenerator<T>* clone() const { return new self(*this); }
};



/**
  Given a factor f and a master value mv located elsewhere, the value
  of this object will be f-mv.

  This class is a hack that shouldn't have to exist, but without it
  there would currently (-r 1.36, 6/2002) be no way to have a value in
  one eye be 1 minus the value in the other, despite the expressions
  supported by ValueGenerator and ValueGenerator_Expression.  The
  trouble is that all ValueGenerators in eyes other than the first one
  are linked to those in the first, usually via
  ValueGenerator_Correlate ("Uncorrelate ...").  This linkage is not
  very flexible, and thus this class provides one other commonly used
  way to link, other than uncorrelation.  A general way to link values
  between eyes is sorely needed.
*/
template<class T=double>
class ValueGenerator_Opposite : public ValueGenerator<T> {
  typedef ValueGenerator_Opposite<T> self;
  T v; /**< Internal storage for generated value; ignored after a relink() */
  T f;
  AllocatingPointer<T> m;

public:
  /** master_value_ptr and uncorrelation_ptr should be const */
  ValueGenerator_Opposite(T* master_value_ptr, T factor=1)
    : ValueGenerator<T>(&v), v(1-*master_value_ptr), f(factor), m(master_value_ptr) { }

#ifndef NO_VALGEN_STRINGS
  /** Construct from a string.

    If present, the ref argument is used in place of the first argument
    expected in the string, and the string is parsed from the beginning
    for the next argument.  This behavior is inherited from
    ValueGenerator_Correlate; it is not clear if it is appropriate here.
  */
  ValueGenerator_Opposite(StringArgs& args, T* ref=0)
    : ValueGenerator<T>(&v) {
    if (ref) m=ref;
    else     args.next(v,m.valueptrref());
    f=args.next(T(1));
    
    reset();  
  }
  
  virtual string stringrep() const {
    return "Opposite " +
      String::stringrep(f) + "-" +
      String::stringrep(*m);
  }
#endif
  
  ValueGenerator_Opposite(const self& other)
    : ValueGenerator<T>(&v), v(other.v), f(other.f), m(other.m) { } /**< Copy constructor */

  virtual bool next() {  reset();  return true;  }
  virtual void reset() { v = f-*m; }
  virtual ValueGenerator<T>* clone() const { return new self(*this); }
};



/** Increment a starting value each time by the given amount. */
template<class T=double>
class ValueGenerator_Increment : public ValueGenerator<T> {
  typedef ValueGenerator_Increment<T> self;
  T v; /**< Internal storage for generated value; ignored after a relink() */
  AllocatingPointer<T> s,inc;

public:
  ValueGenerator_Increment(T  init_value, T  increment=1)
    : ValueGenerator<T>(&v), v(init_value), s(init_value), inc(increment) { }

  ValueGenerator_Increment(T* init_value, T* increment)
    : ValueGenerator<T>(&v), v(*init_value), s(init_value), inc(increment) { }

#ifndef NO_VALGEN_STRINGS
  /** Could extend to accept references for arguments. */
  ValueGenerator_Increment(StringArgs& args)
    : ValueGenerator<T>(&v), inc(1) {
    s   = args.next(T());
    v   = *s;
    inc = args.next(T(1));
  }

  virtual string stringrep() const
    { return "Increment " + String::stringrep(*s) + " " + String::stringrep(*inc);  }
#endif
  
  ValueGenerator_Increment(const self& other)
    : ValueGenerator<T>(&v), v(other.v), s(other.s), inc(other.inc) { } /**< Copy constructor */

  virtual bool next()  { v += *inc; return true; }
  virtual void reset() { v  = *s; } // Resets to starting value

  virtual ValueGenerator<T>* clone() const { return new self(*this); };
};



/**
  Evaluate the given expression using a copy of the given parser each
  time.
*/
#ifndef NO_VALGEN_STRINGS

template<class T=double>
class ValueGenerator_Expression : public ValueGenerator<T> {
  typedef ValueGenerator_Expression<T> self;
  const StringParser* p;
  T v; /**< Internal storage for generated value; ignored after a relink() */
  string expr;
  
public:
  ValueGenerator_Expression(string expression, const StringParser& parser)
    : ValueGenerator<T>(&v), p(parser.clone()), v(p->parse(expression,v)), expr(expression) { }
  
  ValueGenerator_Expression(StringArgs& args)
    : ValueGenerator<T>(&v), p(args.p.clone()), v(0), expr("") {
      expr = args.next(string(""));
      p->parse(expr,v);
    }
  
  virtual string stringrep() const
    { return "Expression " + expr;  }

  ValueGenerator_Expression(const self& other)
    : ValueGenerator<T>(&v), p(other.p->clone()), v(other.v), expr(other.expr) { } /**< Copy constructor */
  
  ~ValueGenerator_Expression() {  delete p;  }
  
  virtual bool next() {  p->parse(expr,v); return true; }
  
  virtual void reset() { next(); } // Resets to current value of expr

  virtual ValueGenerator<T>* clone() const { return new self(*this); };
};
#endif




/******************************************************************************/
/* ValueGeneratorFactory                                                      */
/******************************************************************************/

#ifndef NO_VALGEN_STRINGS


/**
   Abstract base class for ValueGeneratorFactory independent of all template
   parameters except the datatype.
*/
template<class T=double>
class VGenFactory {
public:
  VGenFactory()          { };
  virtual ~VGenFactory() { }

  /**
     Create a ValueGenerator using the given string arguments.
     Notice that the args argument is passed by value and not
     reference, and thus this routine is assuming full responsibility
     for parsing the whole list.  The caller should not assume there
     are further arguments to process when this routine returns.
  */
  virtual ValueGenerator<T>* create(StringArgs args) const =0;

  /** Extended version of create() that can correlate or link the new ValueGenerator with an existing one. */  
  virtual ValueGenerator<T>* create(StringArgs args, T* ref) const =0;
  
protected:
  /** This implementation ignores errors, but derived classes may redefine
      this function if they wish, perhaps to write messages to the console.
      This support should probably be rewritten using msg.h, to
      inherit the global default instance.
  */
  virtual void error(const string&) const {  }
  /** This implementation ignores warnings, but derived classes may redefine
     this function if they wish, perhaps to write messages to the console. */
  virtual void warning(const string&) const {  }
};



/**
  This class can create most or all of the ValueGenerator types
  contained in this file from a (fairly) simple string specification.
  If you need to support more types defined in other files, derive a
  new class from this one and have it call this one for anything your
  class doesn't handle.  Instantiate the template with a different
  Msg::LevelHandler<> if you wish to provide support for error messages.
*/
template<class T=double,
	 class urf=Distributions::uniform,
	 class nrf=Distributions::normal>
class ValueGeneratorFactory : public VGenFactory<T> {
public:
  ValueGeneratorFactory(Msg::LevelHandler<>** mhp_i=&(Msg::LevelHandler<>::default_instance()))
    : VGenFactory<T>() {  mhp = mhp_i;  };
  
  virtual ~ValueGeneratorFactory() { }
  
  /** Constructs and returns a ValueGenerator of the specified type. */
  virtual ValueGenerator<T>* create(StringArgs args) const {
    
    /* Empty string is invalid */
    if (args.empty()) {
      warning("ValueGenerator: empty expression");
      return new ValueGenerator<T>;
    }
    
    /* Convenient shorthand: A single item specifies a numeric constant */
    else if (args.oneremaining())
      return new ValueGenerator<T>(args);
    
    /* Anything else should be the name of the ValueGenerator followed by its arguments */
    else {
      const string& vgclass  = args.next(string(""));

      if      (vgclass=="Random")     {  return new ValueGenerator_Random<T,urf>(          args);  }
      else if (vgclass=="Normal")     {  return new ValueGenerator_Random<T,nrf>(          args);  }
      else if (vgclass=="Increment")  {  return new ValueGenerator_Increment<T>(           args);  }
      else if (vgclass=="Expression") {  return new ValueGenerator_Expression<T>(          args);  }
      else if (vgclass=="Correlate")  {  return new ValueGenerator_Correlate<T,urf,true> ( args);  }
      else {
	error("ValueGenerator: Don't know about type `" + vgclass + "'");
  	return new ValueGenerator<T>;
      }
    }
  }


  /** Extended version of create() that can correlate or link the new ValueGenerator with an existing one. */
  virtual ValueGenerator<T>* create(StringArgs args, T* ref) const {
    if (args.nextis(string("Uncorrelate"))) {
      if (ref) return new ValueGenerator_Correlate<T,urf,true>(args,ref);
      else {
	/* Swallow unneeded args */
	T dummyf;
	ValueGenerator_Correlate<T,urf,true> dummy(args,&dummyf);
      }
    }
    else if (args.nextis(string("Opposite"))) {
      if (ref) return new ValueGenerator_Opposite<T>(args,ref);
      else {
	/* Swallow unneeded args */
	T dummyf;
	ValueGenerator_Opposite<T> dummy(args,&dummyf);
      }
    }
    if (ref) return new ValueGenerator<T>(ref);
    
    /* Otherwise just construct and return a normal ValueGenerator */
    return create(args);
  }
protected:
  void error   (const string& s) const {  message(Msg::Error,s);    }
  void warning (const string& s) const {  message(Msg::Warning,s);  }

private:
  static Msg::LevelHandler<>** mhp;  
  static void message(Msg::MessageLevel m, const string& s)
    {  if (mhp && *mhp) (**mhp)(m,s);  }
};
#endif


/* Static variables */
template<class T, class urf, class nrf> Msg::LevelHandler<>**
ValueGeneratorFactory<T,urf,nrf>::mhp=&(Msg::LevelHandler<>::default_instance());

#endif
