/** \file
  Map of ValueGenerators.

  $Header: /u/nn/cvsroot/lissom/src/valuegenmap.h,v 1.14 2002/10/29 18:32:06 jbednar Exp $
*/

#ifndef __VALUEGENMAP_H__
#define __VALUEGENMAP_H__

#include <map>
#include <string>
using std::string;
#include <algorithm>
#include <functional>

#include "valuegen.h"
#include "generic_stdlib.h"

/*
   Note: the various (*iterator).item constructions were originally
   written in the more natural form iterator->item, but they were
   changed to the uglier notation because the Cray T3E compiler (PE
   3.1) couldn't handle them.
*/

/******************************************************************************/
/* ValueGeneratorMap class hierarchy                                          */
/******************************************************************************/

/**
  This is provided for use with objects which want to maintain a
  simple database of named ValueGenerator objects.  It uses an STL map
  to provide get and set operations on parameters, and also provides
  the ValueGen operations, (making it be a ValueGen itself).  In fact,
  this should really be implemented as a ValueGenMap, since it does
  not really depend on the object types.  Unfortunately, Cray's
  compiler won't currently support clone() for such a case, and thus
  currently this is restricted to items of a single arithmetic type.
*/
template<class T, class K=string>
class ValueGeneratorMap : public ValueGen {
public:
  /** Key type of our local map. */
  typedef K key_type;

  /** Data element type of our local map. */
  typedef ValueGenerator<T>* data_type;

  /** Element type of our local map. */
  typedef std::pair<const key_type, data_type> value_type;

  /** Type of our local map. */
  typedef std::map< key_type, data_type > map_type;

  /** Map's iterator. */
  typedef typename map_type::iterator iterator;

  /** Map's iterator. */
  typedef typename map_type::const_iterator const_iterator; 
  
  ValueGeneratorMap() { }

  ValueGeneratorMap(const ValueGeneratorMap& b)  /* Copy constructor */
    : ValueGen(b) {
    for (const_iterator i=b.begin(); i!=b.end(); i++)
      set((*i).first,(*i).second->clone());
  }
  
  ~ValueGeneratorMap() {  Generic::delete_contents(m);  }

  /** Pass most things right through to the map */
  iterator       begin()       {  return m.begin();  } /**< Underlying map's begin().*/
  const_iterator begin() const {  return m.begin();  } /**< Underlying map's begin().*/
  iterator       end()         {  return m.end();    } /**< Underlying map's end().*/
  const_iterator end()   const {  return m.end();    } /**< Underlying map's end().*/
  /** Insert into underlying map */
  std::pair<iterator,bool> insert(const value_type& x)  {  return m.insert(x);  }

  /**
    Instead of directly using the map's [], which has a tricky
    semantics, or find, which is cumbersome, map access is handled
    through named get and set member functions.
  */
  void set(const key_type& k, const data_type d);
  
  /** Set only if not yet set. */
  void set_default(const key_type& k, const T d);

  /**
    Similar to map's [], but returns actual value (not pointer or
    ValueGenerator), has a default of zero if not found, and never
    adds a new entry.  See Roldan Pozo's C++ notes, lecture 4, for how
    to implement this as operator[] in a way that will still work with
    const references.
  */
  T get(const key_type& k) const;
    
  /**
    Similar to [], but adds a new ValueGenerator if the key is not
    found rather than adding a null pointer.  Unlike get() this
    returns a pointer to the actual data, and should thus only be used
    when absolutely necessary.
  */
  T* getptr(const key_type& k);

  /** Deprecated; exposes underlying datatype, e.g. for debugging */
  const data_type getvalgenptr(const key_type& k) const;

  virtual bool next();
  virtual void reset();

  /** Relink all of the ValueGenerators in this map with those in the given map */
  virtual void relink(const ValueGeneratorMap<T,K>* b);

  /** Incorporate ValueGenerators from another map */
  virtual void merge(const ValueGeneratorMap<T,K>& b);

  /** Construct and return an identical copy of self. */
  virtual ValueGeneratorMap<T,K>* clone() const { return new ValueGeneratorMap(*this); };

private:
  map_type m;
};



template<class T, class K>
void ValueGeneratorMap<T,K>::set(const key_type& k, const data_type d)
{
  iterator existing = m.find(k);
  if (existing == end())
    insert(Generic::make_pair_leftconst(k,d));
  else {
    delete (*existing).second;
    (*existing).second=d;
  }
}


template<class T, class K>
void ValueGeneratorMap<T,K>::set_default(const key_type& k, const T d)
{
  iterator existing = m.find(k);
  if (existing == end()) {
    insert(Generic::make_pair_leftconst(k,new ValueGenerator<T>(d)));
  }
}



template<class T, class K>
T ValueGeneratorMap<T,K>::get(const key_type& k) const
{
  const_iterator existing = m.find(k);
  /** Defaults to zero when not found */
  return (existing == end() ? 0 : (*existing).second->value());
}



template<class T, class K>
T* ValueGeneratorMap<T,K>::getptr(const key_type& k)
{
  const_iterator existing = m.find(k);
  data_type secptr = (existing == end() ?
			(*((insert(Generic::make_pair_leftconst(k,new ValueGenerator<T>(T())))).first)).second
			: (*existing).second);

  return secptr->valueptr();
}



template<class T, class K>
const typename ValueGeneratorMap<T,K>::data_type ValueGeneratorMap<T,K>::getvalgenptr(const key_type& k) const
{
  const_iterator existing = m.find(k);
  /** Defaults to zero when not found */
  return (existing == end() ? 0 : (*existing).second);
}



template<class T, class K>
bool ValueGeneratorMap<T,K>::next()
{
  bool status=true;
  for (iterator i=m.begin(); i!=m.end(); i++) {
    const bool newstatus = (*i).second->next();
    if (!newstatus) status=newstatus;
  }
  return status;
}



template<class T, class K>
void ValueGeneratorMap<T,K>::reset()
{
  for (iterator i=m.begin(); i!=m.end(); i++)
    (*i).second->reset();
}



template<class T, class K>
void ValueGeneratorMap<T,K>::relink(const ValueGeneratorMap<T,K>* b)
{
  if (b)
    for (iterator i=m.begin(); i!=m.end(); i++) {
      const_iterator lookup = b->m.find((*i).first);
      if (lookup != b->end())
	(*i).second->relink((*lookup).second->valueptr());
    }
}



template<class T, class K>
void ValueGeneratorMap<T,K>::merge(const ValueGeneratorMap<T,K>& b)
{
  for (const_iterator i=b.begin(); i!=b.end(); i++)
    set((*i).first,(*i).second->clone());
}
  


/** Short name for typical usage */
typedef ValueGeneratorMap<double> DGenMap;



#endif
