/** \file
  WeightSet class.
  
  $Header: /u/nn/cvsroot/lissom/src/weightset.h,v 1.6 2004/04/16 23:34:42 jbednar Exp $
*/

#ifndef __WEIGHTSET_H__
#define __WEIGHTSET_H__

#include <utility>
#include <string>
using std::string;

#include "matrix.h"
#include "boundingbox.h"
#include "scale.h"
#include "allocatingpointer.h"



/** General base class for a set of weights from one unit in a map
    (e.g. a NeuralRegion) to another */
template <class MatrixT, class BoundsT, class MapBaseType, class MapType>
class WeightSet {
public:
  typedef MatrixT Matrix;
  typedef BoundsT Bounds;
  typedef typename Matrix::size_type size_type;

  /** Constructor taking an externally-stored matrix and bounds */
  WeightSet(const Matrix& matrix, const Bounds& bounds) : m(matrix), b(bounds) {  };

  /** Constructor taking allocated matrix and bounds for us to free later */
  WeightSet(Matrix* matrix, Bounds* bounds) : m(*matrix), mptr(matrix), b(*bounds), bptr(bounds) {  };
    
  const Matrix& matrix() const { return m; }

  const Bounds& bounds() const { return b; }

  /** Allocates and returns a map with the specified weights in their effective positions. */
  template <class MapMatrix>
  MapBaseType*  as_map(const MapMatrix& reference, const Scale::Linear<double>& scale=Scale::Linear<double>()) const {
    typedef Boundary::AAArbitraryRectangle<double,int> Region;
    const Region region(b.aarectangle().corners().yl,
			b.aarectangle().corners().xl,
			b.aarectangle().corners().yl-reference.nrows(),
			b.aarectangle().corners().xl-reference.ncols());
    return new MapType(new Matrix(m),scale,region);
  }
  
private:
  const Matrix& m;  MemoryReference<Matrix> mptr;
  const Bounds& b;  MemoryReference<Bounds> bptr;
};



/** Useful class for working with WeightSets: BoundingBox returning
    inside() true for weights above the specified threshold.  Not sure
    if the various translation and rotation options will really work,
    though (jbednar 04/2004).  */
template<class T=int,class Boolean=bool,class Float=double,
	 class SBB=AARBoundingBox<T,Boolean,Float>, class WtMatrix=int >

class WeightThreshold : public BoundingBox<T,Boolean,Float> {
  typedef BoundingBox<T,Boolean,Float> base;
  typedef WeightThreshold<T,Boolean,Float,SBB,WtMatrix> self;
  typedef AARBoundingBox<T,Boolean,Float> aarbase;
  typedef typename WtMatrix::value_type value_type;
public:
  /** Default constructor; infinite extent if no arguments supplied */
  WeightThreshold(const SBB& bb_, const WtMatrix w_, const value_type threshold=0)
    : b(*bb_.clone()), w(w_), t(threshold) { }

  virtual ~WeightThreshold() {}
  
  /** Returns true if the given coordinates are above the threshold. */
  virtual Boolean inside(T x, T y) const 
    {
      typename base::Rectangle::Points     pts=aarectangle().corners();
      typename base::Rectangle::value_type xl=pts.xl;
      typename base::Rectangle::value_type yl=pts.yl;
      return (b.inside(x,y) && mat::elem(w,x-xl,y-yl)>t);
    }

  /** Scale the size of the box in X and Y dimensions */
  virtual base& scale(T xscale, T yscale)
    {  b.scale(xscale,yscale);  return *this;  }

  /** Rotate the box around the origin */
  virtual base& rotate(Float angle)
    {  b.rotate(angle);  return *this;  }

  /** Translate the origin of the box elsewhere */
  virtual base& translate(T xoffset, T yoffset)
    {  b.translate(xoffset,yoffset);  return *this;  }
  
  /** Returns an axis-aligned rectangle fully enclosing this box, or
      an empty rectangle if no such finite rectangle exists. */
  virtual typename base::Rectangle aarectangle() const {  return b.aarectangle();  }

  virtual base* clone() const {  return new self(*this); };

  virtual std::ostream& put(std::ostream &s) const
    {  s<<"(Weight threshold with underlying bounds ";  s<<b.put(s);  s<<")";  return s;  };
  
private:
  SBB& b;
  WtMatrix w;
  value_type t;
};



#endif /* __WEIGHTSET_H__ */
