/** \file
  Top-level code to generate retinal inputs representing views of a two-dimensional Cartesian world.

  $Header: /u/nn/cvsroot/lissom/src/worldviews.h,v 1.25 2003/01/11 08:15:35 jbednar Exp $
*/

#include <vector>
#include <algorithm>
#include <string>
using std::string;

#include "genericalgs.h"
#include "stringparser.h"
#include "retinalobjs.h"
#include "ipc.h"
#include "cmdparam.h"
#include "globals.h"

/******************************************************************************/
/* DistributedValueGeneratorFactory class                                     */
/******************************************************************************/


/** Multiprocessor-safe wrapper for a random distribution. */
template<class RandomGen>
struct synchronized_distribution {
double operator() ( double mean, double radius )
{
  static double value;
  int  youngest_pe    = ipc_num_processes()-1;
  bool am_youngest_pe = (ipc_my_process() == youngest_pe);
    
  /** Ensure value is ready to read */
  if (am_youngest_pe) value = RandomGen()(mean,radius);
  ipc_barrier();

  /** Ensure all PEs have a private copy of value */
  if (!am_youngest_pe) ipc_get(&value, 1, youngest_pe);
  const double valcopy=value;
  ipc_barrier();
  
  return valcopy;
}
};

/** Version of ValueGeneratorFactory that is safe for multiprocessor use. */
template<class T,
	 class urf=synchronized_distribution<Distributions::uniform>,
	 class nrf=synchronized_distribution<Distributions::normal> >
class DistributedValueGeneratorFactory
  : public ValueGeneratorFactory<T,urf,nrf> { };



/******************************************************************************/
/* WorldViews class                                                           */
/*******************************************************************************/


/** Set of Retinal_Objects representing the views seen by a set of eyes. */
class WorldViews : public ValueGen {
public:
  virtual ~WorldViews() { }

  /** Reset to initial, undefined state. */
  virtual void uninit()  {  Generic::delete_contents(contents);  }

  /** Add a new region with the given name.  If one already exists, just
      changes the height and width to the new values. */
  virtual void define( const string& regionname, const double height, const double width);

  /** Generate default contents for each eye; optional.  Usually you will want
      to check is_empty() first and call this only if so. */
  virtual void generate_default_contents();

  /** Reset each object and the stream of random numbers to their initial values. */
  virtual void reset()  {  reset(random_seed);  }

  /** Reset each object to their initial values, and set the stream of random
      numbers to the given seed value. */
  virtual void reset(long int seed) {
    std::for_each(ISEQ(contents),std::mem_fun(&Retinal_Obj::reset));
    shuffled_rand_reset(seed);
  }

  virtual bool next();

  /** \name    This interface probably needs changing; maybe some can be eliminated */
  //@{
  bool has_been_defined() const {  return !contents.empty();  }
  bool create_object(StringArgs arglist, const string& name,
		     WorldViews& reference_contents, bool link_eyes=true);
  void remove_object(const string& parent="", const string& name="");
  void print_object ( const string& parent="", const string& name="") const;
  /** Returns true if no object has been defined yet */
  bool is_empty() const {  return !std::count_if(ISEQ(contents),std::not1(std::mem_fun(&Retinal_Composite::is_empty)));  }
  //@}

  /** \name    Interface iterating over eyes; should be used only for output */
  //@{

  /** Type to hold view of world */
  typedef Retinal_Composite*          value_type;
  /** Type of container of world views */
  typedef std::vector<value_type>     vector_type;
  /** Iterator over the container of world views */
  typedef vector_type::iterator       iterator;
  /** Iterator over the container of world views */
  typedef vector_type::const_iterator const_iterator;
  /** Underlying container's begin() */
  iterator       begin()       {  return contents.begin();  }
  /** Underlying container's end()   */
  iterator       end()         {  return contents.end();    }
  /** Underlying container's begin() */
  const_iterator begin() const {  return contents.begin();  }
  /** Underlying container's end()   */
  const_iterator end()   const {  return contents.end();    } 
  //@}
  
  /** This interface should be avoided if at all possible, since
     it exposes that a WorldView is a Retinal_Object. */
  const Retinal_Obj* view(int eye) const {
    return contents[eye];
  }
  
  /** \name    External interface to a select few quantities needed by other files. */
  //@{
  
  /** Whether the given object is currently enabled in the given eye. */
  virtual bool   is_active(int objnumber, int eye) const {
    return (objnumber>= int(contents[eye]->size()) ? false :
	    ((*contents[eye])[objnumber]->get_active()));
  }

  /** The angle, in radians, of the given object in the given eye. */
  virtual double angle_of_object(int objnumber, int eye) const {
    if (objnumber< int(contents[eye]->size()))
      return contents[eye][objnumber].angle();
    else {
      const string name = contents[eye]->name();
      ipc_notify(IPC_ONE,IPC_WARNING,"Not enough objects in %s to retrieve object %d",name.c_str(),objnumber);
      return 0;
    }
  }

  /** Returns the nominal angle of the most active object at the given location in the given eye. */
  virtual double angle_of_object_at_location(int i, int j, int eye) const
    {  return contents[eye]->mostactive(i+0.5,j+0.5).angle();  }
  //@}

  
  /** Routines that parse string input for each eye call this function to make
     input_current_eye available for expressions. */
  static void set_current_eye(int eye)
    {  current_eye = eye; }

  /** This function must be called before any other member function call. */
  static void register_params_and_commands( void );

  /** Obsolete; should only be used by distribution_setfn. */
  static cmdstat set_distribution(int dist, bool& changed);
    

private:
  /** Short name for typical usage */
  typedef ValueGenerator<Retinal_Object::Variable> VGen;
  /** Short name for typical usage */
  typedef ValueGenerator_Random<Retinal_Object::Variable,synchronized_distribution<Distributions::uniform> >  UR_VGen;
  
  /** Persistent objects on the retina */
  vector_type contents;

  /* Static parameter */
  static int  random_seed;

  /* Static defaults for retinal objects */
  static string default_command;
  static string default_subcommand;

  static string default_type;
  static string default_angle;
  static string default_position;
  static string default_scale;
  static string default_offset;
  static string default_xsigma;
  static string default_ysigma;
  static string default_phase;
  static string default_size_scale;
  static string default_image_filename;
  static string default_clone_name;

  static string imagepath;

  static int    distribution;
  static double input_offset;
  static int    inputs_pereye;
  static double xsigma;
  static double ysigma;
  static int    current_eye;
  static double scale_input;
  static Tristate input_log;
  static int    input_accum_type;
  static double input_separation_max;
  static Tristate input_separation_max_enforce;
  static double input_separation_min;
  static Tristate input_separation_min_enforce;
};
