/* -*-mode: C++; compile-command: "gmake -C../tests loopspec.test"; -*- */
/**
  \file   Tests the corresponding .h file.
  $Header: /u/nn/cvsroot/lissom/tests/loopspec.c,v 1.3 2004/09/29 17:38:19 jbednar Exp $
*/


#include <vector>
#include <algorithm>
#include <functional>

#include "testutils.h"
#include "../src/loopspec.h"
#include "../src/stringutils.h"


typedef LoopSpec<double> Loop;


/** Function object with stream specification curried away to make it
    compatible with recursive_loops. */
template<class LoopContainer>
class LoopStateToStream {
public:
  LoopStateToStream(const LoopContainer& loops_, std::ostream& os_)
    : loops(loops_), os(os_) { };
  
  /** Combines existing value with new one. */
  void operator() () {
    for (typename LoopContainer::const_iterator i=loops.begin();
	 i!=loops.end(); i++)
      os << (**i) << " ";
    os << "; ";
  }

private:
  const LoopContainer& loops;
  std::ostream& os;
};


/** Compares the output of fixed loops, while loops, and arbitrarily
    recursive loops.  Uses obsolete strstream to work with existing
    compilers that do not implement stringstream. */
void test_loops(int c1, int c2, int c3)
{
  Loop x("x",c1);
  Loop y("y",c2);
  Loop z("z",c3);

  /* Hardcoded for loops */
  std::ostringstream hardcoded_for_loop_stream;
  for (int i=0; i<c1; i++)
    for (int j=0; j<c2; j++)
	for (int k=0; k<c3; k++)
	  hardcoded_for_loop_stream
	    <<  "x=" << i/double(c1)
	    << " y=" << j/double(c2)
	    << " z=" << k/double(c3) << " ; ";;
  hardcoded_for_loop_stream << std::ends;
  

  /* LoopSpec while loops */
  std::ostringstream loopspec_while_loop_stream;
  x.reset();
  while (x.valid()) {
    y.reset();
    while (y.valid()) {
      z.reset();
      while (z.valid()) {
	loopspec_while_loop_stream << x << " " << y << " " << z << " ; ";
	z.next();
      }
      y.next();
    }
    x.next();
  }
  loopspec_while_loop_stream << std::ends;

  
  /* Recursive LoopSpec loops*/
  std::ostringstream recursive_loop_stream;
  typedef  std::vector<Loop*> Loops;
  Loops loops;
  loops.push_back(&x);
  loops.push_back(&y);
  loops.push_back(&z);
  LoopStateToStream<Loops> iteration(loops,recursive_loop_stream);
  LoopSpecs_ExecuteNested(loops, loops.begin(), iteration);
  recursive_loop_stream << std::ends;

  
  /* Output */
  section(("Testing loop "+String::stringrep(c1)+","+String::stringrep(c2)+","+String::stringrep(c3)).c_str());
  const string loopspec_while_loop_string(loopspec_while_loop_stream.str());
  const string hardcoded_for_loop_string(hardcoded_for_loop_stream.str());
  const string recursive_loop_string(recursive_loop_stream.str());

  EXACT_MATCH(loopspec_while_loop_string, hardcoded_for_loop_string);
  EXACT_MATCH(recursive_loop_string,      hardcoded_for_loop_string);
}


int main(int, char**)
{
  section("Sizes");
#ifdef _CRAYMPP
#else  
  EXACT_CHECK( int(sizeof(Loop)), 28);
#endif  

  test_loops(0,0,0);
  test_loops(0,0,1);
  test_loops(1,0,0);
  test_loops(1,0,1);
  test_loops(1,1,0);
  test_loops(1,1,1);
  test_loops(2,1,1);
  test_loops(1,2,1);
  test_loops(1,1,2);
  test_loops(1,2,3);
  test_loops(3,2,1);
  test_loops(2,1,1);
  test_loops(3,2,4);

  section("ValueGenerator");
  {
    double start=100;
    double increment=2.5;
    ValueGenerator_Increment<double> vg(start-increment,increment);
    Loop w("w",7,true,&vg);
    w.reset();
    while (w.valid()) {
      FLOAT_MATCH(w.value(), start+increment*w.count());
      w.next();
    }
  }
  
  return count_errors();
}
