/* -*-mode: C++; compile-command: "gmake -C../tests plotspecification.test"; -*- */
/**
  \file   Tests the corresponding .h file.
  $Header: /u/nn/cvsroot/lissom/tests/plotspecification.c,v 1.8 2001/05/16 01:18:44 jbednar Exp $
*/

#include <utility>
#include <string>
#include <assert.h>

#include "testutils.h"
#include "pixelutils.h"

#include "../src/matriximagefactory.h"
#include "../src/plotspecification.h"
#include "../src/aarectangle.h"
#include "../src/scale.h"

#ifdef EXPLICIT_INSTANTIATION_FOR_STATICS
template class Plot::HSVPixel<Plot::Pixel::FComponent,Plot::Pixel::HueComponent>;
template class Msg::LevelHandler<Msg::MessageLevel,Msg::NumMessageLevels>;
template class PolymorphBase<int>;
template class PolymorphBase<string>;
#endif 


/* Option to test ParameterMap class; otherwise use simple substitute defined here. */
//#define USE_PARAMETERMAP
#ifdef  USE_PARAMETERMAP
#include "../src/parametermap.h"

#else
#include "../src/genericalgs.h"
#include <vector>


/** Simple concrete ParamMap implemented with a PointerMap.  
    Failing lookups fall back to the parent map, if any.  */
class SimpleParamMap : public ParamMap {
public:
  typedef  SimpleParamMap self;
  typedef  Typeless T;
  virtual ~SimpleParamMap() {  Generic::delete_contents(allocated);  }

  SimpleParamMap(self* parent_=0, const string name_="") : parent(parent_)
    {  if (parent) parent->children.set(name_,this);  }

  virtual T* getptr(const key_type& k)
    {        T* p=m.getptr(k);  return (p? p : (parent? parent->getptr(k) : 0));  }
    
  virtual const T* getptr(const key_type& k) const
    {  const T* p=m.getptr(k);  return (p? p : (parent? parent->getptr(k) : 0));  }

  virtual ParamMap* new_child(const string name_="", const bool owned=false) {
    self* p=new self(this,name_);
    if (owned) allocated.push_back(p);
    return p;
  }

  virtual       ParamMap& get_child(const string& name_="")
    {        ParamMap* p=children.getptr(name_);  return (p? *p : *this);  }
  
  virtual const ParamMap& get_child(const string& name_="") const
    {  const ParamMap* p=children.getptr(name_);  return (p? *p : *this);  }
  
  template <class ST>
  self& set_local(const string& name, const ST& value)
    {  m.set(name,new Polymorph<ST>(value));  return *this;  }

private:
  PointerMap<T> m;
  PointerMap<ParamMap,string,false> children;
  std::vector<ParamMap*> allocated;
  self* parent;
};

#endif



/* Wrappers for setting and definition of specific ParamMap classes.
   The Parametermap<> class requires requires that any params to
   be set are defined beforehand, so any such params should be
   defined here.
*/
#ifndef USE_PARAMETERMAP    
#define PARAMMAPTYPE SimpleParamMap
#define PARAMMAP(name) PARAMMAPTYPE name
#define SET(name,val)  set_local(name,val)

#else
#define PARAMMAPTYPE ParameterMap< ParameterDefinition<> >
#define PARAMMAP(name) PARAMMAPTYPE name;\
 name.define_param(ParameterDefinition<>( allocating_Polymorph(string("")),"name"))
#define SET(name,val) define_param(ParameterDefinition<>(allocating_Polymorph(val),name))
#endif



typedef Scale::Linear<float>                LScale;
typedef MatrixType<float>::rectangular      Matrix;
typedef mat::MatrixInterfaceAdapter<Matrix> Adapter;
typedef mat::MatrixInterface<float>         AdapterBase;



int main(int, char**)
{
  section("Construction");
  const Matrix::size_type width=10;
  const Matrix::size_type height=20;
  Matrix lgaussian (height,width);
  Matrix tiltsine  (height,width);
  Matrix sgaussian (height,width);
  Matrix cosine    (height,width);
  
  const double sig=width/5;
  const double sig_sq=sig*sig;
  const double PI=3.314159;
  
  for (Matrix::size_type r=0; r<height; r++)
    for (Matrix::size_type c=0; c<width; c++) {
      const double dc=c-width/2.0;
      const double dr=r-height/2.0;
      const double rad_sq=dc*dc+dr*dr;
      
      lgaussian[r][c]   = exp(-rad_sq/(2*PI*sig_sq)); /* Large Gaussian */
      tiltsine[r][c]    = sin(r/4.0+c/6.0);           /* Tilted sine wave */
      sgaussian[r][c]   = exp(-rad_sq/(PI*sig_sq));   /* Small Gaussian */
      cosine[r][c]      = cos(r);                     /* Cosine wave */
    }


  /* Using MatrixImageFactory directly */
  section("MatrixImageFactory");
  {
    /* These adapters are required for MatrixImageFactory */
    Adapter  lg(lgaussian);
    Adapter  ts(tiltsine);
    Adapter  sg(sgaussian);
      
    /* Three plots with gradually increasing information content */
    const Plot::MatrixImageFactory<AdapterBase> videofactory;
    Plot::MatrixImage<>* imgs    = videofactory(&lg,LScale());
    Plot::MatrixImage<>* imgsc   = videofactory(&lg,LScale(),&ts,LScale());
    Plot::MatrixImage<>* imgscc  = videofactory(&lg,LScale(),&ts,LScale(),&sg,LScale());
    
    /* Versions of the same with paper-based colors */
    const Plot::MatrixImageFactory<AdapterBase> paperfactory(true);
    Plot::MatrixImage<>* imgsp   = paperfactory(&lg,LScale());
    Plot::MatrixImage<>* imgscp  = paperfactory(&lg,LScale(),&ts,LScale());
    Plot::MatrixImage<>* imgsccp = paperfactory(&lg,LScale(),&ts,LScale(),&sg,LScale());
  
    assert (imgs);
    assert (imgsc);
    assert (imgscc);
    
    section("Image file output");
    PPM_WRITE("mif_s",    (*imgs),   "MatrixImageFactory strength plot");
    PPM_WRITE("mif_sc",   (*imgsc),  "MatrixImageFactory strength and color plot");
    PPM_WRITE("mif_scc",  (*imgscc), "MatrixImageFactory strength, color, and confidence plot");
    PPM_WRITE("mif_sp",   (*imgsp),  "MatrixImageFactory strength plot (paper-based)");
    PPM_WRITE("mif_scp",  (*imgscp), "MatrixImageFactory strength and color plot (paper-based)");
    PPM_WRITE("mif_sccp", (*imgsccp),"MatrixImageFactory strength, color, and confidence plot (paper-based)");
  }
  

  /* Using MatrixImageFactory through a PlotSpecification object */
  section("PlotSpecification");
  {
    typedef Plot::PlotSpecification<Adapter,string,LScale> PlotSpec;
    typedef PlotSpec::SubRegion SubRegion;

    /* Define named matrices  */
    PlotSpec::MatrixTable matrices;
    matrices.set("LGaussian",  new Adapter(lgaussian));
    matrices.set("TiltSine",   new Adapter(tiltsine));
    matrices.set("SGaussian",  new Adapter(sgaussian));
    matrices.set("Cosine",     new Adapter(cosine));
  
    /* Define plots selecting matrices by name */
    PARAMMAP(params);
    PARAMMAP(params2);

    /* Select the region 3--10 units from left, and 6--8 units from the top */
    params2.SET("start_c", 3);
    params2.SET("start_r", 6);
    params2.SET("end_c",  10);
    params2.SET("end_r",   8);

    /* Specify and create a variety of plots */
    PlotSpec* specscc = new PlotSpec("",matrices,"CombinedPlot",params, "LGaussian","TiltSine","SGaussian");
    PlotSpec* specoc  = new PlotSpec("",matrices,"OtherPlot",   params, "Cosine",   "TiltSine");
    PlotSpec* specocr = new PlotSpec("",matrices,"OtherPlot2",  params2,"Cosine",   "TiltSine");
    PlotSpec* specref = new PlotSpec("",matrices,"CombinedPlot",params, "LGaussian","TiltSine","SGaussian");

    assert (specscc); Plot::MatrixImage<>* psimgscc = specscc->bitmap_image(); delete specscc;  assert (psimgscc);
    assert (specoc);  Plot::MatrixImage<>* psimgoc  = specoc->bitmap_image();  delete specoc;   assert (psimgoc);
    assert (specocr); Plot::MatrixImage<>* psimgocr = specocr->bitmap_image(); delete specocr;  assert (psimgocr);

    /* Specify and create reference plot */
    matrices.set("__StrengthReference", new Adapter(cosine));
    assert (specref); Plot::MatrixImage<>* psimgref = specref->bitmap_image(); delete specref;  assert (psimgref);

    section("Image file output");
    
    PPM_WRITE("plotspec_scc", (*psimgscc), "PlotSpecification strength, color, and confidence plot");
    PPM_WRITE("plotspec_oc",  (*psimgoc),  "PlotSpecification other plot with same colors");
    PPM_WRITE("plotspec_ocr", (*psimgocr), "PlotSpecification other plot with same colors; subregion only");
    PPM_WRITE("plotspec_ref", (*psimgref), "PlotSpecification strength, color, and confidence plot relative to another");
  }
  
  return count_errors();
}
  
