/** \file
  Routines for testing the functionality and correctness of other files.

  $Header: /u/nn/cvsroot/lissom/tests/testutils.h,v 1.14 2001/07/29 15:42:31 jbednar Exp $
*/

#ifndef __TESTUTILS_H__
#define __TESTUTILS_H__

#include <iostream>
#include <fstream>
#include <string>
#include <cmath>
#include <complex>
#include <cstdlib>


/** Provides support for subtracting strings so that the same compare
    function below will work with them */
int operator-(const std::string&A, const std::string& B) {  return (A==B? 0 : 1);  }




/** (Optionally) increments the number of errors so far and returns
    the previous value. 
*/
inline size_t num_errors(bool increment=false)
{
  static size_t num=0;
  return (increment? num++ : num);
}


/** Report and return the number of errors encountered so far. */
inline size_t count_errors()
{
  static size_t num=num_errors();
  std::ostream& os=(num==0? std::cout : std::cerr);

#ifndef ERRORS_ONLY
  os << std::endl;
#endif  
  os << "There were " << num << " unexpected errors." << std::endl;
  return num;
}



/** Compares the given value to the expected value,
    printing the value and (on mismatch) an error message. */
template <class C, class D, class E>
void compare_with_expected(const char* description, C value, D expected,
			   E tolerance, bool errors_only=false, bool warn_only=false,
			   bool expect_non_match=false)
{
  bool matches=( std::abs(value-expected) <= tolerance );
  num_errors(!warn_only && !matches);
  std::ostream& os=(matches? std::cout : std::cerr);

#ifdef ERRORS_ONLY /* Force true */
  errors_only=true;
#endif
  
  if (expect_non_match) matches = !matches; // Flip the sense 

  if (!matches || !errors_only) {
    os << ( matches? "       " : (warn_only? "Note: " : "Error: ") )
       << description << ":	" << value;
  
    if (!matches)
      os << "	" << (expect_non_match? "UNEXPECTEDLY MATCHES" : "DOES NOT MATCH expected")
	 << " value of " << expected << ".";
    
    os << std::endl;
  }
}



/** Separates a group of tests from others */
inline void section(const char* name)
{
  (void)name;
#ifndef ERRORS_ONLY
  std::cout << std::endl << name << std::endl;
#endif
}



/** Call compare_with_expected using an expression for the value as the description */
#define EXACT_MATCH(value,expected)\
compare_with_expected(#value,value,expected,0,false)


/** Same as EXACT_MATCH except that a looser tolerance is used, appropriate for floating-point comparisons */
#define FLOAT_MATCH(value,expected)\
compare_with_expected(#value,value,expected,0.00001,false)


/** Same as FLOAT_MATCH except that an arbitrarily loose tolerance can be specified */
#define LOOSE_MATCH(value,expected,tolerance)\
compare_with_expected(#value,value,expected,tolerance,false)


/** Like LOOSE_MATCH, but warns if the value matches exactly instead of only to
    within the given tolerance. */
#define RANDOM_MATCH(value,expected,tolerance)\
compare_with_expected(#value,value,expected,tolerance,false); \
compare_with_expected(#value,value,expected,0.000001,false,true,true)


/** Same as EXACT_MATCH except that only a warning is printed.
    Useful for e.g. checking sizes, since one would want to know
    when they change yet a change is not necessarily an error. */
#define EXACT_CHECK(value,expected)\
compare_with_expected(#value,value,expected,0,false,true)


#endif /* __TESTUTILS_H__ */
